/*
// NAME:	   	Com.C
//
// COMMENTS:
//		 Example of a user file format plugin. This is a made-up format
//		 called .com (as in compact), which is defined as 1 byte per 
//		 pixel, where each byte is 3 bits red, 3 bits green and 2 bits blue.
//		 There is a brief header of 128 bytes - the first 4 bytes are the
//		 X dimension of the image, the next 4 bytes are the Y dimension,
//		 and the rest is unused.
*/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>

#include "Com.h"

#define SWAP16(x) x = cpiSwapShort(x)
#define SWAP32(x) x = cpiSwapLong(x)
#define SWAPFLOAT(x) x = cpiSwapFloat(x)

//////////////////////////////////////


static CPI_Bool
ComInit( CPI_Bool /* initforread */ )
{
	return CPI_TRUE;
}


//////////////////////////////////////


static CPI_Bool
ComShutDown( void )
{
	return CPI_TRUE;
}


//////////////////////////////////////


static CPI_PrivateData
ComOpenForRead( const CPI_Metadata    meta,
				CPI_Metadata          imageinfo )
{
	imageState  *retval     = (imageState *)cpiAlloc( sizeof( imageState ) );

	if ( NULL == retval )
		return NULL;

	retval->myDevice = cpiOpen( meta );
	if ( NULL == retval->myDevice )
	{
		cpiError( "Unable to open file for read" );
		cpiFree( retval );
		return NULL;
	}

	if ( sizeof( ComHeader ) != cpiRead( retval->myDevice,
										 (char *)&retval->myHeader,
										 HEADERSIZE ) )
	{
		cpiError( "Couldn't read com image header!" );
		cpiClose( retval->myDevice );
		cpiFree( retval );

		retval = NULL;
	}
	else
	{
		if ( cpiIsLittleEndian() )
			SwapComHeader( &retval->myHeader );

		if ( retval->myHeader.magic == MAGICNUM )
		{
			cpiSetImageFullSize( imageinfo, retval->myHeader.xsize,
								 retval->myHeader.ysize );
			/* these params are fixed by the format definition */
			cpiSetImageBitDepth( imageinfo, 8 );
			cpiSetImageChannels( imageinfo, 3 );
		}
		else
		{
			cpiError( "Magic number doesn't match!" );
			cpiClose( retval->myDevice );
			cpiFree( retval );

			retval = NULL;
		}
	}

	return retval;
}


/*
//////////////////////////////////////
	This format only returns an 8bit image - requests by the user
	for 16 or float versions are handled outside the API, by Rayz.

	For most formats, reading subimages is more efficient if it
	is handled here too if possible. Also, we prefer to
	read in images on a scanline basis, even if (as here) we could
	just read in the whole thing at once. This is for memory
	efficiency.
*/

static CPI_Bool
ComReadImage( CPI_PrivateData handle, CPI_Image *image )
{
	CPI_Uint8	*packedPtr;
	CPI_Uint32	 offset, y;
	imageState  *state		= (imageState *)handle;
	CPI_Uint8	*dataPtr	= (CPI_Uint8 *)image->data;
	CPI_Uint32	 rowSize	= image->myContext.sizeX;
	CPI_Uint32	 end		= image->myContext.offsetY +
							  image->myContext.sizeY;

	/* sanity check */
	if ( image == NULL )
	{
		cpiError( "ComReadImage: destination image is NULL" );
		return CPI_FALSE;
	}

	/* allocate temporary line buffer */
	packedPtr = (CPI_Uint8 *)cpiAlloc( rowSize );

	/* compute offset into file and seek to start of data */
	offset = HEADERSIZE + image->myContext.offsetX + 
			 (image->myContext.offsetY * rowSize);
	cpiSetPosition( state->myDevice, offset, CPI_SEEK_START );

	/* read in each row, and then unpack - the call to cpiGetLine
	   is a safe way to position the data pointer for each row */
	for (y = image->myContext.offsetY; y < end; y++)
	{
		cpiRead( state->myDevice, (char *)packedPtr, rowSize );
		dataPtr = (CPI_Uint8 *)cpiGetLine( image, y);
		UnPackCom( packedPtr, dataPtr, rowSize );
	}

	cpiFree( packedPtr );

	return CPI_TRUE;
}


//////////////////////////////////////
/* open the file, set up and write the header */

static CPI_PrivateData
ComOpenForWrite( const CPI_Metadata		meta,
				 CPI_Metadata			imageinfo )
{
	ComHeader	 header;
	CPI_Int32	 width, height;
	imageState	*retval = (imageState *)cpiAlloc( sizeof( imageState ) );

	retval->myDevice = cpiOpen( meta );

	if ( retval->myDevice == NULL )
	{
		cpiError( "Couldn't open file for write!" );
		cpiFree( retval );
		return NULL;
	}

	/* set up header and write it */
	cpiGetImageFullSize( imageinfo, &width, &height );
	header.xsize = (CPI_Uint32)width;
	header.ysize = (CPI_Uint32)height;

	if ( cpiIsLittleEndian() )
		SwapComHeader( &header );

	if ( sizeof( ComHeader ) != cpiWrite( retval->myDevice,
										  (char *)&header,
										  sizeof( ComHeader ) ) )
	{
		cpiError( "Couldn't write Com Header!" );
		cpiClose( retval->myDevice );
		cpiFree( retval );
		retval = NULL;
	}

	return (CPI_PrivateData *)retval;
}


//////////////////////////////////////
// Output size is just sizeX * sizeY bytes, regardless of number of
// channels or bit depth

static CPI_Bool
ComWriteImage( CPI_PrivateData handle, CPI_Image *image )
{
	imageState	*state	= (imageState *)handle;
	CPI_Bool	 retval = CPI_TRUE;
	CPI_Uint8	*myPackedPtr;
	CPI_Uint64	 writeSize;

	if ( image != NULL )
	{
		// set up output size
		writeSize = image->myContext.sizeX * image->myContext.sizeY *
					sizeof( CPI_Uint8 );

		// allocate buffer for output image
		myPackedPtr = (CPI_Uint8 *)cpiAlloc( writeSize );

		// convert to .com format
		if ( image->myContext.bitsPerPel == 8 )
			PackCom8( myPackedPtr, (CPI_Uint8 *)image->data, 
					writeSize, image->myContext.channels );
		else if ( image->myContext.bitsPerPel == 16 )
			PackCom16( myPackedPtr, (CPI_Uint16 *)image->data, 
					writeSize, image->myContext.channels );
		else
			PackComF( myPackedPtr, (CPI_Float32 *)image->data, 
					writeSize, image->myContext.channels );

		if ( writeSize != cpiWrite( state->myDevice,
									(char *)myPackedPtr,
									writeSize ) )
		{
			cpiError ( "Couldn't write all image data" );
			retval = CPI_FALSE;
		}

		cpiFree( myPackedPtr );
	}
	else
	{
		cpiError( "Can't write NULL COM image!" );
		retval = CPI_FALSE;
	}

	return retval;
}


//////////////////////////////////////


static CPI_Bool
ComCloseImage( CPI_PrivateData handle )
{
	imageState *state = (imageState *)handle;
	
	cpiClose( state->myDevice );
	cpiFree( state );

	return CPI_TRUE;
}


/////////////////////////////////////


static RPI_ImageInfo dinfo = {
	{
		"com",
		"Com",
		"Silicon Grail",
		"1.0", NULL, NULL
	},
	ComInit,
	ComShutDown,
	NULL,				//ComParmChanged
	ComOpenForRead,
	ComReadImage,
	ComReadImage,		//ComReadSubImage
	ComOpenForWrite,
	ComWriteImage,
	NULL,				//WriteSubImage
	ComCloseImage,
	NULL,
	CPI_IMAGE_RGB | CPI_IMAGE_UINT8,
	"com"
};


//////////////////////////////////////


CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	return cpiRegisterImageReader( &dinfo );
}


//////////////////////////////////////


CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageReader( "com" );
}


////////////////////////////////////////


static void
SwapComHeader( ComHeader *header )
{
	SWAP32( header->magic );
	SWAP32( header->xsize );
	SWAP32( header->ysize );
}


//////////////////////////////////////
// These routines take the 8, 16 or float image data and pack it
// into the .com format

// 8 bit source image - take 3 MSB of red, 3 MSB of green and 2 of blue

static CPI_Bool
PackCom8( CPI_Uint8	 *packed,
		  CPI_Uint8  *unpacked,
		  CPI_Uint64  pixels,
		  CPI_Int32	  channels )
{
	CPI_Uint8	r, g, b;

	for ( CPI_Uint32 i = 0; i < pixels; i++ )
	{
		g = b = 0;
		r = ( unpacked[0] >> 5 ) & 0x07;
		if ( channels > 1 )
			g = ( unpacked[1] >> 5 ) & 0x07;
		if ( channels > 2 )
			b = ( unpacked[2] >> 6 ) & 0x03;

		*packed++ = ( r << 5 ) | ( g << 2 ) |  b;

		 unpacked += channels;
	}

	return CPI_TRUE;
}

// 16 bit source image - take 3 MSB of red, 3 MSB of green and 2 of blue

static CPI_Bool
PackCom16( CPI_Uint8	*packed,
		   CPI_Uint16 	*unpacked,
		   CPI_Uint64    pixels,
		   CPI_Int32	 channels )
{
	CPI_Uint8	r, g, b;

	for ( CPI_Uint32 i = 0; i < pixels; i++ )
	{
		g = b = 0;
		r = ( unpacked[0] >> 13 ) & 0x07;
		if ( channels > 1 )
			g = ( unpacked[1] >> 13 ) & 0x07;
		if ( channels > 2 )
			b = ( unpacked[2] >> 14 ) & 0x03;

		*packed++ = ( r << 5 ) | ( g << 2 ) |  b;

		 unpacked += channels;
	}

	return CPI_TRUE;
}

// floating point source image

static CPI_Bool
PackComF( CPI_Uint8	 	*packed,
		  CPI_Float32	*unpacked,
		  CPI_Uint64     pixels,
		  CPI_Int32	  	 channels )
{
	CPI_Uint8	r, g, b;

	for ( CPI_Uint32 i = 0; i < pixels; i++ )
	{
		g = b = 0;
		r = ( (CPI_Uint8)( cpiClampFloat( unpacked[0], 0.0F, 1.0F ) 
							* 7.5 ) ) & 0x07;
		if ( channels > 1 )
			g = ( (CPI_Uint8)( cpiClampFloat( unpacked[1], 0.0F, 1.0F ) 
								* 7.5 ) ) & 0x07;
		if ( channels > 2 )
			b = ( (CPI_Uint8)( cpiClampFloat( unpacked[2], 0.0F, 1.0F ) 
								* 3.5 ) ) & 0x03;

		*packed++ = ( r << 5 ) | ( g << 2 ) |  b;

		 unpacked += channels;
	}

	return CPI_TRUE;
}

/*
   This reads in the packed data and unpacks it into an 8bit per pixel
   RGB image
*/

static CPI_Bool
UnPackCom( CPI_Uint8	*packed,
		   CPI_Uint8 	*unpacked,
		   CPI_Uint64	 pixels )
{
	CPI_Uint8	r, g, b;

	for ( CPI_Uint32 i = 0; i < pixels; i++ )
	{
		r = g = b = 0;
		r = ( *packed >> 5 ) & 0x07;
		g = ( *packed >> 2 ) & 0x07;
		b = *packed & 0x03;

		*unpacked++ = r << 5;
		*unpacked++ = g << 5;
		*unpacked++ = b << 6;

		packed++;
	}

	return CPI_TRUE;
}
