/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Tim Connolly
//  Silicon Grail Corporation
//  710 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:		Jpeg.C
//
// COMMENTS:
//		 The Jpeg File Reader
*/

#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>


#include <CPI/CPI.h>
#ifdef __cplusplus
extern "C" {
#endif

#include <jpeg/jinclude.h>
#include <jpeg/jpeglib.h>
#include <jpeg/jerror.h>

#ifdef __cplusplus
}
#endif


////////////////////////////////////////


typedef struct _jpegReadFile
{
	struct jpeg_source_mgr	 myMembers;
	CPI_DeviceHandle		 myDevice;
	JOCTET					*myBuf;
} jpegReadFile;


////////////////////////////////////////


typedef struct _jpegWriteFile
{
	struct jpeg_destination_mgr	 myMembers;
	CPI_DeviceHandle			 myDevice;
	JOCTET						*myBuf;
} jpegWriteFile;


////////////////////////////////////////


typedef struct _imageState
{
	CPI_DeviceHandle		 myDevice;
	j_compress_ptr			 writeJpeg;
	j_decompress_ptr		 readJpeg;
	struct jpeg_error_mgr	 jpegErr;
	CPI_Uint32				 rows;
	CPI_Uint32				 columns;
	JSAMPROW				 cacheBuffer;
	JSAMPARRAY				 cacheRows;
	CPI_Bool				 myIsStarted;
} imageState;

static CPI_Bool			 JpegInit( CPI_Bool initforread );
static CPI_Bool			 JpegShutDown( void );
static CPI_PrivateData	 JpegOpenForRead( const CPI_Metadata	 meta,
										  CPI_Metadata			 imageinfo );
static CPI_Bool			 JpegReadImage( CPI_PrivateData		 handle,
										CPI_Image			*image );
static CPI_Bool			 JpegReadSubImage( CPI_PrivateData	 handle,
										   CPI_Image		*image );
static CPI_PrivateData	 JpegOpenForWrite( const CPI_Metadata	 meta,
										   const CPI_Metadata	 imginfo );

static CPI_Bool			 JpegWriteImage( CPI_PrivateData	 handle,
										 CPI_Image			*image );

static CPI_Bool			 JpegCloseImage( CPI_PrivateData handle );


////////////////////////////////////////


// Read Functions
// 4K seems to be a good size to read...
#define INPUT_BUF_SIZE  4096
static void		 jpegStartRead( struct jpeg_decompress_struct	*jinfo,
								CPI_DeviceHandle				 device );
static void		 jpegInitRead( j_decompress_ptr info );
static boolean	 jpegFillInputBuffer( j_decompress_ptr info );
static void		 jpegSkipInputData( j_decompress_ptr info, long num_bytes );
static void		 jpegFinishRead( j_decompress_ptr info );

// Write Functions
#define OUTPUT_BUF_SIZE 4096
static void		 jpegStartWrite( struct jpeg_compress_struct	*jinfo,
								 CPI_DeviceHandle				 device,
								 CPI_Int32						 quality,
								 CPI_Int32						 width,
								 CPI_Int32						 height,
								 CPI_Int32						 channels );
static void		 jpegInitWrite( j_compress_ptr info );
static boolean	 jpegEmptyOutputBuffer( j_compress_ptr info );
static void		 jpegFinishWrite( j_compress_ptr info );

////////////////////////////////////////


static jmp_buf jpegJmpBuffer;


////////////////////////////////////////


static void
jpeg_error_message( j_common_ptr info )
{
	static char errBuf[ JMSG_LENGTH_MAX ];

	info->err->format_message( info, errBuf );
	cpiError( errBuf );
}


////////////////////////////////////////


static void
jpeg_error_exit( j_common_ptr info )
{
	jpeg_error_message( info );
	// This routine can't return to the caller, so we are forced to
	// use a jump...
	longjmp( jpegJmpBuffer, 1 );
}


//////////////////////////////////////


static CPI_Bool
JpegInit( CPI_Bool initforread )
{
	if ( CPI_FALSE == initforread )
		cpiAddInteger( "quality",
					   "Jpeg Quality Level",
					   90, 
					   1, 
					   100, 
					   CPI_PARM_DEFAULT,
					   "Adjust the quality of the jpeg image to be written",
					   NULL );

	return CPI_TRUE;
}


//////////////////////////////////////////////

static CPI_Bool
JpegShutDown( void )
{
	return CPI_TRUE;
}


////////////////////////////////////////////


static CPI_PrivateData
JpegOpenForRead( const CPI_Metadata	 meta,
				 CPI_Metadata		 imageinfo )
{
	char		 filename[CPI_PATH_LENGTH];
	imageState	*retval		= (imageState *)cpiAlloc( sizeof ( imageState ) );
	imageState	*cleanptr	= NULL;
	if ( NULL == retval )
		return NULL;

	cpiGetFilename( meta, filename, CPI_PATH_LENGTH );
	retval->myDevice = cpiOpen( meta );
	if ( NULL == retval->myDevice )
	{
		cpiError( "Unable to open file for read" );
		cpiFree( retval );
		return NULL;
	}

	retval->readJpeg	= ( struct jpeg_decompress_struct * ) cpiAlloc(
		sizeof( struct jpeg_decompress_struct ) );
	retval->writeJpeg	= NULL;

	if ( NULL == retval->readJpeg )
	{
		cpiError( "Unable to allocate memory" );
		cpiClose( retval->myDevice );
		cpiFree( retval );
		return NULL;
	}

	retval->readJpeg->err			= jpeg_std_error( &(retval->jpegErr) );
	retval->jpegErr.error_exit		= jpeg_error_exit;
	retval->jpegErr.output_message	= jpeg_error_message;
	retval->myIsStarted				= CPI_FALSE;

	jpeg_create_decompress( retval->readJpeg );

	if ( setjmp( jpegJmpBuffer ) )
	{
		cleanptr	= retval;
		retval		= NULL;
	}

	if ( NULL != retval )
	{
		jpegStartRead( retval->readJpeg, retval->myDevice );

		if ( JPEG_HEADER_OK == jpeg_read_header( retval->readJpeg, true ) )
		{
			cpiSetImageFullSize( imageinfo, retval->readJpeg->image_width,
								 retval->readJpeg->image_height );
			cpiSetImageChannels( imageinfo, retval->readJpeg->num_components );
			cpiSetImageBitDepth( imageinfo, retval->readJpeg->data_precision );
		}
		else
		{
			cleanptr	= retval;
			retval		= NULL;
		}
	}

	if ( NULL != cleanptr )
	{
		jpegReadFile	*reader;

		jpeg_abort_decompress( cleanptr->readJpeg );

		reader = (jpegReadFile *)cleanptr->readJpeg->src;
		if ( NULL != reader->myBuf )
			cpiFree( reader->myBuf );

		if ( NULL != reader )
			cpiFree( reader );

		jpeg_destroy_decompress( cleanptr->readJpeg );
		cpiFree( cleanptr->readJpeg );

		cpiClose( cleanptr->myDevice );
		cpiFree( cleanptr );

		cpiError( "Couldn't retrieve JPEG header" );

		retval = NULL;
	}

	return (CPI_PrivateData)retval;
}


//////////////////////////////////////////////////////////


static CPI_Bool
JpegReadImage( CPI_PrivateData handle, CPI_Image *image )
{
	CPI_Bool	 retval = CPI_FALSE;
	imageState	*state = (imageState *)handle;
	CPI_Int32	 y;
	JSAMPLE		*linebuf;

	if ( setjmp( jpegJmpBuffer ) )
	{
		return CPI_FALSE;
	}

	if ( NULL != image && NULL != state->readJpeg )
	{
		if ( 8 != image->myContext.bitsPerPel )
		{
			cpiError( "Image bit depth doesn't match file" );
			return CPI_FALSE;
		}

		if ( CPI_Int32( state->readJpeg->num_components ) !=
			 image->myContext.channels )
		{
			cpiError( "Image number of channels doesn't match file" );
			return CPI_FALSE;
		}

		if ( CPI_Int32( state->readJpeg->image_width ) !=
			 image->myContext.sizeX )
		{
			cpiError( "Image width doesn't match file width" );
			return CPI_FALSE;
		}

		if ( CPI_Int32( state->readJpeg->image_height ) !=
			 image->myContext.sizeY )
		{
			cpiError( "Image height doens't match file height" );
			return CPI_FALSE;
		}

		if ( 0 != state->readJpeg->output_scanline )
		{
			cpiSetPosition( state->myDevice, 0, CPI_SEEK_START );

			jpeg_abort_decompress( state->readJpeg );

			jpegStartRead( state->readJpeg, state->myDevice );

			if ( JPEG_HEADER_OK == jpeg_read_header( state->readJpeg, true ) )
			{
				state->myIsStarted	= CPI_FALSE;
				retval				= CPI_TRUE;
			}
			else
				retval = CPI_FALSE;
		}
		else
			retval = CPI_TRUE;

		if ( retval && ! state->myIsStarted )
		{
			jpeg_start_decompress( state->readJpeg );
			state->myIsStarted = CPI_TRUE;
		}

		for ( y = 0; retval && y < image->myContext.sizeY; y++ )
		{
			linebuf = (JSAMPLE *)cpiGetLine( image, y );
			if ( 1 != jpeg_read_scanlines( state->readJpeg, &linebuf, 1 ) )
			{
				retval = CPI_FALSE;
				break;
			}
		}

		if ( retval )
			jpeg_finish_decompress( state->readJpeg );
		else
			jpeg_abort_decompress( state->readJpeg );

		state->myIsStarted	= CPI_FALSE;
	}
	else
	{
		cpiError( "JpegReadImage: destination image is NULL" );
		retval = CPI_FALSE;
	}

	return retval;
}


////////////////////////////////////////


static CPI_Bool
JpegReadSubImage( CPI_PrivateData	 handle,
				  CPI_Image			*image )
{
	CPI_Bool	 retval = CPI_FALSE;
	imageState	*state = (imageState *)handle;
	CPI_Int32	 y;
	JSAMPLE		*linebuf;

	if ( NULL != image && image->myContext.sizeY == 0 )
		return CPI_TRUE;

	if ( setjmp( jpegJmpBuffer ) )
	{
		return CPI_FALSE;
	}

	if ( NULL != image && NULL != state->readJpeg )
	{
		if ( 8 != image->myContext.bitsPerPel )
		{
			cpiError( "Image bit depth doesn't match file" );
			return CPI_FALSE;
		}

		if ( CPI_Int32( state->readJpeg->num_components ) !=
			 image->myContext.channels )
		{
			cpiError( "Image number of channels doesn't match file" );
			return CPI_FALSE;
		}

		if ( CPI_Int32( state->readJpeg->image_width ) !=
			 image->myContext.sizeX )
		{
			cpiError( "Image width doesn't match file width" );
			return CPI_FALSE;
		}

		if ( image->myContext.offsetY < state->readJpeg->output_scanline )
		{
			cpiSetPosition( state->myDevice, 0, CPI_SEEK_START );

			jpeg_abort_decompress( state->readJpeg );

			jpegStartRead( state->readJpeg, state->myDevice );

			if ( JPEG_HEADER_OK == jpeg_read_header( state->readJpeg, true ) )
			{
				state->myIsStarted	= CPI_FALSE;
				retval				= CPI_TRUE;
			}
			else
				retval = CPI_FALSE;
		}
		else
			retval = CPI_TRUE;

		if ( retval && ! state->myIsStarted )
		{
			jpeg_start_decompress( state->readJpeg );
			state->myIsStarted = CPI_TRUE;
		}

		// Skip until we get to the line we need...
		y = state->readJpeg->output_scanline;
		for ( ; retval && y < image->myContext.offsetY; ++y )
		{
			linebuf = (JSAMPLE *)cpiGetLine( image, image->myContext.offsetY );
			if ( 1 != jpeg_read_scanlines( state->readJpeg, &linebuf, 1 ) )
			{
				retval = CPI_FALSE;
				break;
			}
		}

		for ( y = 0; retval && y < image->myContext.sizeY; y++ )
		{
			linebuf = (JSAMPLE *)cpiGetLine( image,
											 image->myContext.offsetY + y );

			if ( 1 != jpeg_read_scanlines( state->readJpeg, &linebuf, 1 ) )
			{
				retval = CPI_FALSE;
				break;
			}
		}
	}
	else
	{
		cpiError( "JpegReadImage: destination image is NULL" );
		retval = CPI_FALSE;
	}

	return retval;
}


////////////////////////////////////////


static CPI_PrivateData
JpegOpenForWrite( const CPI_Metadata	meta,
				  const CPI_Metadata  	imageinfo )
{
	CPI_Int32	 ival;
	CPI_Int32  	 width, height, channels, bitdepth;
	imageState	*retval 	= (imageState *)cpiAlloc( sizeof( imageState ) );
	imageState	*cleanptr	= NULL;

	if ( NULL == retval )
		return NULL;
	
	retval->myDevice = cpiOpen( meta );
	if ( NULL == retval->myDevice )
	{
		cpiError( "Unable to open file for write!" );
		cpiFree( retval );
		return NULL;
	}

	// Get the jpeg quality parm
	cpiGetInteger( &ival, "quality", 0 );
	
	// Get the image info....
	cpiGetImageFullSize( imageinfo, &width, &height );
	cpiGetImageChannels( imageinfo, &channels );
	cpiGetImageBitDepth( imageinfo, &bitdepth );

	// Make sure we have a valid # of channels & bitdepth
	if ( bitdepth != 8 )
	{
		cpiError( "JPEG images must be 8 bit!" );
		cpiClose( retval->myDevice );
		cpiFree( retval );
		return NULL;
	}
	else if ( channels != 3 && channels != 1 )
	{
		cpiError( "JPEG images must be 1 or 3 channels!" );
		cpiClose( retval->myDevice );
		cpiFree( retval );
		return NULL;
	}	

	retval->readJpeg	= NULL;
	retval->writeJpeg 	= ( struct jpeg_compress_struct * ) cpiAlloc(
		sizeof( struct jpeg_compress_struct ) ); 	

	if ( NULL == retval->writeJpeg )
	{
		cpiError( "Unable to allocate memory for jpeg write buffer" );
		cpiClose( retval->myDevice );
		cpiFree( retval );
		return NULL;
	}

	retval->writeJpeg->err			= jpeg_std_error( &( retval->jpegErr ) );
	retval->jpegErr.error_exit		= jpeg_error_exit;
	retval->jpegErr.output_message	= jpeg_error_message;
	retval->myIsStarted				= CPI_FALSE;

	jpeg_create_compress( retval->writeJpeg );

	if ( setjmp( jpegJmpBuffer ) )
	{
		cleanptr	= retval;
		retval		= NULL;
	}

	if ( NULL != retval )
	{
		jpegStartWrite( retval->writeJpeg, retval->myDevice, 
						ival, width, height, channels );

		jpeg_start_compress( retval->writeJpeg, TRUE );
	}

	if ( NULL != cleanptr )
	{
		jpegWriteFile	*writer	= NULL;
		
		jpeg_abort_compress( cleanptr->writeJpeg );

		writer = (jpegWriteFile *)cleanptr->writeJpeg->dest;

		if ( NULL != writer->myBuf )
			cpiFree( writer->myBuf );

		if ( NULL != writer )
			cpiFree( writer );
		
		jpeg_destroy_compress( cleanptr->writeJpeg );
		cpiFree( cleanptr->writeJpeg );
		cpiClose( cleanptr->myDevice );
		cpiFree( cleanptr );
	}

	return (CPI_PrivateData)retval;
}


////////////////////////////////////////


static CPI_Bool
JpegWriteImage( CPI_PrivateData	 handle,
				CPI_Image	  	*image )
{
	CPI_Uint32	 y, endY;
	JSAMPLE		*dataPtr;
	CPI_Bool	 retval = CPI_TRUE;
	imageState	*state	= (imageState *)handle;
	
	if ( NULL == image )
	{
		cpiError( "Trying to write from a NULL image!" );
		return CPI_FALSE;
	}

	endY = image->myContext.offsetY + image->myContext.sizeY;

	for ( y = image->myContext.offsetY; retval && y < endY; y++ )
	{
		dataPtr = (JSAMPLE *)cpiGetLine( image, y );

		if ( 1 != jpeg_write_scanlines( state->writeJpeg, &dataPtr, 1 ) )
		{
			retval = CPI_FALSE;
			break;
		}
	}

	return retval;
}


//////////////////////////////////////////////


static CPI_Bool
JpegCloseImage( CPI_PrivateData handle )
{
	imageState	*state	= (imageState *)handle;
	
	if ( NULL != state->readJpeg )
	{
		jpegReadFile	*reader;

		if ( state->myIsStarted &&
			 state->readJpeg->output_scanline >= state->readJpeg->image_height )
			jpeg_finish_decompress( state->readJpeg );
		else
			jpeg_abort_decompress( state->readJpeg );

		reader = (jpegReadFile *)state->readJpeg->src;

		if ( NULL != reader->myBuf )
			cpiFree( reader->myBuf );

		if ( NULL != reader )
			cpiFree( reader );

		jpeg_destroy_decompress( state->readJpeg );
		cpiFree( state->readJpeg );
	}
	else if ( NULL != state->writeJpeg )
	{
		jpegWriteFile	*writer;

		if ( state->writeJpeg->next_scanline < state->writeJpeg->image_height )
			jpeg_abort_compress( state->writeJpeg );
		else
			jpeg_finish_compress( state->writeJpeg );

		jpeg_destroy_compress( state->writeJpeg );

		writer = (jpegWriteFile *)state->writeJpeg->dest;

		if ( NULL != writer->myBuf )
			cpiFree( writer->myBuf );

		if ( NULL != writer )
			cpiFree( writer );

		cpiFree( state->writeJpeg );
	}

	cpiClose( state->myDevice );

	cpiFree( state );

	return CPI_TRUE;
}


//////////////////////////////////////////////

static RPI_ImageInfo dinfo = {
	{
		"jpeg",
		"Jpeg",
		"Silicon Grail",
		"1.0", NULL, NULL
	},

	JpegInit,
	JpegShutDown,
	NULL,
	JpegOpenForRead,
	JpegReadImage,
	JpegReadSubImage,
	JpegOpenForWrite,
    JpegWriteImage,
	JpegWriteImage,
	JpegCloseImage,
	NULL,
	CPI_IMAGE_RGB | CPI_IMAGE_RGBA | CPI_IMAGE_UINT8 | CPI_IMAGE_UINT16 |
	CPI_IMAGE_FLOAT32,
	"jpg;jpeg"
};


/////////////////////////////////////////////


CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	return cpiRegisterImageReader( &dinfo );
}


////////////////////////////////////////////


CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageReader( "jpeg" );
}


////////////////////////////////////////


static void
jpegStartRead( struct jpeg_decompress_struct	*jinfo,
			   CPI_DeviceHandle					 device )
{
	jpegReadFile	*reader = NULL;

	if ( NULL == jinfo )
		return;

	if ( NULL == jinfo->src )
	{
		reader = (jpegReadFile *)cpiAlloc( sizeof( jpegReadFile ) );

		jinfo->src = (struct jpeg_source_mgr *)reader;

		reader->myBuf = (JOCTET *)cpiAlloc(
			INPUT_BUF_SIZE * sizeof( JOCTET ) );
	}

	reader = (jpegReadFile *)jinfo->src;

	reader->myMembers.init_source		= jpegInitRead;
	reader->myMembers.fill_input_buffer	= jpegFillInputBuffer;
	reader->myMembers.skip_input_data	= jpegSkipInputData;
	reader->myMembers.resync_to_restart	= jpeg_resync_to_restart; // Default
	reader->myMembers.term_source		= jpegFinishRead;
	reader->myDevice					= device;
	reader->myMembers.bytes_in_buffer	= 0;
	reader->myMembers.next_input_byte	= NULL;
}


////////////////////////////////////////


static void
jpegInitRead( j_decompress_ptr /*info*/ )
{
	// NOP
}


////////////////////////////////////////


static boolean
jpegFillInputBuffer( j_decompress_ptr info )
{
	jpegReadFile	*reader = (jpegReadFile *)info->src;
	CPI_Uint64		 nbytes;

	nbytes = cpiRead( reader->myDevice, (char *)reader->myBuf, INPUT_BUF_SIZE );

	if ( 0 == nbytes )
	{
		WARNMS( info, JWRN_JPEG_EOF );

		reader->myBuf[0] = (JOCTET) 0xFF;
		reader->myBuf[1] = (JOCTET) JPEG_EOI;
		nbytes = 2;
	}

	reader->myMembers.next_input_byte = reader->myBuf;
	reader->myMembers.bytes_in_buffer = (size_t)( nbytes );

	return TRUE;
}


////////////////////////////////////////


static void
jpegSkipInputData( j_decompress_ptr info, long num_bytes )
{
	jpegReadFile	*reader = (jpegReadFile *)info->src;

	if ( num_bytes > 0 )
	{
		while ( (unsigned long)( num_bytes ) >
				reader->myMembers.bytes_in_buffer )
		{
			num_bytes -= (long)reader->myMembers.bytes_in_buffer;
			jpegFillInputBuffer( info );
		}
		reader->myMembers.next_input_byte += (size_t) num_bytes;
		reader->myMembers.bytes_in_buffer -= (size_t) num_bytes;
	}
}


////////////////////////////////////////


static void
jpegFinishRead( j_decompress_ptr info )
{
	jpegReadFile	*reader = (jpegReadFile *)info->src;

	reader->myDevice = NULL;
}


///////////////////////////////////////


static void
jpegStartWrite( struct jpeg_compress_struct	*jinfo,
				CPI_DeviceHandle			 device,
				CPI_Int32					 quality,
				CPI_Int32					 width,
				CPI_Int32					 height, 
				CPI_Int32					 channels )
{
	jpegWriteFile	*writer = NULL;

	if ( NULL == jinfo )
		return;

	if ( NULL == jinfo->dest )
	{
		writer = (jpegWriteFile *)cpiAlloc( sizeof( jpegWriteFile ) );
		jinfo->dest = (struct jpeg_destination_mgr *)writer;
		writer->myBuf = (JOCTET *)cpiAlloc( OUTPUT_BUF_SIZE * sizeof( JOCTET ) );
	}

	writer->myMembers.init_destination 		= jpegInitWrite;
	writer->myMembers.empty_output_buffer	= jpegEmptyOutputBuffer;
	writer->myMembers.term_destination		= jpegFinishWrite;
	writer->myMembers.next_output_byte		= writer->myBuf;
	writer->myMembers.free_in_buffer		= OUTPUT_BUF_SIZE;
	writer->myDevice						= device;
	
	jinfo->image_width						= width;
	jinfo->image_height						= height;
	jinfo->input_components					= channels;
	jinfo->input_gamma						= 1.0;
#pragma TODO( tjc, "Is this the gamma value which we pass in the context?" )

	if ( channels == 3 )
		jinfo->in_color_space 					= JCS_RGB;
	else if ( channels == 1 )
		jinfo->in_color_space					= JCS_GRAYSCALE;

	jpeg_set_defaults( jinfo );

	jpeg_set_quality( jinfo, quality, TRUE );
}


///////////////////////////////////////


static void
jpegInitWrite( j_compress_ptr /*info*/ )
{
	// NOOP
}


///////////////////////////////////////


static boolean
jpegEmptyOutputBuffer( j_compress_ptr info )
{
	boolean		 	 retval	= TRUE;
	jpegWriteFile	*writer = (jpegWriteFile *)info->dest;
	CPI_Uint64		 nbytes;

	nbytes = cpiWrite( writer->myDevice,
					   (char *)writer->myBuf,
					   OUTPUT_BUF_SIZE );

	if ( nbytes != OUTPUT_BUF_SIZE )
		retval = FALSE;
	
	writer->myMembers.next_output_byte = writer->myBuf;
	writer->myMembers.free_in_buffer = OUTPUT_BUF_SIZE;

	return retval;
}


///////////////////////////////////////
		

static void
jpegFinishWrite( j_compress_ptr info )
{
	jpegWriteFile	*writer = (jpegWriteFile *)info->dest;

	if ( writer->myMembers.free_in_buffer < OUTPUT_BUF_SIZE )
	{
		// We've still got some info left in the buffer, flush it...
		jpegEmptyOutputBuffer( info );
	}

	writer->myDevice = NULL;
}

