/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Tim Connolly
//  Silicon Grail Corporation
//  736 Seward St
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:		Tiff.C
//
// COMMENTS:
//		 The Tiff File Reader
*/

#include <fcntl.h>
#include <stdio.h>
#include <errno.h>

#include "Tiff.h"

//////////////////////////////////////


static const char *compmenu[] =
{
	"Zip Compression",
	"LZW Compression",
	"Jpeg Compression",
	"No Compression",
	NULL
};

static const char *ziplevelmenu[] =
{
	"1 - Fast",
	"2",
	"3",
	"4",
	"5",
	"6 - Default",
	"7",
	"8",
	"9 - Max compression - slow",
	NULL
};

static CPI_Bool
TiffInit( CPI_Bool initforread )
{
	if ( CPI_FALSE == initforread )
	{
		cpiAddMenu( "compression",
					"Compression",
					1,
					compmenu,
					"Set type of compression to use",
					NULL );
		cpiAddMenu( "zip_level",
					"Zip Level",
					5,
					ziplevelmenu,
					"Level of zip compression to use",
					NULL );
		cpiAddInteger( "jpeg_quality",
					   "Jpeg Quality Level",
					   90,
					   1,
					   100,
					   CPI_PARM_DEFAULT,
					   "Adjust the quality of the jpeg image to be written",
					   NULL );
	}

	return CPI_TRUE;
}


//////////////////////////////////////////////


static CPI_Bool
TiffShutDown( void )
{
	return CPI_TRUE;
}


////////////////////////////////////////


static void
TiffParmChanged( const char * /*parmname*/ )
{
	CPI_Bool	 enablezip, enablejpeg;
	CPI_Int32	 ival;

	enablezip	= CPI_FALSE;
	enablejpeg	= CPI_FALSE;

	if ( cpiGetInteger( &ival, "compression", 0 ) )
	{
		switch ( ival )
		{
			case 0:
				enablezip = CPI_TRUE;
				break;

			case 2:
				enablejpeg = CPI_TRUE;
				break;

			default:
				break;
		}
	}

	cpiEnableParameter( "zip_level", enablezip );
	cpiEnableParameter( "jpeg_quality", enablejpeg );
}



////////////////////////////////////////////


static CPI_PrivateData
TiffOpenForRead( const CPI_Metadata	 meta,
				 CPI_Metadata		 imageinfo )
{
	char		 filename[CPI_PATH_LENGTH];
	imageState	*retval = (imageState *)cpiAlloc( sizeof ( imageState ) );

	if ( NULL == retval )
		return NULL;

	cpiGetFilename( meta, filename, CPI_PATH_LENGTH );
	retval->myDevice = cpiOpen( meta );
	if ( NULL == retval->myDevice )
	{
		cpiError( "Unable to open file for read" );
		cpiFree( retval );
		return NULL;
	}

	retval->lastY = 0;

	retval->tif = TIFFClientOpen( filename, "rm", (thandle_t) retval,
								  tiffReadProc, tiffWriteProc, tiffSeekProc,
								  tiffCloseProc, tiffSizeProc, NULL, NULL );

	if ( retval->tif != NULL )
	{
//		TIFFPrintDirectory( retval->tif, stdout, TIFFPRINT_STRIPS |
//							TIFFPRINT_CURVES | TIFFPRINT_CURVES |
//							TIFFPRINT_COLORMAP );

		TIFFGetFieldDefaulted( retval->tif,
							   TIFFTAG_BITSPERSAMPLE,
							   &retval->bitspersample );

		if ( 1 == retval->bitspersample )
		{
			cpiError( "1 Bit Tiff Files Not Supported" );
			TIFFFlushData( retval->tif );
			TIFFClose( retval->tif ); // This will close the file
			cpiFree( retval );
			return NULL;
		}
		TIFFGetFieldDefaulted( retval->tif,
							   TIFFTAG_SAMPLESPERPIXEL,
							   &retval->samplesperpixel );
		TIFFGetFieldDefaulted( retval->tif,
							   TIFFTAG_IMAGEWIDTH,
							   &retval->columns );
		TIFFGetFieldDefaulted( retval->tif,
							   TIFFTAG_IMAGELENGTH,
							   &retval->rows );
		TIFFGetFieldDefaulted( retval->tif,
							   TIFFTAG_COMPRESSION,
							   &retval->compression );
		TIFFGetFieldDefaulted( retval->tif,
							   TIFFTAG_PHOTOMETRIC,
							   &retval->photoMetric );
		TIFFGetFieldDefaulted( retval->tif,
							   TIFFTAG_ORIENTATION,
							   &retval->orientation );

		if ( ORIENTATION_TOPLEFT != retval->orientation &&
			 ORIENTATION_BOTLEFT != retval->orientation )
		{
			cpiError( "Unsupported TIFF Orientation value" );
			TIFFFlushData( retval->tif );
			TIFFClose( retval->tif ); // This will close the file
			cpiFree( retval );
			return NULL;
		}
		else if ( ORIENTATION_BOTLEFT == retval->orientation )
		{
			cpiSetImageSubSupport( imageinfo, true );

			if ( TIFFIsTiled( retval->tif ) )
			{
				cpiError( "The ORIENTATION_BOTLEFT tag is currently not "
						  "supported in conjunction with TILED images" );
				TIFFFlushData( retval->tif );
				TIFFClose( retval->tif ); // This will close the file
				cpiFree( retval );
				return NULL;
			}
		}

		cpiSetImageFullSize( imageinfo, retval->columns, retval->rows );

		// Paletted images are saved as 1 channel but are always 3 channels
		if ( PHOTOMETRIC_PALETTE == retval->photoMetric )
			cpiSetImageChannels( imageinfo, 3 );
		else
			cpiSetImageChannels( imageinfo, retval->samplesperpixel );
		cpiSetImageBitDepth( imageinfo, retval->bitspersample );
	}
	else
	{
		cpiError( "TiffClientOpen failed! Couldn't get TIFF file info!\n" );
		if ( NULL != retval->myDevice )
			cpiClose( retval->myDevice );
		cpiFree( retval );
		retval = NULL;
	}

	return (CPI_PrivateData)retval;
}


//////////////////////////////////////////////////////////


static CPI_Bool
TiffReadImage( CPI_PrivateData handle, CPI_Image *image )
{
	CPI_Bool	 retval	= CPI_FALSE;
	imageState	*state	= (imageState *)handle;

	if ( image != NULL )
	{
		if ( state->bitspersample != image->myContext.bitsPerPel )
		{
			cpiError( "Image bit depth doesn't match file" );
			return CPI_FALSE;
		}
		// We don't care if the # of channels is wrong for a paletted image
		// it will report that it's one channel but it will always map to 3
		if ( state->samplesperpixel != image->myContext.channels &&
			 PHOTOMETRIC_PALETTE != state->photoMetric )
		{
			cpiError( "Image number of channels doesn't match file" );
			return CPI_FALSE;
		}
		if ( state->columns != image->myContext.sizeX )
		{
			cpiError( "Image width doesn't match file width" );
			return CPI_FALSE;
		}
		if ( state->rows != image->myContext.sizeY )
		{
			cpiError( "Image height doens't match file height" );
			return CPI_FALSE;
		}

		if ( TIFFIsTiled( state->tif ) )
		{
			retval = TiffLoadTiles( state->tif, image, state,
									image->myContext.sizeX,
									image->myContext.sizeY, 0, 0,
									image->myContext.bitsPerPel,
									image->myContext.channels );
		}
		else
		{
			retval = TiffLoadLines( state->tif, image, state,
									image->myContext.sizeX,
									image->myContext.sizeY, 0, 0,
									image->myContext.bitsPerPel,
									image->myContext.channels );
		}

		if ( retval )
			retval = doMetrics( state->photoMetric, image, state->tif );

	}
	else
	{
		cpiError( "TiffReadImage: destination image is NULL" );
		retval = CPI_FALSE;
	}

	return retval;
}


////////////////////////////////////////


static CPI_Bool
TiffReadSubImage( CPI_PrivateData	 handle,
				  CPI_Image			*image )
{
	CPI_Bool	 retval	= CPI_FALSE;
	imageState	*state	= (imageState *)handle;

	if ( image != NULL )
	{
		if ( state->bitspersample != image->myContext.bitsPerPel )
		{
			cpiError( "Image bit depth doesn't match file" );
			return CPI_FALSE;
		}
		// We don't care if the # of channels is wrong for a paletted image
		// it will report that it's one channel but it will always map to 3
		if ( state->samplesperpixel != image->myContext.channels &&
			 PHOTOMETRIC_PALETTE != state->photoMetric )
		{
			cpiError( "Image number of channels doesn't match file" );
			return CPI_FALSE;
		}
		if ( TIFFIsTiled( state->tif ) )
		{
			retval = TiffLoadTiles( state->tif, image, state,
									image->myContext.sizeX,
									image->myContext.sizeY,
									image->myContext.offsetX,
									image->myContext.offsetY,
									image->myContext.bitsPerPel,
									image->myContext.channels );
		}
		else
		{
			retval = TiffLoadLines( state->tif, image, state,
									image->myContext.sizeX,
									image->myContext.sizeY,
									image->myContext.offsetX,
									image->myContext.offsetY,
									image->myContext.bitsPerPel,
									image->myContext.channels );
		}

		if ( retval )
			retval = doMetrics( state->photoMetric, image, state->tif );
	}
	else
	{
		cpiError( "TiffReadSubImage: destination image is NULL" );
		retval = CPI_FALSE;
	}

	return retval;
}


////////////////////////////////////////


static CPI_PrivateData
TiffOpenForWrite( const CPI_Metadata	 meta,
				  const CPI_Metadata	 imageinfo )
{
	CPI_Int32	 channels, bitdepth;
	char		 filename[CPI_PATH_LENGTH];
	CPI_Int32	 compParm;
	CPI_Float32	 time;
	imageState	*retval		= (imageState *)cpiAlloc( sizeof ( imageState ) );

	if ( NULL == retval )
		return NULL;

	// Get all of the information about the image we're going to be creating
	// so that we can write the Tiff header...
	cpiGetFilename( meta, filename, CPI_PATH_LENGTH );
	cpiGetImageFullSize( imageinfo, (CPI_Int32 *)&retval->columns,
						 (CPI_Int32 *)&retval->rows );
	cpiGetImageBitDepth( imageinfo, &bitdepth );
	cpiGetImageChannels( imageinfo, &channels );

	retval->bitspersample = (CPI_Uint16) bitdepth;
	retval->samplesperpixel = (CPI_Uint16) channels;
	retval->myDevice = cpiOpen( meta );
	if ( NULL == retval->myDevice )
	{
		cpiError( "Unable to open file for write" );
		cpiFree( retval );
		return NULL;
	}
	retval->tif = TIFFClientOpen( filename, "wm", (thandle_t) retval,
								  tiffReadProc, tiffWriteProc, tiffSeekProc,
								  tiffCloseProc, tiffSizeProc, NULL, NULL );
	if ( NULL == retval->tif )
	{
		cpiError( "TIFFClientOpen failed!  Coudn't open file for write!" );
		cpiFree ( retval );
		retval = NULL;
	}

	// Write the TIFF header information
	TIFFSetField( retval->tif, TIFFTAG_IMAGEWIDTH, retval->columns );
	TIFFSetField( retval->tif, TIFFTAG_IMAGELENGTH, retval->rows );
	TIFFSetField( retval->tif, TIFFTAG_BITSPERSAMPLE, retval->bitspersample );
	if ( 32 == retval->bitspersample )
	{
		TIFFSetField( retval->tif, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_IEEEFP );
	}
	else
	{
		TIFFSetField( retval->tif, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT );
	}
	TIFFSetField( retval->tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT );

	time = 0.0F;

	cpiGetInteger( &compParm, "compression", time );

	switch ( compParm )
	{
        case 0:
			CPI_Int32 zipLevel;

			cpiGetInteger( &zipLevel, "zip_level", time );
			TIFFSetField( retval->tif,
						  TIFFTAG_COMPRESSION,
						  COMPRESSION_DEFLATE );
			TIFFSetField( retval->tif, TIFFTAG_ZIPQUALITY, zipLevel + 1 );
			break;
        case 1:
			TIFFSetField( retval->tif,
						  TIFFTAG_COMPRESSION,
						  COMPRESSION_LZW );
			if ( 32 == retval->bitspersample )
				TIFFSetField( retval->tif, TIFFTAG_PREDICTOR, 1 );
			else
				TIFFSetField( retval->tif, TIFFTAG_PREDICTOR, 2 );
			break;
        case 2:
			CPI_Int32 jpegquality;

			if ( 32 == retval->bitspersample )
			{
				cpiError( "Cannot jpeg compress float images!" );
				TIFFFlushData( retval->tif );
				TIFFClose( retval->tif );
				cpiFree( retval );
				return NULL;
			}

			cpiGetInteger( &jpegquality, "jpeg_quality", time );
			TIFFSetField( retval->tif,
						  TIFFTAG_COMPRESSION,
						  COMPRESSION_JPEG );
			TIFFSetField( retval->tif, TIFFTAG_JPEGQUALITY, jpegquality );
			break;
        default:
			TIFFSetField( retval->tif,
						  TIFFTAG_COMPRESSION,
						  COMPRESSION_NONE );
			break;
	}



	TIFFSetField( retval->tif, TIFFTAG_SAMPLESPERPIXEL,
				  retval->samplesperpixel );
	TIFFSetField( retval->tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB );
	TIFFSetField( retval->tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG );
	TIFFSetField( retval->tif, TIFFTAG_IMAGEDESCRIPTION,
				  "Created by RAYZ" );

	return (CPI_PrivateData) retval;
}


////////////////////////////////////////


static CPI_Bool
TiffWriteImage( CPI_PrivateData	 handle,
				CPI_Image		*image )
{
	CPI_Bool	 retval			= CPI_FALSE;
	imageState	*state			= (imageState *)handle;

	if ( NULL == image )
	{
		cpiError( "Trying to write a NULL image!" );
		cpiFree( state );
		retval = CPI_FALSE;
	}
	else
	{
		int		 y, sy;
		void	*dataPtr		= NULL;

		// Write the TIFF data
		sy = image->myContext.sizeY;

		for ( y = 0; y < sy; ++y )
		{
			dataPtr = cpiGetLine( image, image->myContext.offsetY + y );

			if ( -1 == TIFFWriteScanline( state->tif, dataPtr, y, 0 ) )
			{
				cpiError( "Couldn't write Tiff Scanline" );
				return CPI_FALSE;
			}
		}
		retval = CPI_TRUE;
	}

	return retval;
}


////////////////////////////////////////


static CPI_Bool
TiffWriteSubImage( CPI_PrivateData	 handle,
				   CPI_Image		*image )
{
	CPI_Bool	 retval			= CPI_FALSE;
	imageState	*state			= (imageState *)handle;

	if ( NULL == image )
	{
		cpiError( "Trying to write a NULL image!" );
		cpiFree( state );
		retval = CPI_FALSE;
	}
	else
	{
		int		 y, sy;
		void	*dataPtr		= NULL;
		// Write the TIFF data
		sy = image->myContext.offsetY + image->myContext.sizeY;
		for ( y = image->myContext.offsetY; y < sy; ++y )
		{
			dataPtr = cpiGetLine( image, y );
			TIFFWriteScanline( state->tif, dataPtr, y, 0 );
		}
		retval = CPI_TRUE;
	}

	return retval;
}


//////////////////////////////////////////////


static CPI_Bool
TiffCloseImage( CPI_PrivateData handle )
{
	imageState	*state	= (imageState *)handle;

	TIFFFlushData( state->tif );
	TIFFClose( state->tif );

	cpiFree( state );

	return CPI_TRUE;
}


//////////////////////////////////////////////


static RPI_ImageInfo dinfo = {
	{
		"tiff",
		"Tiff",
		"Silicon Grail",
		"1.0", NULL, NULL
	},

	TiffInit,
	TiffShutDown,
	TiffParmChanged,
	TiffOpenForRead,
	TiffReadImage,
	TiffReadSubImage,
	TiffOpenForWrite,
	TiffWriteImage,
	TiffWriteSubImage,
	TiffCloseImage,
	NULL,
	CPI_IMAGE_RGB | CPI_IMAGE_RGBA | CPI_IMAGE_UINT8 | CPI_IMAGE_UINT16 |
	CPI_IMAGE_FLOAT32,
	"tiff;tif"
};


/////////////////////////////////////////////


CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	return cpiRegisterImageReader( &dinfo );
}


////////////////////////////////////////////


CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageReader( "tiff" );
}


///////////////////////////////////////////////////////////////////////////////
// RAYZ libtiff routines
//////////////////////////////////////////////////////////////////////////////

static tsize_t
tiffReadProc( thandle_t fd, tdata_t buf, tsize_t size )
{
	CPI_Int32  retval	= -1;
	imageState *is		= (imageState *)fd;

	retval = cpiRead( is->myDevice, (char *)buf, size );

	return (tsize_t)retval;
}


///////////////////////////////////////////////


static tsize_t
tiffWriteProc( thandle_t fd, tdata_t buf, tsize_t size )
{
	CPI_Int32	 retval	= -1;
	imageState	*is		= (imageState *)fd;

	if ( cpiWrite( is->myDevice, (char *)buf, size ) )
		retval = size;

	return (tsize_t)retval;
}


//////////////////////////////////////////////


static toff_t
tiffSeekProc( thandle_t fd, toff_t off, int whence )
{
	imageState	*is		= (imageState *)fd;
	CPI_Uint64	 retval	= CPI_Uint64(-1);

	if ( SEEK_SET == whence )
		retval = cpiSetPosition( is->myDevice, off, CPI_SEEK_START );
	else if ( SEEK_CUR == whence )
		retval = cpiSetPosition( is->myDevice, off, CPI_SEEK_CUR );
	else if ( SEEK_END == whence )
		retval = cpiSetPosition( is->myDevice, off, CPI_SEEK_END );

	return (toff_t)( retval );
}


/////////////////////////////////////////////////


static int
tiffCloseProc( thandle_t fd )
{
	imageState *is	= (imageState *)fd;

	cpiClose( is->myDevice );
	is->myDevice = NULL;

	return 1;
}


/////////////////////////////////////////////////


static void
tiffError( const char *module, const char *fmt, va_list ap )
{
	char myError[1024];

	if ( module != NULL )
		sprintf( myError, "%s: ", module );

	vsprintf( myError, fmt, ap );
	cpiError( myError );
}

TIFFErrorHandler _TIFFerrorHandler = tiffError;


/////////////////////////////////////////////////


static void
tiffWarning( const char *module, const char *fmt, va_list ap )
{
	char myWarning[1024];

	if ( module != NULL )
		sprintf( myWarning, "%s: ", module );
	vsprintf( myWarning, fmt, ap );
	cpiWarning( myWarning );
}

TIFFErrorHandler _TIFFwarningHandler = tiffWarning;


//////////////////////////////////////////////////


static toff_t
tiffSizeProc( thandle_t fd )
{
	imageState			*myHandle	= (imageState *)fd;
	CPI_DeviceHandle	 myDevice	= (CPI_DeviceHandle)myHandle->myDevice;
	CPI_Uint64			 retval	= 0;

	if ( ! cpiFileLength( myDevice, &retval ) )
		cpiWarning( "Tiff Warning: Couldn't get the size of the file!" );

	return (toff_t)retval;
}


/////////////////////////////////////////////////


void *
_TIFFmalloc( tsize_t s )
{
	return cpiAlloc( s );
}


//////////////////////////////////////////////////


void
_TIFFfree( tdata_t p )
{
	cpiFree( p );
}


//////////////////////////////////////////////////


void *
_TIFFrealloc( tdata_t p, tsize_t s )
{
	return( cpiReAlloc( p, s ) );
}


//////////////////////////////////////////////////


void
_TIFFmemset( tdata_t p, int v, tsize_t c )
{
	memset( p, v, (size_t)c );
}


//////////////////////////////////////////////////


void
_TIFFmemcpy( tdata_t d, const tdata_t s, tsize_t c )
{
	if ( c > 0 )
		memcpy( d, s, (size_t)c );
}


//////////////////////////////////////////////////


int
_TIFFmemcmp( const tdata_t p1, const tdata_t p2, tsize_t c )
{
	return( memcmp( p1, p2, (size_t)c ) );
}


//////////////////////////////////////////////////


static CPI_Bool
TiffLoadTiles( TIFF			*tif,
			   CPI_Image	*image,
			   imageState	*state,
			   CPI_Uint32	 sizeX,
			   CPI_Uint32	 sizeY,
			   CPI_Uint32	 offX,
			   CPI_Uint32	 offY,
			   CPI_Uint16	 bitdepth,
			   CPI_Uint16	 channels )
{
	CPI_Uint32	 tileWidth, tileHeight;
	char		*tilePtr;
	char		*tmpPtr;
	CPI_Uint32	 thewidth;
	CPI_Uint32	 theheight;
	char		*dataPtr;
	CPI_Uint16	 bpc			= bitdepth / 8;
	CPI_Uint16	 bpp			= bpc * channels;
	CPI_Uint16	 planar			= PLANARCONFIG_CONTIG;
	CPI_Uint32	 endX			= sizeX + offX;
	CPI_Uint32	 endY			= sizeY + offY;

	CPI_Bool retval = CPI_FALSE;

	if ( tif != NULL )
	{
		TIFFGetField( tif, TIFFTAG_PLANARCONFIG, &planar );
		TIFFGetField( tif, TIFFTAG_TILEWIDTH, &tileWidth );
		TIFFGetField( tif, TIFFTAG_TILELENGTH, &tileHeight );

		tilePtr = (char *)cpiAlloc( TIFFTileSize( tif ) );

		if ( PLANARCONFIG_CONTIG == planar )
		{
			for ( CPI_Uint32 y = offY + 1; y < endY + 1; y += tileHeight )
			{
				theheight = endY - y + 1 > tileHeight ? tileHeight
					: endY - y + 1;

				for ( CPI_Uint32 x = offX; x < endX; x += tileWidth )
				{
					thewidth = sizeX - x > tileWidth ? tileWidth : sizeX - x;

					TIFFReadTile( tif, tilePtr, x, y - 1, 0, 0 );
					tmpPtr = tilePtr;
					// If we're not a tile boundary, we need to move up to
					// where we want to be...
					CPI_Uint32 test = ( y - 1 ) % tileHeight;
					if ( test != 0 )
						tmpPtr += test * thewidth * bpp;

					for ( CPI_Uint32 z = 0; z < theheight; z++ )
					{
						dataPtr = ( (char *)cpiGetLine( image, y + z - 1 ) +
									( x * bpp ) );
						memcpy( dataPtr, tmpPtr, thewidth * bpp );
						tmpPtr += bpp * thewidth;
					}
				}
			}
		}
		else
		{
			for ( CPI_Uint32 y = offY + 1; y < endY + 1; y += tileHeight )
			{

				theheight = endY - y + 1 > tileHeight ? tileHeight
					: endY - y + 1;

				for ( CPI_Uint32 x = offX; x < endX; x += tileWidth )
				{
					thewidth = endX - x > tileWidth ? tileWidth : endX - x;

					for ( CPI_Uint32 c = 0; c < channels; c++ )
					{
						TIFFReadTile( tif, tilePtr, x, y - 1, 0, c );
						tmpPtr	= tilePtr;
						// If we're not a tile boundary, we need to move up to
						// where we want to be...
						CPI_Uint32 test = ( y - 1 ) % tileHeight;
						if ( test != 0 )
							tmpPtr += test * thewidth * bpc;

						for ( CPI_Uint32 z = 0; z < theheight; z++ )
						{
							dataPtr = ( (char *)cpiGetLine( image, y + z - 1 )
										+ x * bpp + ( c * bpc ) );
							for ( CPI_Uint32 a = 0;
								  a < thewidth * bpc; a++ )
							{
								dataPtr[ a * channels ] = tmpPtr[ a ];
							}
							tmpPtr += thewidth * bpc;
						}
					}
				}
			}
		}

		cpiFree ( tilePtr );
		retval = CPI_TRUE;
	}
	else
		retval = CPI_FALSE;

	return retval;
}


////////////////////////////////////////////


static CPI_Bool
TiffLoadLines( TIFF			*tif,
			   CPI_Image	*image,
			   imageState	*state,
			   CPI_Uint32	 sizeX,
			   CPI_Uint32	 sizeY,
			   CPI_Uint32	 offX,
			   CPI_Uint32	 offY,
			   CPI_Uint16	 bitdepth,
			   CPI_Uint16	 channels )
{
	CPI_Uint32	 y;
	CPI_Uint8	*dataPtr	= NULL;
	CPI_Uint16	 planar		= PLANARCONFIG_CONTIG;

	TIFFGetField( tif, TIFFTAG_PLANARCONFIG, &planar );

	if ( PLANARCONFIG_CONTIG == planar )
	{
		// We're reading the full image, use ReadScanline directly
		// into our image pointer
		if ( 0 == offX && 0 == offY )
		{
			for ( y = 0; y < sizeY; y++ )
			{
				if ( ORIENTATION_BOTLEFT == state->orientation )
					dataPtr = (CPI_Uint8 *)cpiGetLine( image, sizeY - y - 1 );
				else
					dataPtr = (CPI_Uint8 *)cpiGetLine( image, y );
				TIFFReadScanline( tif, dataPtr, y, 0 );
				state->lastY = y;
			}
		}
		else
		{
			char		*linePtr;
			CPI_Uint32	 startY	= offY;

			linePtr = (char *)cpiAlloc( TIFFScanlineSize( tif ) );

			if ( COMPRESSION_NONE == state->compression )
				startY = offY;
			else if ( state->lastY <= offY )
				startY = state->lastY;
			else
				startY = 0;

			for ( y = startY; y < offY; y++ )
			{
				TIFFReadScanline( tif, linePtr, y, 0 );
				state->lastY = y;
			}

			for ( y = offY; y < ( offY + sizeY ); y++ )
			{
				dataPtr = (CPI_Uint8 *)cpiGetLine( image, y );

				TIFFReadScanline( tif, dataPtr, y, 0 );
				state->lastY = y;
			}
			state->lastY++;
			cpiFree( linePtr );
		}
	}
	else
	{
		char		*linePtr	= (char *)cpiAlloc( TIFFScanlineSize( tif ) );
		CPI_Uint32	 startY	= offY;

		if ( state->lastY <= offY &&
			 state->compression != COMPRESSION_NONE )
			startY = state->lastY;
		else
			startY = 0;

		for ( y = startY; y < offY; y++ )
		{
			TIFFReadScanline( tif, linePtr, y, 0 );
		}

		for ( int c = 0; c < channels; c++ )
		{
			for ( y = offY; y < ( offY + sizeY ); y++ )
			{
				dataPtr = (CPI_Uint8 *)cpiGetLine( image, y );
				TIFFReadScanline( tif, linePtr, y, c );

				for ( int z = 0; z < TIFFScanlineSize( tif ); z++ )
					dataPtr[ z * channels ] = linePtr[z];
			}
		}
		cpiFree( linePtr );
	}
	return CPI_TRUE;
}


////////////////////////////////////////////////


static CPI_Bool
doMetrics( CPI_Uint16 photoMetric, CPI_Image *image, TIFF *tif )
{
	CPI_Bool	retval = CPI_TRUE;

	// Deal with all of the different photometrics
	if ( PHOTOMETRIC_MINISWHITE == photoMetric )
	{
		if ( 8 == image->myContext.bitsPerPel )
		{
			CPI_Uint8 myLut[256];

			for ( int i = 0; i < 256; i++ )
				myLut[i] = 255 - i;

			cpiApplyLut8( image, myLut, -1 );
		}
		else if ( 16 == image->myContext.bitsPerPel )
		{
			CPI_Uint16 myLut[65536];
			for ( int i = 0; i < 65536; i++ )
				myLut[i] = 65535 - i;

			cpiApplyLut16( image, myLut, -1 );
		}
		else if ( 32 == image->myContext.bitsPerPel )
		{
			char *dataPtr = (char *)image->data;

			for ( int x = 0; x < image->myContext.sizeX; x++ )
				for ( int y = 0; y < image->myContext.sizeY; y++ )
				{
					*dataPtr = 1 - *dataPtr;
					dataPtr++;
				}
		}
		else
		{
			cpiError( "Invalid Bit Depth!\n" );
			retval = CPI_FALSE;
		}
	}
	else if ( PHOTOMETRIC_SEPARATED == photoMetric )
	{
		int		 tmpval, black;
		int		 x, y;

		switch ( image->myContext.bitsPerPel )
		{
			case 8:
			{
				CPI_Uint8	*tmpPtr		= (CPI_Uint8 *)image->data;

				for ( y = 0; y < image->myContext.sizeY; ++y )
				{
					for ( x = 0; x < image->myContext.sizeX; ++x )
					{
						black = 255 - int( tmpPtr[3] );

						tmpval = ( int( tmpPtr[0] ) * black + 0x80 ) >> 8;
						tmpval += int( tmpPtr[3] );
						tmpPtr[0] = 255 - ( ( tmpval > 255 ) ? 255 : tmpval );

						tmpval = ( int( tmpPtr[1] ) * black + 0x80 ) >> 8;
						tmpval += int( tmpPtr[3] );
						tmpPtr[1] = 255 - ( ( tmpval > 255 ) ? 255 : tmpval );

						tmpval = ( int( tmpPtr[2] ) * black + 0x80 ) >> 8;
						tmpval += int( tmpPtr[3] );
						tmpPtr[2] = 255 - ( ( tmpval > 255 ) ? 255 : tmpval );

						tmpPtr += 4;
					}
				}
				break;
			}

			case 16:
			{
				CPI_Uint16	*tmpPtr		= (CPI_Uint16 *)image->data;

				for ( y = 0; y < image->myContext.sizeY; ++y )
				{
					for ( x = 0; x < image->myContext.sizeX; ++x )
					{
						black = 65535 - int( tmpPtr[3] );

						tmpval = ( int( tmpPtr[0] ) * black + 0x800 ) >> 16;
						tmpval += int( tmpPtr[3] );
						tmpPtr[0] = 65535 - ( ( tmpval > 65535 ) ? 65535 :
											  tmpval );

						tmpval = ( int( tmpPtr[1] ) * black + 0x800 ) >> 16;
						tmpval += int( tmpPtr[3] );
						tmpPtr[1] = 65535 - ( ( tmpval > 65535 ) ? 65535 :
											  tmpval );

						tmpval = ( int( tmpPtr[2] ) * black + 0x800 ) >> 16;
						tmpval += int( tmpPtr[3] );
						tmpPtr[2] = 65535 - ( ( tmpval > 65535 ) ? 65535 : tmpval );

						tmpPtr += 4;
					}
				}
				break;
			}
			case 32:
			{
				CPI_Float32	*tmpPtr		= (CPI_Float32 *)image->data;
				CPI_Float32	 tmpf, fblack;

				for ( y = 0; y < image->myContext.sizeY; ++y )
				{
					for ( x = 0; x < image->myContext.sizeX; ++x )
					{
						fblack = 1.0F - tmpPtr[3];

						tmpf = tmpPtr[0] * fblack + tmpPtr[3];
						tmpPtr[0] = 1.0F - ( ( tmpf > 1.0F ) ? 1.0F : tmpf );

						tmpf = tmpPtr[1] * fblack + tmpPtr[3];
						tmpPtr[1] = 1.0F - ( ( tmpf > 1.0F ) ? 1.0F : tmpf );

						tmpf = tmpPtr[2] * fblack + tmpPtr[3];
						tmpPtr[2] = 1.0F - ( ( tmpf > 1.0F ) ? 1.0F : tmpf );

						tmpPtr += 4;
					}
				}
				break;
			}

			default:
				break;
		}
	}
	else if ( PHOTOMETRIC_PALETTE == photoMetric )
	{
		CPI_Uint16	*redmap;
		CPI_Uint16	*greenmap;
		CPI_Uint16	*bluemap;
		CPI_Uint8	*dataPtr	= (CPI_Uint8 *)image->data;
		CPI_Uint8	*theimg		= NULL; 
		CPI_Uint8	*mapPtr		= NULL;

		theimg	= (CPI_Uint8 *)cpiAlloc( image->myContext.sizeX *
										 image->myContext.sizeY * 3 );
		mapPtr	= theimg;
		
		if ( !TIFFGetField ( tif, TIFFTAG_COLORMAP,
							 &redmap, &greenmap, &bluemap) )
		{
			cpiError ( "Couldn't get TIFF colormaps!" );
			retval = CPI_FALSE;
		}
		else
		{
			memcpy( theimg, dataPtr,
					image->myContext.sizeX * image->myContext.sizeY * 3 );

			for ( int y = 0; y < image->myContext.sizeY; y++ )
			{
				for ( int x = 0; x < image->myContext.sizeX; x++ )
				{
					dataPtr[0] = redmap[ mapPtr[0] ] >> 8;
					dataPtr[1] = greenmap[ mapPtr[0] ] >> 8;
					dataPtr[2] = bluemap[ mapPtr[0] ] >> 8;
					dataPtr += 3;
					mapPtr++;
				}

				mapPtr += image->myContext.sizeX * 2;
			}
		}
		cpiFree( theimg );
	}

	return retval;
}
