/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  710 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	UserLUT.C
//
// COMMENTS: This demonstrates creating a simple user LUT and
//			 getting it into the 'Display Conversion' menu
//
*/

#include <stdio.h>
#include <CPI.h>

///////////////////////////////////////
// This defines the menu strings which will appear to the user


static const char *gammamenu[] =
{
	"linear",
	"log",
	NULL
};


///////////////////////////////////////
// This inits the plugin and defines the user interface elements,
// if any. These elements are defined in exactly the same way as
// for a node.


static CPI_Bool
DoublerInit( void )
{
	CPI_Bool	retval = CPI_FALSE;

	cpiAddMenu( "images_are",
				"Images Are", 
				0, 
				gammamenu, 
				NULL, 
				NULL );

	cpiAddFloat( "bright",
				 "Brightness", 
				 1.0, 		// default
				 0.0, 		// min value
				 3.0,		// max value
				 CPI_PARM_CLAMPMIN | CPI_PARM_CLAMPMAX,
				 NULL,
				 NULL );

	retval = CPI_TRUE;

	return retval;
}


///////////////////////////////////////

static CPI_Bool
DoublerShutDown( void )
{
	return CPI_TRUE;
}


///////////////////////////////////////
// This is the way you create private data to be passed between
// routines.

#define LOGFORMAT	1

typedef struct _mydata
{
	CPI_Float32	brightness;
	CPI_Int32	isLog;
} myData;

CPI_PrivateData
DoublerSetup( CPI_ImageContext context )
{
	myData	*retval;

	retval = (myData *)cpiAlloc( sizeof( myData ) );
	
	cpiGetInteger( &(retval->isLog), "images_are", 1.0F );
	cpiGetFloat( &(retval->brightness), "bright", 1.0F );

	return retval;
}


///////////////////////////////////////
// This is the counterpart to Setup - here we give back
// memory allocated by Setup

void
DoublerCleanup( CPI_PrivateData    handle )
{
	cpiFree( (myData *)handle );
}


///////////////////////////////////////
// This defines the confidence value of this display conversion. It
// returns a value from 0 to 100 which represents your confidence in
// correctly displaying the current image. You could decide, for 
// example, that your plugin was only interested in displaying log
// images; so you could check the image context and return 100 for
// log images and 0 for linear.

static CPI_Uint8
DoublerConfidence( CPI_ImageContext context )
{
	return 1;
}


///////////////////////////////////////
// this is called for each image viewer update.
// values is the number of color values to convert. 
// For 8bit and 16bit images, you are modifying a collection of
// LUT values. For float, you are modifying the data directly - in 
// other words, this subroutine becomes a function.
// The incoming LUT or pixel values, *src, is modified here and
// returned in *dst. This is because you want your LUT to be
// be part of any upstream color changes, not to replace them.

static void
DoublerCreateLUT( CPI_PrivateData		 myHandle,
				  CPI_Float32			*dst,
				  const CPI_Float32		*src,
				  CPI_Int32			 	 values,
				  CPI_Int32			 	 channels )
{	
	// these next few lines show setting up and unloading data
	// from the PrivateData structure which is passed around
	CPI_Float32		factor;
	myData		   *data = (myData *)myHandle;

	factor = data->brightness;
	values *= channels;

	// now iterate over the number of values, and apply the 
	// LUT or image transform
	if ( data->isLog )
	{
		for ( CPI_Int32 i = 0; i < values; ++i )
			*dst++ = *src++ + factor;
	}
	else
	{
		for ( CPI_Int32 i = 0; i < values; ++i )
			*dst++ = *src++ * factor;
	}
}

////////////////////////////////////
// This is called before each execution, to allow your plugin to
// set up default values based on the context of the incoming image.
// Here, we look at the incoming image and set the user menu to
// 'linear' or 'log' depending on the type of image
//
//	if ( context.dataType == LOGFORMAT )
//		retval->isLog = CPI_TRUE;
//	else
//		retval->isLog = CPI_FALSE;

////////////////////////////////////
// As with nodes, LUTs are described by a structure like
// this one. This must be present.


static RPI_LUTInfo lutInfo = {
	{
		"brighter",
		"Brighter",
		"Silicon Grail",
		"v1.0",
		"",
		""
	},
	DoublerInit,
	DoublerShutDown,
	DoublerSetup,
	DoublerCleanup,
	DoublerConfidence,
	NULL,
	DoublerCreateLUT,
	NULL,
};


//////////////////////////////////
// Register the plugin. You should use this declaration
// to ensure compatibility across NT, linux and IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	return cpiRegisterLUT( &lutInfo );
}


//////////////////////////////////
// Unregister the plugin - the name given must be the same as
// used in the LUTInfo structure

CPIDSOEXPORT void
rpiPluginCleanup( void )
{
	cpiUnregisterLUT( "brighter" );
}
