/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	BlurXY
//
// COMMENTS: Use a Rayz IM to blur an image separately in X & Y
//
*/

#include <stdio.h>
#include <CPI.h>

// menu of actions for optional second input
static const char *actionmenu[] =
{
	"Mix",
	"On/Off",
	NULL
};

// menu of channels to use from optional second input
static const char *chanmenu[] =
{
	"Red channel",
	"Green channel",
	"Blue channel",
	"Alpha channel",
	NULL
};


////////////////////////////////////////////
// The init routine is called once, and typically is used to
// define the user input controls

static CPI_Bool
BlurXYInit( void )
{
	cpiAddFloatPair( "blur_size", "Blur Size",
						"x", "X", 10.0,
						"y", "Y", 10.0,
						CPI_PARM_DEFAULT,
						NULL, NULL );
	cpiAddChannelSelect( NULL, NULL );
	cpiAddMenu( "mask_action", "Mask Action", 0, actionmenu, NULL, NULL );
	cpiAddMenu( "mask_channel", "Mask Channel", 0, chanmenu, NULL, NULL );

	return CPI_TRUE;
}


/////////////////////////////////////////////
// shutdown does nothing in this case

static CPI_Bool
BlurXYShutDown( void )
{
	return CPI_TRUE;
}


//////////////////////////////////////////
// This routine gets called whenever the input changes. When
// that happens, we have to see if there is (still) a 2nd input
// and then enable/disable accordingly.

static void
BlurXYInputChanged( CPI_Uint32 /*inputnum*/ )
{
	CPI_Uint32			numInputs	= 0;
	CPI_Bool			enable		= CPI_FALSE;

	cpiGetNumInputs( &numInputs );

	// enable/disable the mask parameters depending on the presence
	// if the mask input
	if ( numInputs == 2 )
		enable = CPI_TRUE;

	cpiEnableParameter( "mask_action", enable );
	cpiEnableParameter( "mask_channel", enable );
}


//////////////////////////////////////////
// This optional routine defines help labels for the inputs.
// Also, the first letter of each label is used to label the
// input connectors.

static const char *
BlurXYInputLabel( CPI_Uint32    input )
{
	if ( input == 0 )
		return( "Input image" );
	else
		return( "Mask image (optional)" );
}


//////////////////////////////////////////
// This is called when an image is required
// The basic structure of these routines is:
// 1. get the inputs
// 2. get the parameters from the user interface
// 3. allocate metadata and copy the parameters to it
// 4. add the image op(s) to the list of operations

static CPI_ImageOp
BlurXYExec( CPI_Float32 time,
			CPI_Uint8	quality,
			CPI_Uint32	/*output*/,
			CPI_Bool	/*viewerOutput*/,
			CPI_Float32	scaleX,
			CPI_Float32	scaleY )
{
	CPI_ImageOp			retval = NULL;
	CPI_ImageOp			myInputs[2];
	CPI_ImageContext	myContext;
	CPI_Float32			ksizex, ksizey;
	CPI_Uint32			numInputs;
	CPI_Int32			chanSel, action, channel;

	// find out how many inputs we have connected
	cpiGetNumInputs( &numInputs );

	// get the first one, and then if there is a second,
	// get that too
	myInputs[0] = cpiGetInputOp( 0, time, quality, scaleX, scaleY );
	myInputs[1] = NULL;
	if ( numInputs > 1 )
		myInputs[1] = cpiGetInputOp( 1, time, quality, scaleX, scaleY );

	// unload params from user interface
	if ( cpiGetFloat( &ksizex, "blur_size.x", time ) &&
		 cpiGetFloat( &ksizey, "blur_size.y", time ) &&
		 cpiGetChannelSelect( &chanSel ) &&
		 cpiGetInteger( &action, "mask_action", time ) &&
		 cpiGetInteger( &channel, "mask_channel", time ) )
	{
		// allocate metadata and copy values to that
		CPI_Metadata theParms = cpiCreateMetadata();

		cpiSetMetaFloat32( theParms, "kernelxscale", ksizex );
		cpiSetMetaFloat32( theParms, "kernelyscale", ksizey );

		cpiSetMetaInt32( theParms, "channelselect", chanSel );
		cpiSetMetaInt32( theParms, "maskchannel", channel );

		// for internal ops that have filter kernels, such as blur,
		// convolve, and edge, we have to set up the active area
		// as metadata
		if ( cpiGetInputContext( &myContext, 0, time ) )
		{
			cpiSetMetaInt32( theParms, "sizex", myContext.sizeX );
			cpiSetMetaInt32( theParms, "sizey", myContext.sizeY );
			cpiSetMetaInt32( theParms, "offx", myContext.offsetX );
			cpiSetMetaInt32( theParms, "offy", myContext.offsetY );
		}

		if ( numInputs > 1 )
		{
			if ( 0 == action )
				cpiSetMetaString( theParms, "masktype", "MIX" );
			else 
				cpiSetMetaString( theParms, "masktype", "ONOFF" );
		}
		else
			cpiSetMetaString( theParms, "masktype", "NONE" );

		// add the 'blur' op to the list
		// in this case, we are using the Rayz internal 'blur' op
		// so this operation is not defined in this file.
		// The documentation gives a list of Rayz ops and their
		// metadata parameters which you may use.
		if ( NULL != myInputs[0] )
			retval = cpiAddImageOp( "blur", theParms, myInputs, 2 );
		else
			cpiError( "Input A is not connected" );
	
		cpiDeleteMetadata( theParms );
	}
	else
		cpiError( "Couldn't get BlurXY parameter(s)" );
		
	return retval;
}


//////////////////////////////////////////
// This structure, which is required, defines to Rayz
// the various routines which are part of this plugin.
// Routines which are not required are set to NULL

static RPI_NodeInfo ninfo = {
	{
		"blurxy",		// node name
		"BlurXY",		// menu name
		"Silicon Grail",
		"1.0", 
		NULL, 
		NULL
	},

	BlurXYInit,		// NodeInit
	BlurXYShutDown,	// NodeShutDown
	NULL,			// ParmChanged
	BlurXYInputChanged,
	NULL,			// GetRegion
	NULL,			// GetFullSize
	NULL,			// GetRange
	NULL,			// GetErrors
	BlurXYExec,		// Node Execute
	BlurXYInputLabel,
	NULL,			//OutputLabel
	NULL,			// ViewerLabel
	NULL,			// ViewerOutputs
	1,				// min inputs
	2,				// max inputs
	1,				// min outputs
	1,				// max outputs
	CPI_FALSE		// is the node growable? (like Mcomp)
};


/////////////////////////////////////////////////
// register the node
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	return cpiRegisterNode( &ninfo );
}


//////////////////////////////////////////////
// unregister the node

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterNode( "blurxy" );
}
