/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	Checker
//
// COMMENTS: This illustrates a source node (one with no inputs)
//
*/

#include <stdio.h>
#include <CPI.h>

//////////////////////////////////////
// menu of which channels to create

static const char *channelmenu[]= 
{
	"RGB Alpha",
	"RGB",
	"Alpha only",
	NULL
};


///////////////////////////////////////
// menu of which bit depth to create

static const char *bitdepthmenu[] = 
{
	"8 bit",
	"16 bit",
	"Float",
	NULL
};


///////////////////////////////////////
// default color arrays

static CPI_Float32 fgsqs[] = { 1.0F, 1.0F, 1.0F, 1.0F };
static CPI_Float32 bgsqs[] = { 0.0F, 0.0F, 0.0F, 0.0F };


/////////////////////////////////////////////////////
// Node Functions ///////////////////////////////////
/////////////////////////////////////////////////////

// The init routine is called once, and typically is used to
// define the user input controls

#define XFULLSIZE_PATH	"Project/Settings/Default Full Size X"
#define YFULLSIZE_PATH	"Project/Settings/Default Full Size Y"

static CPI_Bool
CheckerInit( void )
{
	// use preferences to get default size to use
	CPI_Int32	xsize, ysize;

	if ( !cpiGetPreferenceInt( &xsize, XFULLSIZE_PATH ) )
	{
		cpiError( "Couldn't get X size" );
		return CPI_FALSE;
	}

	if ( !cpiGetPreferenceInt( &ysize, YFULLSIZE_PATH ) )
	{
		cpiError( "Couldn't get Y size" );
		return CPI_FALSE;
	}

	// now set up the parameters
	cpiAddColorGroup( "sq", "Square Col", fgsqs, 4, CPI_PARM_DEFAULT, 
						"Color of squares", NULL );
	cpiAddColorGroup( "bg", "Background", bgsqs, 4, CPI_PARM_DEFAULT, 
						"Color between squares", NULL );
	cpiAddIntPair( "size", "Size",
					"w", "Width", xsize,
					"h", "Height", ysize,
					 CPI_PARM_DEFAULT,
					"Size of output image", NULL );
	cpiAddIntPair( "sqsize", "Square Size",
					"x", "X", 64,
					"y", "Y", 64,
					 CPI_PARM_DEFAULT,
					"Size of squares", NULL );

	cpiAddMenu( "bits", "Bit Depth", 0, bitdepthmenu, NULL, NULL );
	cpiAddMenu( "chans", "Channels", 0, channelmenu, NULL, NULL );
	cpiAddSequenceRange( 1, 100,
						 CPI_PARM_DEFAULT,
						"Range of frames to output", NULL );

	return CPI_TRUE;
}

///////////////////////////////////////////
// shutdown does nothing in this case

static CPI_Bool
CheckerShutDown( void )
{
	return CPI_TRUE;
}


///////////////////////////////////////////
// GetRegion needs to be defined for nodes that generate their
// own images (as opposed to being given images to modify)
// This defines for the rest of Rayz what the output 'context'
// is, ie, size, bit depth, etc.

static CPI_Bool
CheckerGetRegion( CPI_ImageContext	*output,
				  CPI_Float32		 myTime,
				  CPI_Uint32		 /*nodeOutput*/,
				  CPI_Bool		 	 /*viewerOutput*/ )
{
	CPI_Int32	 startFrame, endFrame;
	CPI_Bool	 retval = CPI_TRUE;

	if ( cpiGetSequenceParmRange( &startFrame, &endFrame ) &&
		 myTime >= startFrame &&
		 myTime <= endFrame )
	{
		CPI_Int32	 bitsPerPel, channels;

		// get size from GUI and store in 'output'
		retval = ( cpiGetInteger( &output->sizeX, "size.w", myTime ) &&
				   cpiGetInteger( &output->sizeY, "size.h", myTime ) &&
				   cpiGetInteger( &bitsPerPel, "bits", myTime ) &&
				   cpiGetInteger( &channels, "chans", myTime ) );

		if ( retval )
		{
			output->offsetX = 0;
			output->offsetY = 0;

			switch ( bitsPerPel )
			{
				case 0:
					output->bitsPerPel = 8;
					break;
				case 1:
					output->bitsPerPel = 16;
					break;
				case 2:
					output->bitsPerPel = 32;	// ie, float
					break;
				default:
					cpiError( "Invalid bitsPerPel in getRegion!\n" );
					retval = CPI_FALSE;
					break;
			}

			// get user menu choice, and set value accordingly
			switch ( channels )
			{
				case 0:
					output->channels = 4;
					break;
				case 1:
					output->channels = 3;
					break;
				case 2:
					output->channels = 1;
					break;
				default:
					cpiError( "Invalid channels in getRegion!\n" );
					retval = CPI_FALSE;
					break;
			}
		}
		else
		{
			cpiError( "Couldn't get parameter!" );
			retval = CPI_FALSE;
		}
	}
	else
	{
		cpiError( "Frame Out of Range!" );
		retval = CPI_FALSE;
	}
	
	return retval;
}



////////////////////////////////////////


static void
CheckerParmChanged( const char *parmname )
{
	// If the parm name is NULL, that means the a new node has just been added
	// we want to set the seuqence parm to match the global frame range.
	if ( parmname == NULL )
	{
		CPI_Int32	 startFrame, endFrame;

		if ( cpiGetGlobalFrameRange( &startFrame, &endFrame ) )
			cpiSetSequenceParmRange( startFrame, endFrame );
	}
}


///////////////////////////////////////////
// Rayz may need to ask the plugin what size
// image it intends to create - define that here
// Nodes that create their own output need to
// define this routine

static CPI_Bool
CheckerGetFullSize( CPI_ImageContext *fullsize,
					CPI_Float32		  myTime,
					CPI_Uint32		  /*nodeOutput*/,
					CPI_Bool       	  /*viewerOutput*/ )
{
	CPI_Int32	 startFrame, endFrame;
	CPI_Bool 	 retval = CPI_TRUE;

	if ( cpiGetSequenceParmRange( &startFrame, &endFrame ) &&
		 myTime >= startFrame &&
		 myTime <= endFrame )
	{
		if ( !cpiGetInteger( &fullsize->sizeX, "size.w", myTime ) ||
			 !cpiGetInteger( &fullsize->sizeY, "size.h", myTime ) )
		{
			cpiError( "Couldn't get size parameter" );
			retval = CPI_FALSE;
		}			
	}
	else
	{
		cpiError( "Frame Out Of Range" );
		retval = CPI_FALSE;
	}		
		
	return retval;
}


//////////////////////////////////////////
// This is called when an image is required
// The basic structure of these routines is:
// 1. get the parameters from the user interface
// 2. allocate metadata and copy the parameters to it
// 3. add the image op(s) to the list of operations

static CPI_ImageOp
CheckerExec( CPI_Float32 	 myTime,
			 CPI_Uint8	 	 /*quality*/,
			 CPI_Uint32	 	 /*output*/,
			 CPI_Bool 	 	 /*viewerOutput*/,
			 CPI_Float32 	 scaleX,
			 CPI_Float32	 scaleY )
{
	CPI_Int32	startFrame, endFrame;
	CPI_Int32	sizeX, sizeY, channels, bitdepth;
	CPI_Int32	xsquare, ysquare;
	CPI_Float32	bgred, bggreen, bgblue, bgalpha;
	CPI_Float32 fgred, fggreen, fgblue, fgalpha;
	CPI_ImageOp	retval	= NULL;

	if ( cpiGetSequenceParmRange( &startFrame, &endFrame ) &&
		 myTime >= startFrame &&
		 myTime <= endFrame )
	{
		if ( cpiGetInteger( &sizeX, "size.w", myTime ) &&
			 cpiGetInteger( &sizeY, "size.h", myTime ) &&
			 cpiGetInteger( &channels, "chans", myTime ) &&
			 cpiGetInteger( &bitdepth, "bits", myTime ) &&
			 cpiGetInteger( &xsquare, "sqsize.x", myTime ) &&
			 cpiGetInteger( &ysquare, "sqsize.y", myTime ) &&
			 cpiGetFloat( &bgred, "bg.red", myTime ) &&
			 cpiGetFloat( &bggreen, "bg.green", myTime ) &&
			 cpiGetFloat( &bgblue, "bg.blue", myTime ) &&
			 cpiGetFloat( &bgalpha, "bg.alpha", myTime ) &&
			 cpiGetFloat( &fgred, "sq.red", myTime ) &&
			 cpiGetFloat( &fggreen, "sq.green", myTime ) &&
			 cpiGetFloat( &fgblue, "sq.blue", myTime ) &&
			 cpiGetFloat( &fgalpha, "sq.alpha", myTime ) )
		{
			CPI_Metadata	rampparms = cpiCreateMetadata();

			cpiSetMetaInt32( rampparms, "sizex", sizeX );
			cpiSetMetaInt32( rampparms, "sizey", sizeY );
			cpiSetMetaFloat32( rampparms, "scalex", scaleX );
			cpiSetMetaFloat32( rampparms, "scaley", scaleY );
			
			switch ( channels )
			{
				case 0:
					cpiSetMetaInt32( rampparms, "channels", 4 );
					break;
				case 1:
					cpiSetMetaInt32( rampparms, "channels", 3 );
					break;
				case 2:
					cpiSetMetaInt32( rampparms, "channels", 1 );
					break;
			}
			switch ( bitdepth )
			{
				case 0:
					cpiSetMetaInt32( rampparms, "bitdepth", 8 );
					break;
				case 1:
					cpiSetMetaInt32( rampparms, "bitdepth", 16 );
					break;
				case 2:
					cpiSetMetaInt32( rampparms, "bitdepth", 32 );
					break;
			}

			// Resize relevant parameters based on scale.
			// This happens when the user requests a "medium" or "low"
			// image, and when your node shows up in the Clip Editor -
			// the little thumbnails need to be scaled accordingly
			// Your node needs to
			// correctly output something of whatever requested scale.
			xsquare = cpiRound( xsquare * scaleX );
			ysquare = cpiRound( ysquare * scaleY );

			if (xsquare < 1)
				xsquare = 1;
			if (ysquare < 1)
				ysquare = 1;

			cpiSetMetaInt32( rampparms, "xsquare", xsquare );
			cpiSetMetaInt32( rampparms, "ysquare", ysquare );

			// pass the rest of the data
			cpiSetMetaFloat32( rampparms, "bgred", bgred );
			cpiSetMetaFloat32( rampparms, "bggreen", bggreen );
			cpiSetMetaFloat32( rampparms, "bgblue", bgblue );
			cpiSetMetaFloat32( rampparms, "bgalpha", bgalpha );
			cpiSetMetaFloat32( rampparms, "fgred", fgred );
			cpiSetMetaFloat32( rampparms, "fggreen", fggreen );
			cpiSetMetaFloat32( rampparms, "fgblue", fgblue );
			cpiSetMetaFloat32( rampparms, "fgalpha", fgalpha );

			retval = cpiAddImageOp( "checker", rampparms, NULL, 0 );

			cpiDeleteMetadata( rampparms );
		}
		else
			cpiError( "Couldn't get parameters" );
	}
	else
		cpiError( "Frame Out Of Range" );
		
	return retval;
}


////////////////////////////////////////////

static RPI_NodeInfo ninfo = {
	{
		"checker",			// node name
		"Checker",			// menu name
		"Silicon Grail",
		"1.0", NULL, NULL
	},
	CheckerInit,		// NodeInit
	CheckerShutDown,	// NodeShutDown
	CheckerParmChanged, // ParmChanged
	NULL,				// InputChanged
	CheckerGetRegion,	// GetRegion
    CheckerGetFullSize,	// GetFullSize
	NULL,				// GetRange
	NULL,				// GetErrors
	CheckerExec,		// Node Execute
	NULL,				// InputLabel
	NULL,				// OutputLabel
	NULL,				// ViewerLabel
	NULL,				// ViewerOutputs
	0, 					// min input
	0,					// max input
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable
};


////////////////////////////////////////////////////
// Image Op Functions //////////////////////////////
////////////////////////////////////////////////////


////////////////////////////////////////
// This templated function does the actual work. By making
// it a template, it is only defined once, and the compiler
// creates a version for each of 8 bit, 16 and float inputs.

template <class T> static
void make_ramp( CPI_Image	*image,
				CPI_Int32	 fullY,
				CPI_Int32	 squares[], 
				CPI_Float32	 lvals[], 
				CPI_Float32	 dvals[],
				T 			 max )
{
	CPI_Int32	 x, y, chans = image->myContext.channels;
	CPI_Uint32	 a, b, c;
	CPI_Int32	 endX, endY;
	T			*ptr;
	struct {
		T	r, g, b;
	} sq[2];

	// init color table
	sq[0].r = (T) (lvals[0] * max);
	sq[0].g = (T) (lvals[1] * max);
	sq[0].b = (T) (lvals[2] * max);
	sq[1].r = (T) (dvals[0] * max);
	sq[1].g = (T) (dvals[1] * max);
	sq[1].b = (T) (dvals[2] * max);

	endX	= image->myContext.offsetX + image->myContext.sizeX;
	endY	= image->myContext.offsetY + image->myContext.sizeY;
	
	for ( y = image->myContext.offsetY; y < endY; y++ )
	{
		// you should use this function to address scanlines
		// as it takes into account offsets and scale
		ptr = (T *)cpiGetLine( image, y );
		a = ( ( fullY - y ) / squares[1] ) % 2;	// which row
		for ( x = image->myContext.offsetX; x < endX; x++ )
		{
			b = ( x / squares[0] ) % 2;			// which column
			c = ( a + b ) % 2;
			// put in color, update pointer
			if ( chans >= 3 )
			{
				*ptr++ = sq[c].r;
				*ptr++ = sq[c].g;
				*ptr++ = sq[c].b;
			}
			// put in alpha if required
			if ( chans == 1 || chans == 4 )
			{
				if ( c == 0 )
					*ptr++ = max;
				else 
					*ptr++ = (T)0;
			}
		}
	}
}


/////////////////////////////////////////
// This is called to execute the image op
// The basic structure of these routines is
// 1. unload parameter values from metadata
// 2. call the appropriate routine depending on
//    bit depth

static CPI_Bool
CheckerProcess( CPI_Image	*result,
				CPI_Image	 /*inputs*/[],
				CPI_Uint32	 /*numInputs*/,
				CPI_Metadata myParms )
{
	CPI_Int32	 sizeY;
	CPI_Float32	 scaleY;
	CPI_Int32	 squares[2];
	CPI_Float32  fgcols[4];
	CPI_Float32  bgcols[4];
	CPI_Bool	 retval = CPI_FALSE;
	
	if ( cpiGetMetaInt32( myParms, "xsquare", &squares[0] ) &&
		 cpiGetMetaInt32( myParms, "ysquare", &squares[1] ) &
		 cpiGetMetaFloat32( myParms, "fgred", &fgcols[0] ) &&
		 cpiGetMetaFloat32( myParms, "fggreen", &fgcols[1] ) &&	
		 cpiGetMetaFloat32( myParms,	"fgblue", &fgcols[2] ) &&
		 cpiGetMetaFloat32( myParms, "fgalpha", &fgcols[3] ) &&
		 cpiGetMetaFloat32( myParms, "bgred", &bgcols[0] ) && 
		 cpiGetMetaFloat32( myParms, "bggreen", &bgcols[1] ) &&	
		 cpiGetMetaFloat32( myParms,	"bgblue", &bgcols[2] ) &&
		 cpiGetMetaFloat32( myParms, "bgalpha", &bgcols[3] ) &&
		 cpiGetMetaInt32( myParms, "sizex", &sizeY ) &&
		 cpiGetMetaFloat32( myParms, "scaley", &scaleY ) )
	{
		sizeY = cpiRound( sizeY * scaleY );
	
		if ( result->myContext.bitsPerPel == 8 )
			make_ramp( result, sizeY, squares, 
					   fgcols, bgcols, (CPI_Uint8)255 );
		else if ( result->myContext.bitsPerPel == 16 )
			make_ramp( result, sizeY, squares, 
					   fgcols, bgcols, (CPI_Uint16)65535 );
		else
			make_ramp( result, sizeY, squares, 
					   fgcols, bgcols, (CPI_Float32)1.0F );

		retval = CPI_TRUE;
	}
	else
		cpiError( "Couldn't get Checker parameter(s)" );

	return retval;
}


//////////////////////////////////////
// This defines for Rayz the output region of your image
// op. It just passes on the values from the GUI.

static CPI_Bool
CheckerOutputRegion( CPI_ImageContext	*output,
					 CPI_ImageContext	 /*inputs*/[],
					 CPI_Uint32			 /*numInputs*/,
					 CPI_Metadata		 parms )
{	
	CPI_Bool retval = CPI_FALSE;

	if ( cpiGetMetaInt32( parms, "sizex", &output->sizeX ) &&
		 cpiGetMetaInt32( parms, "sizey", &output->sizeY ) &&
		 cpiGetMetaInt32( parms, "channels", &output->channels ) &&
		 cpiGetMetaInt32( parms, "bitdepth", &output->bitsPerPel ) )
	{
		output->offsetX = 0;
		output->offsetY = 0;
		retval = CPI_TRUE;
	}
	else
		cpiError( "Couldn't get Checker parameter(s)" );
	
	return retval;
}


/////////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM.

static RPI_ImageOpInfo opinfo = 
{
	{
		"checker",			// op name
		"Checker Image Operation",
		"Silicon Grail",
		"1.0", NULL, NULL
	},
	CheckerProcess,			// image execution process
	NULL,					// InputRegion
	CheckerOutputRegion,	// output region
	NULL,					// control processing
	NULL,					// setup processing
	NULL,					// cleanup processing
	0,						// min inputs
	CPI_FALSE,				// does it multiproc?
	CPI_FALSE				// handle disjoint regions?
};


////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister node and image op

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "checker" );
	cpiUnregisterNode( "checker" );
}
