/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  710 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	CircleRamp
//
// COMMENTS:
//
*/

#include <stdio.h>
#include <math.h>

#include <CPI.h>

//////////////////////////////////////

static const char *channelmenu[]= 
{
	"RGBA",
	"RGB",
	"Alpha",
	NULL
};


///////////////////////////////////////


static const char *bitdepthmenu[] = 
{
	"8 bit",
	"16 bit",
	"Float",
	NULL
};


///////////////////////////////////////


static CPI_Float32 centerdefs[] = { 1.0F, 1.0F, 1.0F, 1.0F };
static CPI_Float32 edgedefs[] = { 0.0F, 0.0F, 0.0F, 0.0F };


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
// Node Functions /////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

static CPI_Bool
CircleRampInit( void )
{
	cpiAddColorGroup( "center_colors",
					  "Center Colors", 
					  centerdefs, 
					  4, 
					  CPI_PARM_DEFAULT,
					  "Adjust the color at the center of the ramp",
					  NULL );
	cpiAddColorGroup( "edge_colors",
					  "Edge Colors", 
					  edgedefs, 
					  4, 
					  CPI_PARM_DEFAULT,
					  "Adjust the color at the edges of the ramp",
					  NULL );

	cpiAddFloatPair( "center",
					 "Center", 
					 "x",
					 "X",
					 0.5, 
					 "y",
					 "Y",
					 0.5, 
					 CPI_PARM_ANIMATABLE,
					 "Adjust the X/Y position of the center of the ramp",
					 NULL );
	cpiAddFloatPair( "radius",
					 "Radius", 
					 "in",
					 "In", 
					 0.0, 
					 "out",
					 "Out", 
					 1.0, 
					 CPI_PARM_ANIMATABLE,
					 "Adjust the In/Out radius of the ramp",
					 NULL );
	cpiAddFloat( "aspect",
				 "Aspect", 
				 1.0, 
				 0.01, 
				 2, 
				 CPI_PARM_ANIMATABLE | CPI_PARM_CLAMPMIN,
				 "Adjust the shape of the ramp",
				 NULL );
	cpiAddIntPair( "size",
				   "Size",
				   "width",
				   "Width", 
				   512, 
				   "height",
				   "Height", 
				   512, 
				   CPI_PARM_DEFAULT,
				   "Adjust the X/Y size of the output image",
				   NULL );
	cpiAddMenu( "bit_depth",
				"Bit Depth", 
				0, 
				bitdepthmenu,
				"Bit depth to be output",
				NULL );
	cpiAddMenu( "channels",
				"Channels", 
				0,
				channelmenu,
				"Number of channels to output",
				NULL );

	cpiAddSequenceRange( 1, 
						 1, 
						 CPI_PARM_DEFAULT, 
						 "Frame Numbers To Output",
						 NULL );

	return CPI_TRUE;
}

///////////////////////////////////////////


static CPI_Bool
CircleRampShutDown( void )
{
	return CPI_TRUE;
}


///////////////////////////////////////////


static CPI_Bool
CircleRampGetRegion( CPI_ImageContext	*output,
					 CPI_Float32		 myTime,
					 CPI_Uint32			 /*nodeOutput*/,
					 CPI_Bool		 	 /*viewerOutput*/ )
{
	CPI_Int32	 startFrame, endFrame;
	CPI_Bool	 retval = CPI_TRUE;

	if ( cpiGetSequenceParmRange( &startFrame, &endFrame ) &&
		 startFrame <= myTime &&
		 endFrame >= myTime )
	{
		CPI_Int32	bitsPerPel, channels;
	
		if ( cpiGetInteger( &output->sizeX, "size.width", myTime ) &&
			 cpiGetInteger( &output->sizeY, "size.height", myTime ) &&
			 cpiGetInteger( &bitsPerPel, "bit_depth", myTime ) &&
			 cpiGetInteger( &channels, "channels", myTime ) )
		{
			output->offsetX = 0;
			output->offsetY = 0;

			switch ( bitsPerPel )
			{
				case 0:
					output->bitsPerPel = 8;
					break;
				case 1:
					output->bitsPerPel = 16;
					break;
				case 2:
					output->bitsPerPel = 32;
					break;
				default:
					cpiError( "Invalid bitsPerPel in getRegion!\n" );
					retval = CPI_FALSE;
					break;
			}

			switch ( channels )
			{
				case 0:
					output->channels = 4;
					break;
				case 1:
					output->channels = 3;
					break;
				case 2:
					output->channels = 1;
					break;
				default:
					cpiError( "Invalid channels in getRegion!\n" );
					retval = CPI_FALSE;
					break;
			}

			output->isPremultiplied = false;
		}
		else
		{
			cpiError( "Couldn't get parameter(s)" );
			retval = CPI_FALSE;
		}
	}
	else
	{
		cpiError( "Frame Out Of Range" );
		retval = CPI_FALSE;
	}
	
	return retval;
}


///////////////////////////////////////////


static void
CircleRampParmChanged( const char *name )
{
	// Only sync with the global range when the node is first placed down
	if ( name == NULL )
	{
		CPI_Int32	 startFrame, endFrame;

		if ( cpiGetGlobalFrameRange( &startFrame, &endFrame ) )
			cpiSetSequenceParmRange( startFrame, endFrame );
	}
}


////////////////////////////////////////


static CPI_Bool
CircleRampGetFullSize( CPI_ImageContext	*fullsize,
					   CPI_Float32		 myTime,
					   CPI_Uint32		 /*output*/,
					   CPI_Bool		 	 /*viewerOutput*/ )
{
	CPI_Int32	 startFrame, endFrame;
	CPI_Bool 	 retval = CPI_FALSE;

	if ( cpiGetSequenceParmRange( &startFrame, &endFrame ) &&
		 myTime >= startFrame &&
		 myTime <= endFrame )
	{
		if ( cpiGetInteger( &fullsize->sizeX, "size.width", myTime ) &&
			 cpiGetInteger( &fullsize->sizeY, "size.height", myTime ) )
			retval = CPI_TRUE;
		else
			cpiError( "Couldn't get full size parameters" );
	}
	else
		cpiError( "Frame Out Of Range" );
	
	return retval;
}


//////////////////////////////////////////


static CPI_ImageOp
CircleRampExec( CPI_Float32		myTime,
				CPI_Uint8		/*quality*/,
				CPI_Uint32		/*output*/,
				CPI_Bool		/*viewerOutput*/,
				CPI_Float32		scaleX,
				CPI_Float32		scaleY )
{
	CPI_Int32	 startFrame, endFrame;
	CPI_ImageOp	 retval	= NULL;

	if ( cpiGetSequenceParmRange( &startFrame, &endFrame ) &&
		 myTime >= startFrame &&
		 myTime <= endFrame )
	{
		CPI_Int32	 sizeX, sizeY, channels, bitdepth;
		CPI_Float32	 centerX, centerY;
		CPI_Int32	 intCenterX, intCenterY;
		CPI_Float32	 radiusX, radiusY;
		CPI_Float32	 aspect;
		CPI_Float32	 edgered, edgegreen, edgeblue, edgealpha;
		CPI_Float32  centerred, centergreen, centerblue, centeralpha;
		CPI_Float32	 maxdim;

		if ( cpiGetInteger( &sizeX, "size.width", myTime ) &&
			 cpiGetInteger( &sizeY, "size.height", myTime ) &&
			 cpiGetInteger( &channels, "channels", myTime ) &&
			 cpiGetInteger( &bitdepth, "bit_depth", myTime ) &&
			 cpiGetFloat( &centerX, "center.x", myTime ) &&
			 cpiGetFloat( &centerY, "center.y", myTime ) &&
			 cpiGetFloat( &radiusX, "radius.in", myTime ) &&
			 cpiGetFloat( &radiusY, "radius.out", myTime ) &&
			 cpiGetFloat( &aspect, "aspect", myTime ) &&
			 cpiGetFloat( &edgered, "edge_colors.red", myTime ) &&
			 cpiGetFloat( &edgegreen, "edge_colors.green", myTime ) &&
			 cpiGetFloat( &edgeblue, "edge_colors.blue", myTime ) &&
			 cpiGetFloat( &edgealpha, "edge_colors.alpha", myTime ) &&
			 cpiGetFloat( &centerred, "center_colors.red", myTime ) &&
			 cpiGetFloat( &centergreen, "center_colors.green", myTime ) &&
			 cpiGetFloat( &centerblue, "center_colors.blue", myTime ) &&
			 cpiGetFloat( &centeralpha, "center_colors.alpha", myTime ) )
		{
			CPI_Metadata	rampparms = cpiCreateMetadata();

			sizeX = cpiRound( sizeX * scaleX );
			sizeY = cpiRound( sizeY * scaleY );

			switch ( channels )
			{
				case 0:
					cpiSetMetaInt32( rampparms, "channels", 4 );
					break;
				case 1:
					cpiSetMetaInt32( rampparms, "channels", 3 );
					break;
				case 2:
					cpiSetMetaInt32( rampparms, "channels", 1 );
					break;
			}

			switch ( bitdepth )
			{
				case 0:
					cpiSetMetaInt32( rampparms, "bitdepth", 8 );
					break;
				case 1:
					cpiSetMetaInt32( rampparms, "bitdepth", 16 );
					break;
				case 2:
					cpiSetMetaInt32( rampparms, "bitdepth", 32 );
					break;
			}

			intCenterX = cpiRound(centerX * sizeX);
			intCenterY = cpiRound((1.0 - centerY) * sizeY);

			cpiSetMetaInt32( rampparms, "sizex", sizeX );
			cpiSetMetaInt32( rampparms, "sizey", sizeY );
			cpiSetMetaInt32( rampparms, "centerx", intCenterX);
			cpiSetMetaInt32( rampparms, "centery", intCenterY );
			cpiSetMetaFloat32( rampparms, "radiusx", radiusX );
			cpiSetMetaFloat32( rampparms, "radiusy", radiusY );
			cpiSetMetaFloat32( rampparms, "aspect", aspect );
			cpiSetMetaFloat32( rampparms, "edgered", edgered );
			cpiSetMetaFloat32( rampparms, "edgegreen", edgegreen );
			cpiSetMetaFloat32( rampparms, "edgeblue", edgeblue );
			cpiSetMetaFloat32( rampparms, "edgealpha", edgealpha );
			cpiSetMetaFloat32( rampparms, "centerred", centerred );
			cpiSetMetaFloat32( rampparms, "centergreen", centergreen );
			cpiSetMetaFloat32( rampparms, "centerblue", centerblue );
			cpiSetMetaFloat32( rampparms, "centeralpha", centeralpha );

			maxdim = 0.5 * ( ( sizeX > sizeY ) ? sizeY : sizeX ); 
			cpiSetMetaFloat32( rampparms, "maxdim", maxdim );

			retval = cpiAddImageOp( "circleramp", rampparms, NULL, 0 );

			cpiDeleteMetadata( rampparms );
		}
		else
			cpiError( "Couldn't get parameter(s)" );
	}
	else
		cpiError( "Frame Out Of Range" );
		
	return retval;
}


////////////////////////////////////////////

static RPI_NodeInfo ninfo = {
	{
		"circleramp",
		"CircleRamp",
		"Silicon Grail",
		"1.0", NULL, NULL
	},
	CircleRampInit,
	CircleRampShutDown,
	CircleRampParmChanged,
	NULL,		// InputChanged
	CircleRampGetRegion,
    CircleRampGetFullSize,
	NULL,		// CircleRampGetRange,
	NULL,
	CircleRampExec,
	NULL,		// InputLabel,
	NULL,		// OutputLabel,
	NULL,		// ViewerOutputLabel
	NULL,		// NumViewerOutputs
	0, 
	0,
	1,
	1,
	CPI_FALSE
};


//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// Image Op Functions ////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////


template <class T> static
void make_ramp( CPI_Image	*image, 
				CPI_Float32	 radii[],
				CPI_Int32	 center[], 
			    CPI_Uint32	 chans,
				CPI_Float32	 aspect, 
				CPI_Float32	 cvals[], 
				CPI_Float32	 evals[],
				CPI_Float32	 maxdim,
				T 			 max )
{
	CPI_Uint32	 i, x, y;
	CPI_Float32	 dx, dy, d, xcen, ycen;
	CPI_Float32	 xscale, yscale;
	CPI_Float32	 minval=0.0, maxval=1.0;
	CPI_Float32  minrad=radii[0], maxrad=radii[1], scalefac;
	CPI_Uint32	 endX, endY;
	CPI_Uint8	 parmbase;
	T			*ptr, val;

	ptr		= (T *)image->data;
	xcen	= (CPI_Float32)center[0];
	ycen	= (CPI_Float32)center[1];
	endX	= image->myContext.offsetX + image->myContext.sizeX;
	endY	= image->myContext.offsetY + image->myContext.sizeY;
	
	// next 3 lines just for boundary conditions
	d = maxrad - minrad;
	if ( fabs( d ) < 0.0001 )
		minrad = maxrad - 0.0001;

	scalefac = ( maxval - minval ) / ( maxrad - minrad );
	xscale = yscale = 1.0;

	if ( aspect < 1.F )
	   	yscale = 1 / aspect;
	else
		xscale = aspect; 

	if ( chans == 1 )
		parmbase = 3;
	else
		parmbase = 0;

	for ( y = image->myContext.offsetY;
		  y < endY && !cpiIsProcessingAborted(); y++ )
	{
		ptr = (T *)cpiGetLine( image, y );
		for ( x = image->myContext.offsetX; x < endX; x++ )
		{
			// for each (x, y) compute d, the distance to the
			// center of the image
			dx = ( (CPI_Float32)x + 0.5F - xcen ) * xscale;
			dy = ( (CPI_Float32)y + 0.5F - ycen ) * yscale;

			d = sqrt( dx * dx + dy * dy ) / maxdim;

			// make it so that 0.0 <= d <= 1.0
			if ( d < minrad ) d = minrad;
			else if ( d > maxrad ) d = maxrad;
			d = ( d - minrad ) * scalefac + minval;
			
			// put in color, update pointer
			for ( i = 0; i < chans; i++ )
			{
				val = (T)( ( ( (1 - d) * cvals[i + parmbase] ) 
							 + (d * evals[i + parmbase]) ) * max );
				*ptr++ = val;
			}
		}
	}
}


/////////////////////////////////////////


static CPI_Bool
CircleRampProcess( CPI_Image	*result,
				   CPI_Image	 /*inputs*/[],
				   CPI_Uint32	 /*numInputs*/,
				   CPI_Metadata	 myParms )
{
	CPI_Int32	center[2];
	CPI_Float32	radius[2];
	CPI_Float32 aspect;
	CPI_Float32 centercols[4];
	CPI_Float32 edgecols[4];
	CPI_Float32	maxdim;
	CPI_Bool	retval	= CPI_FALSE;

	if ( cpiGetMetaInt32( myParms, "centerx", &center[0] ) &&
		 cpiGetMetaInt32( myParms, "centery", &center[1] ) &&
		 cpiGetMetaFloat32( myParms, "radiusx", &radius[0] ) &&
		 cpiGetMetaFloat32( myParms, "radiusy", &radius[1] ) &&
		 cpiGetMetaFloat32( myParms, "aspect", &aspect ) &&
		 cpiGetMetaFloat32( myParms, "centerred", &centercols[0] ) &&
		 cpiGetMetaFloat32( myParms, "centergreen", &centercols[1] ) &&	
		 cpiGetMetaFloat32( myParms, "centerblue", &centercols[2] ) &&
		 cpiGetMetaFloat32( myParms, "centeralpha", &centercols[3] ) &&
		 cpiGetMetaFloat32( myParms, "edgered", &edgecols[0] ) &&
		 cpiGetMetaFloat32( myParms, "edgegreen", &edgecols[1] ) &&	
		 cpiGetMetaFloat32( myParms, "edgeblue", &edgecols[2] ) &&
		 cpiGetMetaFloat32( myParms, "edgealpha", &edgecols[3] ) &&
		 cpiGetMetaFloat32( myParms, "maxdim", &maxdim ) )
	{
		if ( result->myContext.bitsPerPel == 8 )
			make_ramp( result, radius, center, 
					   result->myContext.channels, aspect, centercols, 
					   edgecols, maxdim, (CPI_Uint8)255 );
		else if ( result->myContext.bitsPerPel == 16 )
			make_ramp( result, radius, center, 
					   result->myContext.channels, aspect, centercols, 
					   edgecols, maxdim, (CPI_Uint16)65535 );
		else if ( result->myContext.bitsPerPel == 32 )
			make_ramp( result, radius, center, 
					   result->myContext.channels, aspect, centercols, 
					   edgecols, maxdim, (CPI_Float32)1.0F );
		else
			cpiError( "Result image has an invalid bit depth!" );

		retval = CPI_TRUE;
	}
	else
		cpiError( "Circleramp couldn't get parameter(s)" );

	return retval;
}


//////////////////////////////////////


static CPI_Bool
CircleRampControlProcessing( CPI_Image		*/*result*/,
							 CPI_Image	 	 /*inputs*/[],
							 CPI_Uint32	 	 /*numInputs*/,
							 CPI_Metadata	 /*myParms*/ )
{
	CPI_Bool	retval = CPI_TRUE;

	retval = cpiCompute( true );

	return retval;
}


//////////////////////////////////////


static CPI_Bool
CircleRampOutputRegion( CPI_ImageContext	*output,
						CPI_ImageContext	 /*inputs*/[],
						CPI_Uint32			 /*numInputs*/,
						CPI_Metadata		 parms )
{	
	CPI_Bool retval = CPI_FALSE;

	if ( cpiGetMetaInt32( parms, "sizex", &output->sizeX ) &&
		 cpiGetMetaInt32( parms, "sizey", &output->sizeY ) && 
		 cpiGetMetaInt32( parms, "channels", &output->channels ) &&
		 cpiGetMetaInt32( parms, "bitdepth", &output->bitsPerPel ) )
	{
		output->offsetX = 0;
		output->offsetY = 0;
		retval			= CPI_TRUE;
	}
	else
		cpiError( "Circleramp couldn't get parameter(s)" );
	
	return retval;
}


/////////////////////////////////////


static RPI_ImageOpInfo opinfo = 
{
	{
		"circleramp",
		"CircleRamp Image Operation",
		"Silicon Grail",
		"1.0", NULL, NULL
	},
	CircleRampProcess,
	NULL,	//CircleRampInputRegion
	CircleRampOutputRegion,
	CircleRampControlProcessing,
	NULL,	//CircleRampSetupProcessing
	NULL,	//CircleRampCleanupProcessing
	0,
	CPI_TRUE,
	CPI_FALSE,
};


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
// Overlay Functions //////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


#define INNER_CIRCLE_ID		1
#define OUTER_CIRCLE_ID		2
#define CROSS_ID			3

typedef struct _overlayState
{
	CPI_Int32	fullSizeX;
	CPI_Int32	fullSizeY;
	CPI_Float32	aspect;
	CPI_Float32	centerX;
	CPI_Float32 centerY;
	CPI_Float32	radiusIn;
	CPI_Float32 radiusOut;
	CPI_Float32	selectedX;
	CPI_Float32 selectedY;
	CPI_Uint32	selectedObject;
	CPI_Float32	myTime;
} overlayState;


//////////////////////////////////////


static CPI_PrivateData
CircleRampCreateInstance( void )
{
	overlayState	*retval;

	retval = (overlayState *)cpiAlloc( sizeof( overlayState ) );
	
	return retval;
}


///////////////////////////////////////


static void
CircleRampDestroyInstance( CPI_PrivateData handle )
{
	overlayState *state = (overlayState *)handle;

	cpiFree( state );
}


///////////////////////////////////////


static void
CircleRampRender( CPI_PrivateData 	handle, 
				  CPI_Bool 			/*isPicking*/ )
{
	CPI_Float32	  inRad, outRad, imageAspectX, imageAspectY;
	overlayState *state = (overlayState *)handle;
	
	inRad = state->radiusIn * 0.5;
	outRad = state->radiusOut * 0.5;

	if ( inRad <= 0 )
		inRad = 0.075;
	
	if ( outRad <= 0 )
		outRad = 0.1;
	 
	if ( outRad > 1 )
		outRad = 1;

	if ( state->fullSizeX > state->fullSizeY )
	{
		imageAspectX	= ( state->fullSizeY * inRad ) / state->fullSizeX;
		imageAspectY	= inRad;
	}
	else if ( state->fullSizeX < state->fullSizeY )
	{
		imageAspectX	= inRad;
		imageAspectY	= ( state->fullSizeX * inRad ) / state->fullSizeY;
	}
	else
	{
		imageAspectX	= inRad;
		imageAspectY	= inRad;
	}

	imageAspectX *= ( 1 / state->aspect );
	imageAspectY *= state->aspect;

	cpiStartNewObject( INNER_CIRCLE_ID );
	cpiSetColor( 0.0, 0.0, 1.0 );
	cpiDrawEllipse( state->centerX, 
					state->centerY, 
					imageAspectX, 
					imageAspectY );

	if ( state->fullSizeX > state->fullSizeY )
	{
		imageAspectX	= ( state->fullSizeY * outRad ) / state->fullSizeX;
		imageAspectY	= outRad;
	}
	else if ( state->fullSizeX < state->fullSizeY )
	{
		imageAspectX	= outRad;
		imageAspectY	= ( state->fullSizeX * outRad ) / state->fullSizeY;
	}
	else
	{
		imageAspectX	= outRad;
		imageAspectY	= outRad;
	}

	imageAspectX *= ( 1 /state->aspect );
	imageAspectY *= state->aspect;

	cpiStartNewObject( OUTER_CIRCLE_ID );
	cpiDrawEllipse( state->centerX, 
					state->centerY, 
					imageAspectX, 
					imageAspectY );

	cpiStartNewObject( CROSS_ID );
	cpiDrawLine( state->centerX - 0.05, 
				 state->centerY, 
				 state->centerX + 0.05, 
				 state->centerY );
	cpiDrawLine( state->centerX, 
				 state->centerY - 0.05,
				 state->centerX, 
				 state->centerY + 0.05 );
}


//////////////////////////////////////////


static void
CircleRampSourceOutOfDate( CPI_PrivateData 	handle, 
						   CPI_ImageContext	fullsize,
						   CPI_Float32 		theTime )
{
	overlayState *state = (overlayState *)handle;

	cpiGetFloat( &state->centerX, "center.x", theTime );
	cpiGetFloat( &state->centerY, "center.y", theTime );
	cpiGetFloat( &state->radiusIn, "radius.in", theTime );
	cpiGetFloat( &state->radiusOut, "radius.out", theTime );
	cpiGetFloat( &state->aspect, "aspect", theTime );
	
	state->fullSizeX = fullsize.sizeX;
	state->fullSizeY = fullsize.sizeY;

	state->myTime = theTime;
}


///////////////////////////////////////////


static void
CircleRampLeftMouseDown( CPI_PrivateData	handle,
						 CPI_Uint32			objectID,
						 CPI_Float32		x,
						 CPI_Float32		y,
						 CPI_Uint32			/*keyMod*/ )
{
	overlayState	*state = (overlayState *)handle;

	state->selectedObject	= objectID;

	state->selectedX		= x;
	state->selectedY		= y;
	
	cpiSetSelectLock( CPI_TRUE );	
}


/////////////////////////////////////////


static void
CircleRampLeftMouseDrag( CPI_PrivateData	 handle,
						 CPI_Uint32			 /*objectID*/,
						 CPI_Float32		 x,
						 CPI_Float32		 y,
						 CPI_Uint32			 /*keyMod*/,
						 CPI_Bool			 isContinuous )
{
	overlayState	*state = (overlayState *)handle;

	if ( state->selectedObject == CROSS_ID )
	{
		state->centerX = x;
		state->centerY = y;
	}
	else if ( state->selectedObject == OUTER_CIRCLE_ID )
	{
		state->radiusOut = ( hypot( x - state->centerX, y - state->centerY ) *
							 2 );
	}
	else if ( state->selectedObject == INNER_CIRCLE_ID )
	{
		state->radiusIn = ( hypot( x - state->centerX, y - state->centerY ) * 
							2 );
	}

	if ( isContinuous )
	{
		cpiSetFloat( state->centerX, "center.x", state->myTime, CPI_TRUE );
		cpiSetFloat( state->centerY, "center.y", state->myTime, CPI_TRUE  );
		cpiSetFloat( state->radiusOut, "radius.out", state->myTime, CPI_TRUE );
		cpiSetFloat( state->radiusIn, "radius.in", state->myTime, CPI_TRUE );
	}
	
	cpiDoRedraw();   
}


/////////////////////////////////////////


static void
CircleRampLeftMouseUp( CPI_PrivateData	 handle, 
					   CPI_Uint32		 /*objectID*/,
					   CPI_Float32		 /*x*/,
					   CPI_Float32		 /*y*/,
					   CPI_Uint32		 /*keyMod*/ )
{
	overlayState	*state = (overlayState *)handle;
	
	cpiSetSelectLock( CPI_FALSE );
	
	cpiSetFloat( state->centerX, "center.x", state->myTime, CPI_TRUE );
	cpiSetFloat( state->centerY, "center.y", state->myTime, CPI_TRUE  );
	cpiSetFloat( state->radiusOut, "radius.out", state->myTime, CPI_TRUE );
	cpiSetFloat( state->radiusIn, "radius.in", state->myTime, CPI_TRUE );
}


////////////////////////////////////////


static RPI_OverlayInfo oinfo = {
	{
		"circleramp",
		"CircleRamp",
		"Silicon Grail",
		"1.0", NULL, NULL
	},
	NULL,		// Init
	NULL,		// Shutdown
	CircleRampCreateInstance,
	CircleRampDestroyInstance,
	CircleRampSourceOutOfDate,
	NULL,		// CircleRampFullSizeChanged,
	CircleRampSourceOutOfDate,
	CircleRampRender,
	CircleRampLeftMouseDown,
	CircleRampLeftMouseDrag,
	CircleRampLeftMouseUp,
	NULL,		// CircleRampMouseEnter,
	NULL,		// CircleRampMouseDrag,
	NULL,		// CircleRampMouseExit,
	"circleramp",
	CPI_FALSE
};


//////////////////////////////////////////


CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = true;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	if ( retval )
		retval = cpiRegisterOverlay( &oinfo );

	return retval;	
}


//////////////////////////////////////////////


CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "circleramp" );
	cpiUnregisterNode( "circleramp" ); 
	cpiUnregisterOverlay( "circleramp" );
}
