/*
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:    Clamp
//
// COMMENTS: This clamps the rgb values of an image
//
*/

#include <stdio.h>
#include <math.h>
#include <CPI.h>

#define MAX8	(CPI_Uint8)255
#define MAX16	(CPI_Uint16)65535

/////////////////////////////////////////
// Init is called when the node is created and is 
// typically used to define the user input controls.
// Here, there is nothing to do.

static CPI_Bool
ClampInit( void )
{
	cpiAddFloatRGBAOGroup( "minimum_list", "Min Value",
	                        0.0, 0.0, 1.0, 5, CPI_PARM_ANIMATABLE,
	                       "Minimum color values", NULL );

	cpiAddFloatRGBAOGroup( "maximum_list", "Max Value",
	                        1.0, 0.0, 1.0, 5, CPI_PARM_ANIMATABLE,
	                       "Maximum color values", NULL );

	return CPI_TRUE;
}


//////////////////////////////////////////
// This is called when the node is actually executed
// This gets the input image, and then adds the image
// operation to the list of ops to be executed

static CPI_ImageOp
ClampExec( CPI_Float32		myTime,
		   CPI_Uint8		quality,
		   CPI_Uint32		nodeOutput,
		   CPI_Bool	 		viewerOutput,
		   CPI_Float32		scaleX,
		   CPI_Float32   	scaleY )

{
	CPI_ImageOp	retval	= NULL;
	CPI_ImageOp myInputs;
	CPI_Float32	minR, minG, minB, minA;
	CPI_Float32	maxR, maxG, maxB, maxA;
	
	myInputs = cpiGetInputOp( 0, myTime, quality, scaleX, scaleY );

	// get the clamp values from the user
	// Note: don't need to get or use the overall value for each
	// group, because that is already factored in to the rgba values
	if ( cpiGetFloat( &minR, "minimum_list.red", myTime ) &&
		 cpiGetFloat( &minG, "minimum_list.green", myTime ) &&
		 cpiGetFloat( &minB, "minimum_list.blue", myTime ) &&
		 cpiGetFloat( &minA, "minimum_list.alpha", myTime ) &&
		 cpiGetFloat( &maxR, "maximum_list.red", myTime ) &&
		 cpiGetFloat( &maxG, "maximum_list.green", myTime ) &&
		 cpiGetFloat( &maxB, "maximum_list.blue", myTime ) &&
		 cpiGetFloat( &maxA, "maximum_list.alpha", myTime ) )
	{
		// create metadata and copy values to it
		CPI_Metadata    nodeparm = cpiCreateMetadata();

		cpiSetMetaFloat32( nodeparm, "minR", minR );
		cpiSetMetaFloat32( nodeparm, "minG", minG );
		cpiSetMetaFloat32( nodeparm, "minB", minB );
		cpiSetMetaFloat32( nodeparm, "minA", minA );
		cpiSetMetaFloat32( nodeparm, "maxR", maxR );
		cpiSetMetaFloat32( nodeparm, "maxG", maxG );
		cpiSetMetaFloat32( nodeparm, "maxB", maxB );
		cpiSetMetaFloat32( nodeparm, "maxA", maxA );

		if ( myInputs != NULL )
			retval = cpiAddImageOp( "clamp", nodeparm, &myInputs, 1 );

		cpiDeleteMetadata( nodeparm );
	}
	else
		cpiError( "Couldn't get Clamp parameter(s)" );
	
	return retval;
}


////////////////////////////////////////////
// This structure, which is required, defines to Rayz
// the various routines which are part of this plugin.
// Routines which are not required are set to NULL

static RPI_NodeInfo ninfo = {
	{
		"clamp",				// node name
		"Clamp",				// menu name
		"Joe Programmer",		// author
		"v1.0"					// version
	},
	ClampInit,			// NodeInit
	NULL,				// NodeShutDown
	NULL,				// ParmChanged
	NULL,				// InputChanged
	NULL,				// GetRegion
    NULL,				// GetFullSize
	NULL,				// GetRange
	NULL,				// GetErrors
	ClampExec,			// Node Execute
	NULL,				// InputLabel
	NULL,				// OutputLabel
	NULL,				// ViewerLabel
	NULL,               // ViewerOutputs
	1, 					// min inputs
	1,					// max inputs
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable (eg Mcomp)
};



//////////////////////////////////////////////////////////////////////
// Image Op Functions ////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

// This is the actual image manipulation routine. It is a templated
// function so that we can write it once, and let the compiler build
// a version for each of 8, 16 and float pixel types.
// This just copies input to result

template <class T> static
void clamp( CPI_Image	&input,
			CPI_Image	*result, 
			T 			 min[],
			T			 max[] )
{
	CPI_Int32	 x, y, endX, endY, c;
	CPI_Int32	 channels = result->myContext.channels;
	T			*pin, *pout, p;

	endX	= result->myContext.offsetX + result->myContext.sizeX;
	endY	= result->myContext.offsetY + result->myContext.sizeY;
	
	for ( y = result->myContext.offsetY; y < endY; y++ )
	{
		// Notice that we use cpiGetLine to avoid any sub-image issues
		// as rayz may ask us to process a small part of a larger image
		pin		= (T *)cpiGetLine( &input, y );
		pout	= (T *)cpiGetLine( result, y );

		for ( x = result->myContext.offsetX; x < endX; x++ )
		{
			for ( c = 0; c < channels; c++ )
			{
				/* Threshold the pixel */
				p = *pin;
				if ( p < min[c] )
					p = min[c];
				else if (p > max[c] )
					p = max[c];
				*pout = p;

				pin++;
				pout++;
			}
		}
	}
}


/////////////////////////////////////////
// This is called to execute the image op
// It just calls the appropriate copy routine, 
// based on the input bit depth

static CPI_Bool
ClampProcess( CPI_Image		*result,
			  CPI_Image	 	 inputs[],
			  CPI_Uint32	 numInputs,
			  CPI_Metadata	 myParms )
{
	CPI_Float32	 minR, minG, minB, minA;
	CPI_Float32	 maxR, maxG, maxB, maxA;
	CPI_Int32	 i;
	CPI_Bool	 retval = CPI_FALSE;
	
	// get the values from metadata
	if ( cpiGetMetaFloat32( myParms, "minR", &minR ) &&
		 cpiGetMetaFloat32( myParms, "minG", &minG ) &&
		 cpiGetMetaFloat32( myParms, "minB", &minB ) &&
		 cpiGetMetaFloat32( myParms, "minA", &minA ) &&
		 cpiGetMetaFloat32( myParms, "maxR", &maxR ) &&
		 cpiGetMetaFloat32( myParms, "maxG", &maxG ) &&
		 cpiGetMetaFloat32( myParms, "maxB", &maxB ) &&
		 cpiGetMetaFloat32( myParms, "maxA", &maxA ) )
	{
		// set up to call appropriate routine
		// no bounds checking or sanity checks on values
		if ( result->myContext.bitsPerPel == 8 )
		{
			CPI_Uint8	min[10], max[10];

			min[0] = cpiRound( minR * MAX8 );
			min[1] = cpiRound( minG * MAX8 );
			min[2] = cpiRound( minB * MAX8 );
			min[3] = cpiRound( minA * MAX8 );

			max[0] = cpiRound( maxR * MAX8 );
			max[1] = cpiRound( maxG * MAX8 );
			max[2] = cpiRound( maxB * MAX8 );
			max[3] = cpiRound( maxA * MAX8 );
			for ( i = 4; i < 10; i++)
			{
				min[i] = (CPI_Uint8)0;
				max[i] = MAX8;
			}

			clamp( inputs[0], result, min, max );
		}
		else if ( result->myContext.bitsPerPel == 16 )
		{
			CPI_Uint16	min[10], max[10];

			min[0] = cpiRound( minR * MAX16 );
			min[1] = cpiRound( minG * MAX16 );
			min[2] = cpiRound( minB * MAX16 );
			min[3] = cpiRound( minA * MAX16 );

			max[0] = cpiRound( maxR * MAX16 );
			max[1] = cpiRound( maxG * MAX16 );
			max[2] = cpiRound( maxB * MAX16 );
			max[3] = cpiRound( maxA * MAX16 );
			for ( i = 4; i < 10; i++)
			{
				min[i] = (CPI_Uint16)0;
				max[i] = MAX16;
			}

			clamp( inputs[0], result, min, max );
		}
		else
		{
			CPI_Float32	min[10], max[10];

			min[0] = minR;
			min[1] = minG;
			min[2] = minB;
			min[3] = minA;

			max[0] = maxR;
			max[1] = maxG;
			max[2] = maxB;
			max[3] = maxA;
			for ( i = 4; i < 10; i++)
			{
				min[i] = 0.0F;
				max[i] = 1.0F;
			}

			clamp( inputs[0], result, min, max );
		}

		retval = CPI_TRUE;
	}
	else
		cpiError( "Couldn't get clamp parameter(s)" );

	return retval;
}


/////////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM

static RPI_ImageOpInfo opinfo = 
{
	{
		"clamp",			// command name
		"",			        // not used
		"Joe Programmer",	// author
		"0.1"				// version
	},
	ClampProcess,			// image execution process
	NULL,					// InputRegion
	NULL,					// OutputRegion,
	NULL,                   // ControlProcessing
	NULL,                   // SetupProcessing
	NULL,                   // CleanupProcessing
	1,						// min number of inputs required
	CPI_TRUE,				// Can this plugin multiproc?
	CPI_FALSE				// Disjoint regions?
};


////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister node and image op

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "clamp" );
	cpiUnregisterNode( "clamp" );
}
