/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	CropCircle
//
// COMMENTS: This demonstrates creating and using the overlays
// 			 to interactively control user parameter values. It
//			 crops an image to a user-defined circle. No anti-
//			 aliasing is attempted.
//
*/

#include <stdio.h>
#include <math.h>
#include <CPI.h>

//////////////////////////////////////
// The init routine is called once, and typically is used to
// define the user input controls

static CPI_Bool
CropCircleInit( void )
{
	cpiAddFloatPair( "center", "Center", 
					 "x", "X", 0.5, "y", "Y", 0.5, 
					  CPI_PARM_ANIMATABLE,
					  NULL, NULL );
	cpiAddFloat( "radius", "Radius", 1.0, 0.001, 1.0, 
				 CPI_PARM_ANIMATABLE | CPI_PARM_CLAMPMIN,
				 NULL, NULL );

	return CPI_TRUE;
}


//////////////////////////////////////////
// This is called when an image is required
// The basic structure of these routines is:
// 1. get the parameters from the user interface
// 2. allocate metadata and copy the parameters to it
// 3. add the image op(s) to the list of operations

static CPI_ImageOp
CropCircleExec( CPI_Float32		myTime,
				CPI_Uint8		quality,
				CPI_Uint32		output,
				CPI_Bool	 	viewerOutput,
				CPI_Float32		scaleX,
				CPI_Float32   	scaleY )

{
	CPI_ImageOp	 retval	= NULL;
	CPI_ImageOp  myInputs;
	CPI_Float32	 centerX, centerY, radius;
	CPI_Int32	 fullX, fullY;
	CPI_Int32	 intCenterX, intCenterY;
	
	// get input. If not connected, abandon ship
	myInputs = cpiGetInputOp( 0, myTime, quality, scaleX, scaleY );
	if ( myInputs == NULL )
		return retval;

	if ( !cpiGetInputFullSize( &fullX, &fullY, 0, myTime ) )
		return retval;
	
	fullX = cpiRound( fullX * scaleX );
	fullY = cpiRound( fullY * scaleY );
	
	// unload values from GUI
	if ( cpiGetFloat( &centerX, "center.x", myTime ) &&
		 cpiGetFloat( &centerY, "center.y", myTime ) &&
		 cpiGetFloat( &radius, "radius", myTime ) )
	{
		intCenterX = cpiRound( fullX * centerX );
		intCenterY = fullY - cpiRound( fullY * centerY );

		// allocate metadata and copy parameter values to it
		CPI_Metadata    nodeparms = cpiCreateMetadata();

		if ( fullX >= fullY )
			radius = radius * 0.5 * fullY;
		else 
			radius = radius * 0.5 * fullX;
	
		cpiSetMetaInt32( nodeparms, "centerx", intCenterX );
		cpiSetMetaInt32( nodeparms, "centery", intCenterY );
		cpiSetMetaFloat32( nodeparms, "radius", radius );
	   
		// add the "cropcircle" op to the list
		retval = cpiAddImageOp( "cropcircle", nodeparms, &myInputs, 1 );

		cpiDeleteMetadata( nodeparms );
	}
	else
		cpiError( "Couldn't get CropCircle parameter(s)" );
		
	return retval;
}


////////////////////////////////////////////
// This structure, which is required, defines to Rayz
// the various routines which are part of this plugin.
// Routines which are not required are set to NULL

static RPI_NodeInfo ninfo = {
	{
		"cropcircle",		// plugin name
		"CropCircle",		// menu name
		"Silicon Grail",	// author
		"v1.0"				// version
	},
	CropCircleInit,		// NodeInit
	NULL,				// NodeShutDown
	NULL,				// ParmChanged
	NULL,				// InputChanged
	NULL,				// GetRegion
    NULL,				// GetFullSize
	NULL,				// GetRange
	NULL,				// GetErrors
	CropCircleExec,		// Node Execute
	NULL,				// InputLabel
	NULL,				// OutputLabel
	NULL,				// ViewerLabel
	NULL,				// ViewerOutputs
	1, 					// min inputs
	1,					// max inputs
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable (eg Mcomp)
};


//////////////////////////////////////////////////
// Image Op Functions ////////////////////////////
//////////////////////////////////////////////////

////////////////////////////////////////
// This templated function does the actual work. By making
// it a template, it is only defined once, and the compiler
// creates a version for each of 8 bit, 16 and float inputs.

template <class T> static
void cropcircle( CPI_Image	   &input,
				 CPI_Image	   *result, 
				 CPI_Int32 		center[],
				 CPI_Float32 	radius,
				 T 			    max)
{
	CPI_Int32    i, x, y, chans = input.myContext.channels;
	CPI_Float32  dx, dy, d;
	CPI_Int32	 endX, endY;
	T           *pin, *pout;

	endX    = input.myContext.offsetX + input.myContext.sizeX;
	endY    = input.myContext.offsetY + input.myContext.sizeY;

	for ( y = input.myContext.offsetY; y < endY; y++ )
	{
		// use the GetLine() function to get the address of
		// a scanline
		pin  = (T *)cpiGetLine( &input, y );
		pout = (T *)cpiGetLine( result, y );
		for ( x = input.myContext.offsetX; x < endX; x++ )
		{
			// for each (x, y) compute d, the distance to the
			// center of the image
			dx = x - center[0];
			dy = y - center[1];

			// if outside circle, set to black
			// this doesn't anti-alias - that's outside the
			// scope of this example
			d = sqrt( dx * dx + dy * dy );
			if ( d > radius )
			{
				for (i = 0; i < chans; i++ )
				{
					*pout++ = (T)0;
					 pin++;
				}
			}
			else
			{
				for (i = 0; i < chans; i++ )
					*pout++ = *pin++;
			}
		}
	}
}


/////////////////////////////////////////
// This is called to execute the image op
// The basic structure of these routines is
// 1. unload parameter values from metadata
// 2. call the appropriate routine depending on
//    bit depth

#define MAX8	(CPI_Uint8)255
#define MAX16	(CPI_Uint16)65535
#define MAXF	(CPI_Float32)1.0F

static CPI_Bool
CropCircleProcess( CPI_Image	*result,
				   CPI_Image 	 inputs[],
				   CPI_Uint32	 numInputs,
				   CPI_Metadata  myParms )
{
	CPI_Float32	 radius;
	CPI_Int32	 center[2];
	CPI_Bool	 retval = CPI_FALSE;
	
	if ( cpiGetMetaInt32( myParms, "centerx", &center[0] ) &&
		 cpiGetMetaInt32( myParms, "centery", &center[1] ) &&
		 cpiGetMetaFloat32( myParms, "radius", &radius ) )
	{
		if ( result->myContext.bitsPerPel == 8 )
			cropcircle( inputs[0], result, center, radius, MAX8 );
		else if ( result->myContext.bitsPerPel == 16 )
			cropcircle( inputs[0], result, center, radius, MAX16 );
		else
			cropcircle( inputs[0], result, center, radius, MAXF );

		retval = CPI_TRUE;
	}
	else
		cpiError( "Couldn't get CropCircle parameter(s)" );
		
	return retval;
}


/////////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM.

static RPI_ImageOpInfo opinfo = 
{
	{
		"cropcircle",		// command name
		"",					// not used
		"silicon grail",	// author
		"1.0"				// version
	},
	CropCircleProcess,		// image execution process
	NULL,					// InputRegion
	NULL,					// OutputRegion,
	NULL,					// ControlProcessing
	NULL,					// SetupProcessing
	NULL,					// CleanupProcessing
	1,						// min number of inputs required
	CPI_FALSE,				// Can this plugin multiproc?
	CPI_FALSE				// Disjoint regions?
};


///////////////////////////////////////////////////////////
// Overlay Functions //////////////////////////////////////
///////////////////////////////////////////////////////////
/*
   This section defines the overlay routines. They are based
   around a structure, just as the node and image op are.
*/


// typically, you define an overlayState structure to hold
// the current state of what you are drawing. This is then
// handed from function to function as private data.

#define CIRCLE_ID		1
#define CROSS_ID		2

typedef struct _overlayState
{
	CPI_Int32   fullSizeX;
	CPI_Int32   fullSizeY;
	CPI_Float32	centerX;
	CPI_Float32 centerY;
	CPI_Float32	radius;
	CPI_Float32	startRad;
	CPI_Uint32	selectedObject;
	CPI_Float32	myTime;
} overlayState;


//////////////////////////////////////
// This creates the overlay you are working with, by allocating
// the defined data structure and passing it on.

static CPI_PrivateData
CropCircleCreateInstance( void )
{
	overlayState	*retval;

	retval = (overlayState *)cpiAlloc( sizeof( overlayState ) );
	
	return retval;
}


///////////////////////////////////////
// cleans up before exiting

static void
CropCircleDestroyInstance( CPI_PrivateData handle )
{
	overlayState *state = (overlayState *)handle;

	cpiFree( state );
}


///////////////////////////////////////
// You need to define a render routine. This will use the
// drawing commands which you have available - see
// include/CPI_Overlay.h for a complete list.

static void
CropCircleRender( CPI_PrivateData handle, CPI_Bool isPicking )
{
	CPI_Float32	  radius, xAspect, yAspect, xHairSize, yHairSize;
	overlayState *state = (overlayState *)handle;
	
	radius = state->radius * 0.5;

	if ( radius <= 0 )
		radius = 0.075;
	
	if ( radius > 1 )
		radius = 1;

	// correct for aspect ratio to get a round circle
	if ( state->fullSizeX >= state->fullSizeY )
	{
		xAspect = ( state->fullSizeY * radius ) / state->fullSizeX;
		yAspect = radius;
	}
	else
	{
		xAspect = radius;
		yAspect = ( state->fullSizeX * radius ) / state->fullSizeY;
	}

	// each collection of lines or shapes that you want to be 
	// separately selectable should be delimited by a call to
	// cpiStartNewObject. The ID that you pass will be returned
	// in the selection routine, so you know what the user has
	// selected.
	cpiStartNewObject( CIRCLE_ID );
	cpiSetColor( 0.1, 0.1, 1.0 );
	cpiDrawEllipse( state->centerX, state->centerY, xAspect, yAspect );

	// define a second 'object' - the crosshairs
	xHairSize = 0.1 * xAspect;
	yHairSize = 0.1 * yAspect;
	cpiStartNewObject( CROSS_ID );
	cpiDrawLine( state->centerX - xHairSize, 
				 state->centerY, 
				 state->centerX + xHairSize, 
				 state->centerY );
	cpiDrawLine( state->centerX, 
				 state->centerY - yHairSize,
				 state->centerX, 
				 state->centerY + yHairSize );
}


//////////////////////////////////////////
// If the node is asked to recompute for any reason (such
// as a parameter change, either in this node or elsewhere
// in Rayz), this routine is called. Here, we refresh our
// user input parameters.

static void
CropCircleSourceOutOfDate( CPI_PrivateData	 handle, 
						   CPI_ImageContext	 fullsize,
						   CPI_Float32 		 theTime )
{
	overlayState *state = (overlayState *)handle;

	cpiGetFloat( &state->centerX, "center.x", theTime );
	cpiGetFloat( &state->centerY, "center.y", theTime );
	cpiGetFloat( &state->radius, "radius", theTime );

	state->fullSizeX = fullsize.sizeX;
	state->fullSizeY = fullsize.sizeY;

	state->myTime = theTime;
}


////////////////////////////////////////////
// When an input image changes, this routine is called to
// update any values which may be based on that image.

static void
CropCircleSourceChanged( CPI_PrivateData handle, 
						 CPI_ImageContext fullsize,
						 CPI_Float32 theTime )
{
	CropCircleSourceOutOfDate( handle, fullsize, theTime );
}


// compute the radius of a circle based on delta x and delta y
// cursor values

static CPI_Float32
getRad( CPI_Float32 x, CPI_Float32 y, CPI_Float32 cx, CPI_Float32 cy)
{
	return ( hypot( x - cx, y - cy ) * 2 );
}


///////////////////////////////////////////
// These routines are called to service the appropriate
// hardware activity. They are only called if you have a current
// object under the cursor. The input values are
// handle - your private structure 
// objectID - the ID of the overlay image that is under the cursor
// x,y - the current x,y position of the cursor, ranging from 0.0
// to 1.0
// keyMod - any modifier key which is pressed. These are defined
// in include/CPI_OverlayProvider.h, as
// CPI_CTRL_KEY	 - control key
// CPI_ALT_KEY   - alt key
// CPI_SHIFT_KEY - shift key

static void
CropCircleLeftMouseDown( CPI_PrivateData	handle,
						 CPI_Uint32			objectID,
						 CPI_Float32		x,
						 CPI_Float32		y,
						 CPI_Uint32			keyMod )
{
	overlayState	*state = (overlayState *)handle;

	state->selectedObject	= objectID;

	// compute radius to here, and store difference as error
	// factor for mouse dragging
	state->startRad = getRad( x, y, state->centerX, state->centerY );
	
	// prevent other objects from getting events, for example
	// if you drag this one across another one
	cpiSetSelectLock( CPI_TRUE );	
}


/////////////////////////////////////////
// same as before, but now we are dragging the cursor, so we
// need to update the overlay image with cpiDoRedraw()

static void
CropCircleLeftMouseDrag( CPI_PrivateData	 handle,
						 CPI_Uint32			 objectID,
						 CPI_Float32		 x,
						 CPI_Float32		 y,
						 CPI_Uint32			 keyMod,
						 CPI_Bool			 isContinuous )
{
	overlayState	*state = (overlayState *)handle;

	if ( state->selectedObject == CROSS_ID )
	{
		state->centerX = x;
		state->centerY = y;
	}
	else if ( state->selectedObject == CIRCLE_ID )
	{
		CPI_Float32		newRad, deltaRad;

		newRad = getRad( x, y, state->centerX, state->centerY );
		deltaRad = newRad - state->startRad;
		state->radius += deltaRad;
		state->startRad = newRad;
	}

	if ( isContinuous )
	{
		cpiSetFloat( state->centerX, "center.x", state->myTime, CPI_TRUE );
		cpiSetFloat( state->centerY, "center.y", state->myTime, CPI_TRUE );
		cpiSetFloat( state->radius, "radius", state->myTime, CPI_TRUE );	
	}
	
	cpiDoRedraw();   
}


/////////////////////////////////////////
// unlock the current selection and update current values

static void
CropCircleLeftMouseUp( CPI_PrivateData	 handle, 
					   CPI_Uint32		 objectID,
					   CPI_Float32		 x,
					   CPI_Float32		 y,
					   CPI_Uint32		 keyMod )
{
	overlayState *state = (overlayState *)handle;

	cpiSetSelectLock( CPI_FALSE );

	cpiSetFloat( state->centerX, "center.x", state->myTime, CPI_TRUE );
	cpiSetFloat( state->centerY, "center.y", state->myTime, CPI_TRUE );
	cpiSetFloat( state->radius, "radius", state->myTime, CPI_TRUE );
}


////////////////////////////////////////
// As with nodes and image ops, overlays are defined
// to Rayz with a structure. This structure must be
// present, but the routines it defines are optional.
// Missing routines must be marked with NULL

static RPI_OverlayInfo oinfo = {
	{
		"cropcircle",			// internal name
		"",						// menu name (in Image Viewer menu)
		// the menu name is only used for overlays which are
		// not attached to nodes (see below, and FieldChart.C)
		"Silicon Grail",
		"v1.0"
	},
	NULL,						// Init
	NULL,						// Shutdown
	CropCircleCreateInstance,	// CreateInstance
	CropCircleDestroyInstance,	// DestroyInstance
	CropCircleSourceChanged,	// SourceChanged
	NULL,						// FullSizeChanged
	CropCircleSourceOutOfDate,	// SourceOutOfDate
	CropCircleRender,			// do Render
	CropCircleLeftMouseDown,	// LeftMouseDown
	CropCircleLeftMouseDrag,	// LeftMouseDrag
	CropCircleLeftMouseUp,		// LeftMouseUp
	NULL,						// MouseEnter
	NULL,						// MouseDrag
	NULL,						// MouseExit
	"cropcircle",				// node this is attached to
	CPI_FALSE
	// if this string is NULL, the overlay is not attached to any
	// node, but simply exists globally within every image viewer.
	// In this case, the menu name above will be used to show the
	// overlay in the right mouse menu of each image viewer.
	// See FieldChart.C.
	// It is also possible to attach this overlay to other nodes
	// besides this one - you could attach an overlay to the
	// blur node, if that had some meaning.
};


////////////////////////////////////
// register the node, the image op and the overlay
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	if ( retval )
		retval = cpiRegisterOverlay( &oinfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister everything

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "cropcircle" );
	cpiUnregisterNode( "cropcircle" );
	cpiUnregisterOverlay( "cropcircle" );
}
