/*
//	Demonstrate context and regions - Overlay a mask onto an image
//  This also demonstrates advanced user interface building
*/

#include <stdio.h>
#include <unistd.h>
#include <math.h>

#include <CPI.h>

static const char *chanmenu[] =
{
	"Red",
	"Green",
	"Blue",
	NULL
};

#define REDSCREEN	0
#define GREENSCREEN	1
#define BLUESCREEN	2

// Set up parameters and nested parameters
static CPI_Bool
DespillInit( void )
{
	cpiAddMenu( "back", "Backing Color", 1, chanmenu, NULL, NULL );
	cpiAddFloat( "prop", "Proportion", 0.5, 0.0, 1.0,
				 CPI_PARM_ANIMATABLE | CPI_PARM_CLAMPMIN | CPI_PARM_CLAMPMAX,
				 NULL, NULL );

	return CPI_TRUE;
}


//////////////////////////////////////////
// This is called when the node is actually executed

static CPI_ImageOp
DespillExec( CPI_Float32	 myTime,
			 CPI_Uint8       quality,
			 CPI_Uint32      output,
			 CPI_Bool        viewerOutput,
			 CPI_Float32     scaleX,
			 CPI_Float32     scaleY )
{
	CPI_Float32		pcent;
	CPI_Int32		chan;
	CPI_ImageOp		retval	= NULL;
	CPI_ImageOp		myInputs;

	myInputs = cpiGetInputOp( 0, myTime, quality, scaleX, scaleY );

	if ( cpiGetFloat( &pcent, "prop", myTime ) &&
		 cpiGetInteger( &chan, "back", myTime ) )
	{
		CPI_Metadata    nodeparms = cpiCreateMetadata();

		if ( NULL != nodeparms )
		{
			cpiSetMetaInt32( nodeparms, "channel", chan );
			cpiSetMetaFloat32( nodeparms, "pcent", pcent );

			if ( myInputs != NULL )
				retval = cpiAddImageOp( "despill", nodeparms, &myInputs, 1 );

			cpiDeleteMetadata( nodeparms );
		}
	}
	else
		cpiError( "Unable to retrieve parameters" );

	return retval;
}


////////////////////////////////////////


static CPI_Bool
DespillGetErrors( CPI_Float32 time )
{
	CPI_ImageContext 	 input;
	CPI_Bool			 retval	= CPI_FALSE;

	if ( cpiGetInputContext( &input, 0, time ) )
	{
		if ( 3 <= input.channels )
			retval = CPI_TRUE;
		else
			cpiError( "Input must have at least 3 channels" );
	}
	else
		cpiError( "Couldn't get input context!" );

	return retval;
}


////////////////////////////////////////////
// This structure defines aspects of the node - user subroutines,
// number of inputs, etc

static RPI_NodeInfo ninfo = {
	{
		"despill",				// node name
		"Despill",				// menu name
		"Silicon Grail",		// author
		"v1.0",					// version
		NULL, NULL
	},
	DespillInit,		// NodeInit
	NULL,				// NodeShutDown
	NULL,				// ParmChanged
	NULL,				// InputChanged
	NULL,				// GetRegion
    NULL,				// GetFullSize
	NULL,				// GetRange
	DespillGetErrors,	// GetErrors
	DespillExec,		// Node Execute
	NULL,				// InputLabel
	NULL,				// OutputLabel
	NULL,
	NULL,
	1,					// min inputs
	1,					// max inputs
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable (eg Mcomp)
};


////////////////////////////////////////


template <class T>
class redProcess
{
public:
	inline redProcess( CPI_Float32 pcent )
	{
		myPercent		= pcent;
		myOnePercent	= 1.F - pcent;
	}
	
	inline void	 average( T	&r, T &g, T &b )
	{
		T	 ave = (T)( g * myPercent + b * myOnePercent );

		if ( r > ave )
			r = ave;
	}

private:
	CPI_Float32	 myPercent, myOnePercent;
};


////////////////////////////////////////


template <class T>
class greenProcess
{
public:
	inline greenProcess( CPI_Float32 pcent )
	{
		myPercent		= pcent;
		myOnePercent	= 1.F - pcent;
	}
	
	inline void	 average( T	&r, T &g, T &b )
	{
		T	 ave = (T)( r * myPercent + b * myOnePercent );

		if ( g > ave )
			g = ave;
	}

private:
	CPI_Float32	 myPercent, myOnePercent;
};


////////////////////////////////////////


template <class T>
class blueProcess
{
public:
	inline blueProcess( CPI_Float32 pcent )
	{
		myPercent		= pcent;
		myOnePercent	= 1.F - pcent;
	}
	
	inline void	 average( T	&r, T &g, T &b )
	{
		T	 ave = (T)( r * myPercent + g * myOnePercent );

		if ( b > ave )
			b = ave;
	}

private:
	CPI_Float32	 myPercent, myOnePercent;
};


////////////////////////////////////////


template <class T, class AVE>
class despillProcess
{
public:
	despillProcess( CPI_Float32 pcent )
		: myAve( pcent )
	{
	}

	CPI_Bool	 process( CPI_Image	*input,
						  CPI_Image	*result );

private:
	AVE			 myAve;
};


////////////////////////////////////////


template <class T, class AVE>
CPI_Bool
despillProcess<T,AVE>::process( CPI_Image	*input,
								CPI_Image	*result )
{
	CPI_Int32	 c, x, y, endX, endY;
	CPI_Int32	 chans = result->myContext.channels;
	T			*pin;
	T			*pout;
	T			 red, green, blue;

	endX	= result->myContext.offsetX + result->myContext.sizeX;
	endY	= result->myContext.offsetY + result->myContext.sizeY;

	for ( y = result->myContext.offsetY; y < endY; ++y )
	{
		// Notice that we use cpiGetLine to avoid any sub-image issues
		// as rayz may ask us to process a small part of a larger image
		pin		= (T *)cpiGetLine( input, y );
		pout	= (T *)cpiGetLine( result, y );

		for ( x = result->myContext.offsetX; x < endX; ++x )
		{
			red		= pin[0];
			green	= pin[1];
			blue	= pin[2];

			myAve.average( red, green, blue );

			pout[0] = red;
			pout[1] = green;
			pout[2] = blue;

			// copy each channel
			for ( c = 3; c < chans; ++c )
				pout[c] = pin[c];

			pin		+= chans;
			pout	+= chans;
		}
	}

	return CPI_TRUE;
}


/////////////////////////////////////////
// This is the image operation.


#define BITDEPTHSWITCH( PROCESS )											\
switch ( result->myContext.bitsPerPel )										\
{																			\
	case 8:																	\
	{																		\
		despillProcess<CPI_Uint8,PROCESS<CPI_Uint8> >	 dospill( pcent );	\
		retval = dospill.process( &inputs[0], result );						\
		break;																\
	}																		\
	case 16:																\
	{																		\
		despillProcess<CPI_Uint16,PROCESS<CPI_Uint16> >	 dospill( pcent );	\
		retval = dospill.process( &inputs[0], result );						\
		break;																\
	}																		\
	case 32:																\
	{																		\
		despillProcess<CPI_Float32,PROCESS<CPI_Float32> > dospill( pcent );	\
		retval = dospill.process( &inputs[0], result );						\
		break;																\
	}																		\
	default:																\
		cpiError( "Unknown bit depth" );									\
		break;																\
}
	
static CPI_Bool
DespillProcess( CPI_Image		*result,
				CPI_Image		 inputs[],
				CPI_Uint32		 numInputs,
				CPI_Metadata	 myParms )
{
	CPI_Int32	chan;
	CPI_Float32	pcent;
	CPI_Bool	retval	= CPI_FALSE;

	if ( cpiGetMetaInt32( myParms, "channel", &chan ) &&
		 cpiGetMetaFloat32( myParms, "pcent", &pcent ) )
	{
		// We require 3 or more channels, since we are calculating with
		// red, green, and blue
		if ( result->myContext.channels >= 3 )
		{
			// Sanity check that the images are correct - they should be
			// in rayz, but if we are used somewhere else like scripting, who knows
			if ( result->myContext.channels		== inputs[0].myContext.channels &&
				 result->myContext.bitsPerPel	== inputs[0].myContext.bitsPerPel &&
				 result->myContext.offsetX		== inputs[0].myContext.offsetX &&
				 result->myContext.offsetY		== inputs[0].myContext.offsetY &&
				 result->myContext.sizeX		== inputs[0].myContext.sizeX &&
				 result->myContext.sizeY		== inputs[0].myContext.sizeY )
			{
				// Figure out bit depth and type of despill to do
				if ( REDSCREEN == chan )
				{
					BITDEPTHSWITCH( redProcess );
				}
				else if ( GREENSCREEN == chan )
				{
					BITDEPTHSWITCH( greenProcess );
				}
				else if ( BLUESCREEN == chan )
				{
					BITDEPTHSWITCH( greenProcess );
				}
				else
					cpiError( "Unknown channel selection for the despill backing color" );
			}
			else
				cpiError( "Input and Result images differ" );
		}
		else
			cpiError( "Unable to process images with fewer than 3 channels" );
	}
	else
		cpiError( "Unable to retrieve correct metadata" );

	return CPI_TRUE;
}


/////////////////////////////////////
// Structure which defines the image op

static RPI_ImageOpInfo opinfo =
{
	{
		"despill",			// command name
		"",			        // not used
		"Silicon Grail",	// author
		"v1.0",				// version
		NULL, NULL
	},
	DespillProcess,			// image execution process
	NULL,					// InputRegion
	NULL,					// OutputRegion
	NULL,
	NULL,
	NULL,
	1,						// min number of inputs required
	CPI_TRUE				// Can this plugin multiproc?
};


CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;
}


CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "despill" );
	cpiUnregisterNode( "despill" );
}
