/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	EarthQuake
//
// COMMENTS: Does camera shake. Shows how to do this by not
//			 writing any image code, but instead just using
//			 the built-in Rayz transform image op.
//
*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <CPI.h>

////////////////////////////////////////////
// The init routine is called once, and typically is used to
// define the user input controls

static CPI_Bool
QuakeInit( void )
{
	cpiAddFloat( "roll", "Max Roll", 1.0F, 0.0F, 45.0F, 
				 CPI_PARM_ANIMATABLE | CPI_PARM_CLAMPMIN, 
				 "Maximum Z rotation value in degrees", NULL);
	cpiAddFloatPair( "shift", "Max Shift", 
					 "x", "X", 5.0F, "y", "Y", 5.0F, 
					 CPI_PARM_ANIMATABLE,
					 "Maximum X & Y displacement in pixels", NULL );
	cpiAddFloat( "speed", "Speed", 1.0, 0.0, 10.0, 
				 CPI_PARM_ANIMATABLE, 
				 "Rate of change of movement", NULL );
	cpiAddInteger( "seed", "Seed", 1, 0, 1000, 
				   CPI_PARM_ANIMATABLE, 
				   "Value to control randomness", NULL );

	return CPI_TRUE;
}


/////////////////////////////////////////////
// shutdown does nothing in this case

static CPI_Bool
QuakeShutDown( void )
{
	return CPI_TRUE;
}


//////////////////////////////////////////
// return a random number between -1 & +1 with a
// gaussian distribution
// It would be even better to use fractal noise here...

inline CPI_Float32
fastRandomFloat( CPI_Uint32	&seed )
{
	CPI_Uint32 tmp;

	seed = seed*1664525 + 1013904223;
	tmp = 0x3f800000 | ( 0x007fffff & seed );
	return (*(CPI_Float32 *)&tmp) - 1.0F;
}


////////////////////////////////////////


static CPI_Float32
newrand( CPI_Uint32 &seed )
{
	static CPI_Uint32  iset = 0;
	static CPI_Float32 gset = 0.0F;
	CPI_Float32		   fac, rsq, v1, v2, retval;

	if( iset == 0 )
	{
		// Pick 2 numbers in a unit circle
		rsq = 2.0F;
		while( rsq >= 1.0F )
		{
			v1 = 2.0F * fastRandomFloat( seed ) - 1.0F;
			v2 = 2.0F * fastRandomFloat( seed ) - 1.0F;
			rsq = v1 * v1 + v2 * v2;
		}

		fac = sqrt( -2.0F * log( rsq ) / rsq );

		// Make the Box-Muller transformation to get 2
		// normal deviates.  Return one and save the other for next time.
		gset = v1 * fac;
		iset = 1;

		retval = v2 * fac;
	}
	else
	{
		iset	= 0;
		retval	= gset;
	}

	return retval;
}


////////////////////////////////////////


////////////////////////////////////
// This is called when an image is required
// The basic structure of these routines is:
// 1. get the inputs
// 2. get the parameters from the user interface
// 3. allocate metadata and copy the parameters to it
// 4. add the image op(s) to the list of operations

static CPI_ImageOp
QuakeExec( CPI_Float32	 time,
		   CPI_Uint8	 quality,
		   CPI_Uint32	 output,
		   CPI_Bool 	 viewerOutput,
		   CPI_Float32	 scaleX,
		   CPI_Float32	 scaleY )
{
	CPI_ImageOp			retval = NULL;
	CPI_ImageOp			myInput;
	CPI_Float32			maxZrot, maxX, maxY, speed;
	CPI_Float32			xoff, yoff, roll, px, py;
	CPI_Int32			sizeX, sizeY;
	CPI_Int32			seed;

	// get the input image
	myInput = cpiGetInputOp( 0, time, quality, scaleX, scaleY );

	// get the parameters from the user
	if ( cpiGetFloat( &maxZrot, "roll", time ) &&  
		 cpiGetFloat( &maxX, "shift.x", time ) &&
		 cpiGetFloat( &maxY, "shift.y", time ) && 
		 cpiGetFloat( &speed, "speed", time ) && 
		 cpiGetInteger( &seed, "seed", time ) )
	{
		CPI_Uint32	realSeed  = (CPI_Uint32)seed;
	
		// compute the transform values
		roll = maxZrot * newrand( realSeed ) * speed;
		xoff = maxX * newrand( realSeed ) * speed;
		yoff = maxY * newrand( realSeed ) * speed;

		// get the input size - pivot and translates need
		// to be given in pixel values
		if ( cpiGetInputFullSize( &sizeX, &sizeY, 0, time ) )
		{
			px = CPI_Float32(sizeX) / 2.0F;
			py = CPI_Float32(sizeY) / 2.0F;

			// set up the metadata
			// metadata variable names are defined in the
			// accompanying html docs
			CPI_Metadata theParms = cpiCreateMetadata();

			cpiSetMetaFloat32( theParms, "movex", xoff );
			cpiSetMetaFloat32( theParms, "movey", yoff );
			cpiSetMetaFloat32( theParms, "rotatez", roll );
			cpiSetMetaFloat32( theParms, "pivotx", px);
			cpiSetMetaFloat32( theParms, "pivoty", py );
			cpiSetMetaString( theParms, "order", "RTS" );

			// add the 'transform' op to the list. This is an internal
			// Rayz image op
			if ( NULL != myInput )
				retval = cpiAddImageOp( "transform", theParms, &myInput, 1 );

			cpiDeleteMetadata( theParms );
		}
		else
			cpiError( "Couldn't get full size from input A" );
	}
	else
		cpiError( "Couldn't get EarthQuake parameter(s)" );

	return retval;
}


//////////////////////////////////////////
// This structure, which is required, defines to Rayz
// the various routines which are part of this plugin.
// Routines which are not required are set to NULL

static RPI_NodeInfo ninfo = {
	{
		"earthquake",	// node name
		"EarthQuake",	// menu name
		"Silicon Grail",
		"1.0", NULL, NULL
	},
	
	QuakeInit,		// Node Init
	QuakeShutDown,	// ShutDown
	NULL,			// ParmChanged
	NULL,			// InputChanged
	NULL,			// QuakeGetRegion,
	NULL,			// QuakeGetFullSize,
	NULL,			// QuakeGetRange,
	NULL,			// QuakeGetErrors,
	QuakeExec,		// Execute
	NULL,			// InputLabel
	NULL,			// OutputLabel
	NULL,			// ViewerLabel
	NULL,			// ViewerOutputs
	1,				// min inputs
	1,				// max inputs
	1,				// min outputs
	1,				// max outputs
	CPI_FALSE		// growable?
};
	

// And that's it - there is no Image Op to define, because we
// are using a Rayz internal op. Just register the node and
// we're done.


/////////////////////////////////////////////////
// register the node
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	return cpiRegisterNode( &ninfo );
}


//////////////////////////////////////////////
// unregister the node

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterNode( "EarthQuake" );
}
