/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:    Enable
//
// COMMENTS: Demonstrate enabling and disabling of parameters
//           The input is just copied to the output
//
*/

#include <string.h>
#include <stdio.h>
#include <math.h>
#include <CPI.h>


/////////////////////////////////////////
// Define the interface controls

static const char *unitmenu[]=
{
	"pixels",
	"fractions",
	NULL
};

static CPI_Bool
EnableInit( void )
{
	cpiAddMenu( "units", "Units", 0, unitmenu,
				"Choose which units to enter values in", NULL );

	cpiAddIntPair( "pixels", "Pixels",
				   "x", "Width", 10, "y", "Height", 10,
				   CPI_PARM_DEFAULT, NULL, NULL );

	cpiAddFloatPair( "fractions", "Fractions",
			"x", "Width", 0.1, "y", "Height", 0.1,
			CPI_PARM_DEFAULT, NULL, NULL );

	return CPI_TRUE;
}

//////////////////////////////////////////
// This is called whenever a parameter is changed in the GUI.
// It comes in with the internal name of the parameter.

static void
EnableParmChanged( const char *name )
{
	CPI_Int32	units, pval;
	CPI_Float32	fval;

	// this is the situation when the node is initialized
	if ( name == NULL )
	{
		cpiEnableParameter( "pixels", CPI_TRUE );
		cpiEnableParameter( "fractions", CPI_FALSE );
		return;
	}

	// if the relevant param is changed, pick up its new value
	// and enable/disable based on that
	if ( strcmp( name, "units" ) == 0 &&
		 cpiGetInteger( &units, "units", 1.0F ) )
	{
		if ( units == 0 )
		{
			cpiEnableParameter( "pixels", CPI_TRUE );
			cpiEnableParameter( "fractions", CPI_FALSE );
		}
		else
		{
			cpiEnableParameter( "pixels", CPI_FALSE );
			cpiEnableParameter( "fractions", CPI_TRUE );
		}
	}
	// if a value is changed, update the other set to stay
	// in sync
	// These are based on the arbitrary relationship that
	// fractions = pixels / 100
	//       and so
	// pixels = fractions * 100
	else if ( strcmp( name, "pixels.x" ) == 0 &&
			  cpiGetInteger( &pval, "pixels.x", 0.0F ) )
	{
		fval = pval / 100.0F;
		cpiSetFloat( fval, "fractions.x", 0.0F, CPI_FALSE );
	}
	else if ( strcmp( name, "pixels.y" ) == 0 &&
			  cpiGetInteger( &pval, "pixels.y", 0.0F ) )
	{
		fval = pval / 100.0F;
		cpiSetFloat( fval, "fractions.y", 0.0F, CPI_FALSE );
	}
	else if ( strcmp( name, "fractions.x" ) == 0 &&
			  cpiGetFloat( &fval, "fractions.x", 0.0F ))
	{
		pval = cpiRound( fval * 100.0F );
		cpiSetInteger( pval, "pixels.x", 0.0F, CPI_FALSE );
	}
	else if ( strcmp( name, "fractions.y" ) == 0 &&
			  cpiGetFloat( &fval, "fractions.y", 0.0F ) )
	{
		pval = cpiRound( fval * 100.0F );
		cpiSetInteger( pval, "pixels.y", 0.0F, CPI_FALSE );
	}
}

//////////////////////////////////////////
// This is called when the node is actually executed
// This gets the input image, and then adds the image
// operation to the list of ops to be executed

static CPI_ImageOp
EnableExec( CPI_Float32		myTime,
			CPI_Uint8		quality,
			CPI_Uint32		nodeOutput,
			CPI_Bool	 	viewerOutput,
			CPI_Float32		scaleX,
			CPI_Float32   	scaleY )

{
	CPI_ImageOp	retval	= NULL;
	CPI_ImageOp myInputs;
	
	myInputs = cpiGetInputOp( 0, myTime, quality, scaleX, scaleY );

	if ( myInputs != NULL )
	   	retval = cpiAddImageOp( "enable", NULL, &myInputs, 1 );

	return retval;
}


////////////////////////////////////////////
// This structure, which is required, defines to Rayz
// the various routines which are part of this plugin.
// Routines which are not required are set to NULL

static RPI_NodeInfo ninfo = {
	{
		"enable",				// node name
		"Enable",				// menu name
		"Joe Programmer",		// author
		"v1.0"					// version
	},
	EnableInit,			// NodeInit
	NULL,				// NodeShutDown
	EnableParmChanged,	// ParmChanged
	NULL,				// InputChanged
	NULL,				// GetRegion
    NULL,				// GetFullSize
	NULL,				// GetRange
	NULL,				// GetErrors
	EnableExec,			// Node Execute
	NULL,				// InputLabel
	NULL,				// OutputLabel
	NULL,				// ViewerLabel
	NULL,               // ViewerOutputs
	1, 					// min inputs
	1,					// max inputs
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable (eg Mcomp)
};



//////////////////////////////////////////////////////////////////////
// Image Op Functions ////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

// This is the actual image manipulation routine. It is a templated
// function so that we can write it once, and let the compiler build
// a version for each of 8, 16 and float pixel types.
// This just copies input to result

template <class T> static
void simple( CPI_Image	&input,
			 CPI_Image	*result, 
			 T 			 max)
{
	CPI_Int32	 x, y, endX, endY, i, j = 0;
	CPI_Int32	 chans = result->myContext.channels;
	T			*pin, *pout;

	endX	= result->myContext.offsetX + result->myContext.sizeX;
	endY	= result->myContext.offsetY + result->myContext.sizeY;
	
	for ( y = result->myContext.offsetY; y < endY; y++ )
	{
		// Notice that we use cpiGetLine to avoid any sub-image issues
		// as rayz may ask us to process a small part of a larger image
		pin     = (T *)cpiGetLine( &input, y );
		pout    = (T *)cpiGetLine( result, y );
		j 		= 0;

		for ( x = result->myContext.offsetX; x < endX; x++ )
		{
			// copy each channel
			for ( i = 0; i < chans; i++, j++ )
				pout[j] = pin[j];
		}
	}
}


/////////////////////////////////////////
// This is called to execute the image op
// It just calls the appropriate copy routine, 
// based on the input bit depth

static CPI_Bool
EnableProcess( CPI_Image	*result,
			   CPI_Image	 inputs[],
			   CPI_Uint32	 numInputs,
			   CPI_Metadata  myParms )
{
	if ( result->myContext.bitsPerPel == 8 )
		simple( inputs[0], result, (CPI_Uint8)255 );
	else if ( result->myContext.bitsPerPel == 16 )
		simple( inputs[0], result, (CPI_Uint16)65535 );
	else
		simple( inputs[0], result, (CPI_Float32)1.0F );

		
	return CPI_TRUE;
}


/////////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM

static RPI_ImageOpInfo opinfo = 
{
	{
		"enable",			// command name
		"",			        // not used
		"Joe Programmer",	// author
		"0.1"				// version
	},
	EnableProcess,			// image execution process
	NULL,					// InputRegion
	NULL,					// OutputRegion,
	NULL,                   // ControlProcessing
	NULL,                   // SetupProcessing
	NULL,                   // CleanupProcessing
	1,						// min number of inputs required
	CPI_TRUE,				// Can this plugin multiproc?
	CPI_FALSE				// Disjoint regions?
};


////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister node and image op

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "enable" );
	cpiUnregisterNode( "enable" );
}
