/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:    Errors
//
// COMMENTS: Demonstrates generating user errors
//
*/

/* 
   This defines a GetErrors routine. This routine, if defined, is
   called before any processing begins, so the programmer can 
   validate inputs, values, etc.

   It is also possible to generate errors at runtime from anywhere
   else in the plugin.

   In this example, we check for two error conditions. The first is
   based on the user incorrectly connecting the node. In this example,
   we want to disallow floating point inputs, but more commonly you
   might require an alpha channel, or 3 inputs, or something like that.

   The second error is based on an incorrect value received from the
   GUI - to illustrate that, we have a menu choice that simply
   says error/no error
*/

#include <stdio.h>
#include <CPI.h>


/////////////////////////////////////////
// menu of actions 
static const char *errormenu[]=
{
	"Error",
	"No Error",
	NULL
};

////////////////////////////////////////
// The init routine is called once, and typically is used to
// define the user input controls

static CPI_Bool
ErrorsInit( void )
{
	cpiAddDisplayString("a", " ", "This node errors on float image input ");
	cpiAddMenu( "menu", "ErrorTest", 1, errormenu, NULL, NULL );

	return CPI_TRUE;
}

//////////////////////////////////////////
// This is called before processing begins. It checks to see if
// an error was requested from the menu, and if so generates one
// using cpiError()
// It also checks to see if the input image is floating point -
// if so, it generates an error.

static CPI_Bool
ErrorsGetErrors( CPI_Float32	time )
{
	CPI_Int32			errortest;
	CPI_Bool			retval = CPI_TRUE;
	CPI_ImageContext	inputContext;

	// get menu choice
	if ( cpiGetInteger( &errortest, "menu", time ) )
	{
		if ( errortest == 0 )
		{
			cpiError( "We have an error!" );
			retval = CPI_FALSE;
		}
		else
		{
			// get input image context
			cpiGetInputContext( &inputContext, 0, time );

			if ( inputContext.bitsPerPel == 32 )
			{
				cpiError( "We want to error on float inputs!" );
				retval = CPI_FALSE;
			}
		}
	}
	else
	{
		cpiError( "Couldn't get Errors Menu value" );
		retval = CPI_FALSE;
	}

	return retval;
}


//////////////////////////////////////////
// This is called when an image is required
// The basic structure of these routines is:
// 1. get the inputs
// 2. get the parameters from the user interface
// 3. allocate metadata and copy the parameters to it
// 4. add the image op(s) to the list of operations
// In this case, we have no parameters to worry about

static CPI_ImageOp
ErrorsExec( CPI_Float32		myTime,
			CPI_Uint8		quality,
			CPI_Uint32		nodeOutput,
			CPI_Bool	 	viewerOutput,
			CPI_Float32		scaleX,
			CPI_Float32   	scaleY )

{
	CPI_ImageOp	retval	= NULL;
	CPI_ImageOp myInputs;
	
	myInputs = cpiGetInputOp( 0, myTime, quality, scaleX, scaleY );

	if ( myInputs != NULL )
	   	retval = cpiAddImageOp( "errors", NULL, &myInputs, 1 );

	return retval;
}


////////////////////////////////////////////
// This structure defines aspects of the node - user subroutines,
// number of inputs, etc

static RPI_NodeInfo ninfo = {
	{
		"errors",				// node name
		"Errors",				// menu name
		"Joe Programmer",		// author
		"v1.0"					// version
	},
	ErrorsInit,			// NodeInit
	NULL,				// NodeShutDown
	NULL,				// ParmChanged
	NULL,				// InputChanged
	NULL,				// GetRegion
    NULL,				// GetFullSize
	NULL,				// GetRange
	ErrorsGetErrors,	// GetErrors
	ErrorsExec,			// Node Execute
	NULL,				// InputLabel
	NULL,				// OutputLabel
	NULL,				// ViewerLabel
	NULL,				// ViewerOutputs
	1, 					// min inputs
	1,					// max inputs
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable (eg Mcomp)
};



//////////////////////////////////////////////////////////////////////
// Image Op Functions ////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////
// This templated function does the actual work. By making
// it a template, it is only defined once, and the compiler
// creates a version for each of 8 bit, 16 and float inputs.

template <class T> static
void copy( CPI_Image	&input,
			 CPI_Image	*result, 
			 T 			 max)
{
	CPI_Int32	 x, y, endX, endY, i, j = 0;
	CPI_Int32	 chans = result->myContext.channels;
	T			*pin, *pout;

	endX	= result->myContext.offsetX + result->myContext.sizeX;
	endY	= result->myContext.offsetY + result->myContext.sizeY;
	
	// this routine just copies its input to its output
	for ( y = result->myContext.offsetY; y < endY; y++ )
	{
		// Notice that we use cpiGetLine to avoid any sub-image issues
		// as rayz may ask us to process a small part of a larger image
		pin		= (T *)cpiGetLine( &input, y );
		pout	= (T *)cpiGetLine( result, y );
		j		= 0;

		for ( x = result->myContext.offsetX; x < endX; x++ )
		{
			// copy each channel
			for ( i = 0; i < chans; i++, j++ )
				pout[j] = pin[j];
		}
	}
}


/////////////////////////////////////////
// This is called to execute the image op
// The basic structure of these routines is
// 1. unload parameter values from metadata
// 2. call the appropriate routine depending on bit depth

static CPI_Bool
ErrorsProcess( CPI_Image	*result,
			  CPI_Image	 	 inputs[],
			  CPI_Uint32	 numInputs,
			  CPI_Metadata   myParms )
{
	if ( result->myContext.bitsPerPel == 8 )
		copy( inputs[0], result, (CPI_Uint8)255 );
	else if ( result->myContext.bitsPerPel == 16 )
		copy( inputs[0], result, (CPI_Uint16)65535 );
	else
		copy( inputs[0], result, (CPI_Float32)1.0F );

		
	return CPI_TRUE;
}


/////////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM.

static RPI_ImageOpInfo opinfo = 
{
	{
		"errors",			// command name
		"",			        // not used
		"Joe Programmer",	// author
		"0.1"				// version
	},
	ErrorsProcess,			// image execution process
	NULL,					// InputRegion
	NULL,					// OutputRegion,
	NULL,					// control processing
	NULL,					// setup processing
	NULL,					// CleanupProcessing
	1,						// min number of inputs required
	CPI_TRUE,				// Can this plugin multiproc?
	CPI_FALSE				// Supports Disjoint regions?
};


////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister node and image op

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "errors" );
	cpiUnregisterNode( "errors" );
}
