/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	Grid
//
// COMMENTS: This illustrates a source node (one with no inputs)
//
*/

#include <stdio.h>
#include <CPI.h>

//////////////////////////////////////
// menu of which channels to create

static const char *channelmenu[]= 
{
	"RGB Alpha",
	"RGB",
	"Alpha only",
	NULL
};


///////////////////////////////////////
// menu of which bit depth to create

static const char *bitdepthmenu[] = 
{
	"8 bit",
	"16 bit",
	"Float",
	NULL
};


///////////////////////////////////////
// default color arrays

static CPI_Float32 lines[] = { 1.0F, 1.0F, 1.0F, 1.0F };
static CPI_Float32 bgcol[] = { 0.0F, 0.0F, 0.0F, 0.0F };


/////////////////////////////////////////////////////
// Node Functions ///////////////////////////////////
/////////////////////////////////////////////////////

// The init routine is called once, and typically is used to
// define the user input controls

static CPI_Bool
GridInit( void )
{
	cpiAddColorGroup( "line", "Line Color", lines, 4, CPI_PARM_DEFAULT, 
						"Color of squares", NULL );
	cpiAddColorGroup( "bg", "Background", bgcol, 4, CPI_PARM_DEFAULT, 
						"Color between squares", NULL );
	cpiAddIntPair( "size", "Size",
					"w", "Width", 720,
					"h", "Height", 576,
					 CPI_PARM_DEFAULT,
					"Size of output image", NULL );
	cpiAddInteger( "linewid", "Line Width",
					5, 1, 50, CPI_PARM_CLAMPMINMAX,
					"Width of each line in pixels", NULL );
	cpiAddInteger( "linegap", "Line Gap",
					25, 1, 150, CPI_PARM_CLAMPMINMAX,
					"Width of each line in pixels", NULL );

	cpiAddMenu( "bits", "Bit Depth", 0, bitdepthmenu, NULL, NULL );
	cpiAddMenu( "chans", "Channels", 0, channelmenu, NULL, NULL );
	cpiAddSequenceRange( 1, 100,
						CPI_PARM_DEFAULT,
						"Range of frames to output", NULL );

	return CPI_TRUE;
}

///////////////////////////////////////////
// shutdown does nothing in this case

static CPI_Bool
GridShutDown( void )
{
	return CPI_TRUE;
}


///////////////////////////////////////////
// GetRegion needs to be defined for nodes that generate their
// own images (as opposed to being given images to modify)
// This defines for the rest of Rayz what the output 'context'
// is, ie, size, bit depth, etc.

static CPI_Bool
GridGetRegion( CPI_ImageContext	*output,
			   CPI_Float32		 myTime,
			   CPI_Uint32		 nodeOutput,
			   CPI_Bool		 	 viewerOutput )
{
	CPI_Int32	bitsPerPel, channels;
	CPI_Bool	retval = CPI_TRUE;

	// get size from GUI and store in 'output'
	if ( cpiGetInteger( &output->sizeX, "size.w", myTime ) &&
		 cpiGetInteger( &output->sizeY, "size.h", myTime ) &&
		 cpiGetInteger( &bitsPerPel, "bits", myTime ) &&
		 cpiGetInteger( &channels, "chans", myTime ) )
	{
		output->offsetX = 0;
		output->offsetY = 0;

		// get user menu choice, and set value accordingly
		switch ( bitsPerPel )
		{
			case 0:
				output->bitsPerPel = 8;
			break;
			case 1:
				output->bitsPerPel = 16;
				break;
			case 2:
				output->bitsPerPel = 32;	// ie, float
				break;
			default:
				cpiError( "Invalid bitsPerPel in getRegion!\n" );
				retval = CPI_FALSE;
				break;
		}

		// get user menu choice, and set value accordingly
		switch ( channels )
		{
			case 0:
				output->channels = 4;
				break;
			case 1:
				output->channels = 3;
				break;
			case 2:
				output->channels = 1;
				break;
			default:
				cpiError( "Invalid channels in getRegion!\n" );
				retval = CPI_FALSE;
				break;
		}
	}
	else
	{
		cpiError( "Couldn't get Grid parameter(s)" );
		retval = CPI_FALSE;
	}

	return retval;
}


///////////////////////////////////////////
// Rayz may need to ask the plugin what size
// image it intends to create - define that here
// Nodes that create their own output need to
// define this routine

static CPI_Bool
GridGetFullSize( CPI_ImageContext *fullsize,
					CPI_Float32		  myTime,
					CPI_Uint32		  nodeOutput,
					CPI_Bool          viewerOutput )
{
	CPI_Bool retval = CPI_TRUE;

	retval = ( cpiGetInteger( &fullsize->sizeX, "size.w", myTime ) &&
			   cpiGetInteger( &fullsize->sizeY, "size.h", myTime ) );

	return retval;
}


//////////////////////////////////////////
// This is called when an image is required
// The basic structure of these routines is:
// 1. get the parameters from the user interface
// 2. allocate metadata and copy the parameters to it
// 3. add the image op(s) to the list of operations

static CPI_ImageOp
GridExec( CPI_Float32 	myTime,
		  CPI_Uint8     quality,
		  CPI_Uint32    nodeOutput,
		  CPI_Bool     	viewerOutput,
		  CPI_Float32   scaleX,
		  CPI_Float32   scaleY )
{
	CPI_Int32	sizeX, sizeY, channels, bitdepth, linewid;
	CPI_Int32	lineX, lineY, gapX, gapY, linegap;
	CPI_Float32	bgred, bggreen, bgblue, bgalpha;
	CPI_Float32 linered, linegreen, lineblue, linealpha;
	CPI_ImageOp	retval	= NULL;
	   
	if ( cpiGetInteger( &sizeX, "size.w", myTime ) &&
		 cpiGetInteger( &sizeY, "size.h", myTime ) &&
		 cpiGetInteger( &channels, "chans", myTime ) &&
		 cpiGetInteger( &bitdepth, "bits", myTime ) &&
		 cpiGetInteger( &linewid, "linewid", myTime ) &&
		 cpiGetInteger( &linegap, "linegap", myTime ) &&
		 cpiGetFloat( &bgred, "bg.red", myTime ) &&
		 cpiGetFloat( &bggreen, "bg.green", myTime ) &&
		 cpiGetFloat( &bgblue, "bg.blue", myTime ) &&
		 cpiGetFloat( &bgalpha, "bg.alpha", myTime ) &&
		 cpiGetFloat( &linered, "line.red", myTime ) &&
		 cpiGetFloat( &linegreen, "line.green", myTime ) &&
		 cpiGetFloat( &lineblue, "line.blue", myTime ) &&
		 cpiGetFloat( &linealpha, "line.alpha", myTime ) )
	{
		CPI_Metadata	rampparms = cpiCreateMetadata();

		cpiSetMetaInt32( rampparms, "sizex", sizeX );
		cpiSetMetaInt32( rampparms, "sizey", sizeY );

		switch ( channels )
		{
			case 0:
				cpiSetMetaInt32( rampparms, "channels", 4 );
				break;
			case 1:
				cpiSetMetaInt32( rampparms, "channels", 3 );
				break;
			case 2:
				cpiSetMetaInt32( rampparms, "channels", 1 );
				break;
		}
		switch ( bitdepth )
		{	
			case 0:
				cpiSetMetaInt32( rampparms, "bitdepth", 8 );
				break;
			case 1:
				cpiSetMetaInt32( rampparms, "bitdepth", 16 );
				break;
			case 2:
				cpiSetMetaInt32( rampparms, "bitdepth", 32 );
				break;
		}

		// Resize relevant parameters based on scale.
		// This happens when the user requests a "medium" or "low"
		// image, and when your node shows up in the Clip Editor -
		// the little thumbnails need to be scaled accordingly
		// Your node needs to
		// correctly output something of whatever requested scale.
		lineX = cpiRound( linewid * scaleX );
		lineY = cpiRound( linewid * scaleY );
		if ( lineX < 1 )
			lineX = 1;
		if ( lineY < 1 )
			lineY = 1;

		cpiSetMetaInt32( rampparms, "linewidX", lineX );
		cpiSetMetaInt32( rampparms, "linewidY", lineY );

		gapX = cpiRound( linegap * scaleX );
		gapY = cpiRound( linegap * scaleY );
		if ( gapX < 1 )
			gapX = 1;
		if ( gapY < 1 )
			gapY = 1;

		cpiSetMetaInt32( rampparms, "linegapX", gapX );
		cpiSetMetaInt32( rampparms, "linegapY", gapY );

		// pass the rest of the data
		cpiSetMetaFloat32( rampparms, "bgred", bgred );
		cpiSetMetaFloat32( rampparms, "bggreen", bggreen );
		cpiSetMetaFloat32( rampparms, "bgblue", bgblue );
		cpiSetMetaFloat32( rampparms, "bgalpha", bgalpha );
		cpiSetMetaFloat32( rampparms, "linered", linered );
		cpiSetMetaFloat32( rampparms, "linegreen", linegreen );
		cpiSetMetaFloat32( rampparms, "lineblue", lineblue );
		cpiSetMetaFloat32( rampparms, "linealpha", linealpha );

		retval = cpiAddImageOp( "grid", rampparms, NULL, 0 );

		cpiDeleteMetadata( rampparms );
	}
	else
		cpiError( "Couldn't get Grid parameter(s)" );
	
	return retval;
}


////////////////////////////////////////////

static RPI_NodeInfo ninfo = {
	{
		"grid",			// node name
		"Grid",			// menu name
		"Silicon Grail",	// author
		"v1.0"				// version
	},
	GridInit,		// NodeInit
	GridShutDown,	// NodeShutDown
	NULL,				// ParmChanged
	NULL,				// InputChanged
	GridGetRegion,	// GetRegion
    GridGetFullSize,	// GetFullSize
	NULL,				// GetRange
	NULL,				// GetErrors
	GridExec,		// Node Execute
	NULL,				// InputLabel
	NULL,				// OutputLabel
	NULL,				// ViewerLabel
	NULL,				// ViewerOutputs
	0, 					// min input
	0,					// max input
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable
};


////////////////////////////////////////////////////
// Image Op Functions //////////////////////////////
////////////////////////////////////////////////////


////////////////////////////////////////
// This templated function does the actual work. By making
// it a template, it is only defined once, and the compiler
// creates a version for each of 8 bit, 16 and float inputs.

// To gain some efficiency, define a separate routine for each type
// of image, (that is 1, 3 or 4 channels) Then use templates to
// generate code for 8, 16 and 32 bit versions of each

template <class T> static
void grid1chan( CPI_Image	*image, 
				CPI_Int32	 lineWid[], 
				CPI_Int32	 lineGap[], 
				T 			 max )
{
	CPI_Int32	 x, y, endX, endY, a, b, xwidth, ywidth;
	T			*ptr;

	endX	= image->myContext.offsetX + image->myContext.sizeX;
	endY	= image->myContext.offsetY + image->myContext.sizeY;
	
	xwidth = lineWid[0] + lineGap[0];
	ywidth = lineWid[1] + lineGap[1];
	// check each scanline and pixel to see what color it is
	for ( y = image->myContext.offsetY; y < endY; y++ )
	{
		a = ( (endY - y - 1) % ywidth);
		ptr = (T *)cpiGetLine( image, y );
		for ( x = image->myContext.offsetX; x < endX; x++ )
		{
			b = x % xwidth;
			if ( a >= lineGap[1] || b >= lineGap[0] )
				ptr[0] = max;
			else
				ptr[0] = (T)0;
			ptr++;
		}
	}

}

template <class T> static
void grid3chan( CPI_Image	*image, 
				CPI_Int32	 lineWid[], 
				CPI_Int32	 lineGap[], 
				CPI_Float32	 lvals[], 
				CPI_Float32	 dvals[],
				T 			 max )
{
	CPI_Int32	 x, y;
	CPI_Int32	 endX, endY, a, b, xwidth, ywidth;
	T			*ptr, bgR, bgG, bgB, lineR, lineG, lineB;

	// init colors
	bgR   = (T) (dvals[0] * max);
	bgG   = (T) (dvals[1] * max);
	bgB   = (T) (dvals[2] * max);
	lineR = (T) (lvals[0] * max);
	lineG = (T) (lvals[1] * max);
	lineB = (T) (lvals[2] * max);

	endX	= image->myContext.offsetX + image->myContext.sizeX;
	endY	= image->myContext.offsetY + image->myContext.sizeY;
	
	xwidth = lineWid[0] + lineGap[0];
	ywidth = lineWid[1] + lineGap[1];
	// check each scanline and pixel to see what color it is
	for ( y = image->myContext.offsetY; y < endY; y++ )
	{
		a = ( (endY - y - 1) % ywidth);
		ptr = (T *)cpiGetLine( image, y );
		for ( x = image->myContext.offsetX; x < endX; x++ )
		{
			b = x % xwidth;
			if ( a >= lineGap[1] || b >= lineGap[0] )
			{
					ptr[0] = lineR;
					ptr[1] = lineG;
					ptr[2] = lineB;
			}
			else
			{
					ptr[0] = bgR;
					ptr[1] = bgG;
					ptr[2] = bgB;
			}
			ptr += 3;
		}
	}

}

template <class T> static
void grid4chan( CPI_Image	*image, 
				CPI_Int32	 lineWid[], 
				CPI_Int32	 lineGap[], 
				CPI_Float32	 lvals[], 
				CPI_Float32	 dvals[],
				T 			 max )
{
	CPI_Int32	 x, y;
	CPI_Int32	 endX, endY, a, b, xwidth, ywidth;
	T			*ptr, bgR, bgG, bgB, lineR, lineG, lineB;

	// init colors
	bgR   = (T) (dvals[0] * max);
	bgG   = (T) (dvals[1] * max);
	bgB   = (T) (dvals[2] * max);
	lineR = (T) (lvals[0] * max);
	lineG = (T) (lvals[1] * max);
	lineB = (T) (lvals[2] * max);

	endX	= image->myContext.offsetX + image->myContext.sizeX;
	endY	= image->myContext.offsetY + image->myContext.sizeY;
	
	xwidth = lineWid[0] + lineGap[0];
	ywidth = lineWid[1] + lineGap[1];
	// check each scanline and pixel to see what color it is
	for ( y = image->myContext.offsetY; y < endY; y++ )
	{
		a = ( (endY - y - 1) % ywidth);
		ptr = (T *)cpiGetLine( image, y );
		for ( x = image->myContext.offsetX; x < endX; x++ )
		{
			b = x % xwidth;
			if ( a >= lineGap[1] || b >= lineGap[0] )
			{
					ptr[0] = lineR;
					ptr[1] = lineG;
					ptr[2] = lineB;
					ptr[3] = max;
			}
			else
			{
					ptr[0] = bgR;
					ptr[1] = bgG;
					ptr[2] = bgB;
					ptr[3] = (T)0;
			}
			ptr += 4;
		}
	}

}

/////////////////////////////////////////
// This is called to execute the image op
// The basic structure of these routines is
// 1. unload parameter values from metadata
// 2. call the appropriate routine depending on
//    bit depth

static CPI_Bool
GridProcess( CPI_Image	*result,
				CPI_Image	 inputs[],
				CPI_Uint32	 numInputs,
				CPI_Metadata myParms )
{
	CPI_Int32	 linewid[2], linegap[2];
	CPI_Float32  linecols[4];
	CPI_Float32  bgcols[4];
	CPI_Bool	 retval = CPI_FALSE;
	
	if ( cpiGetMetaInt32( myParms, "linewidX", &linewid[0] ) &&
		 cpiGetMetaInt32( myParms, "linewidY", &linewid[1] ) &&
		 cpiGetMetaInt32( myParms, "linegapX", &linegap[0] ) &&
		 cpiGetMetaInt32( myParms, "linegapY", &linegap[1] ) &&
		 cpiGetMetaFloat32( myParms, "linered", &linecols[0] ) &&
		 cpiGetMetaFloat32( myParms, "linegreen", &linecols[1] ) &&	
		 cpiGetMetaFloat32( myParms,	"lineblue", &linecols[2] ) &&
		 cpiGetMetaFloat32( myParms, "linealpha", &linecols[3] ) &&
		 cpiGetMetaFloat32( myParms, "bgred", &bgcols[0] ) &&
		 cpiGetMetaFloat32( myParms, "bggreen", &bgcols[1] ) &&	
		 cpiGetMetaFloat32( myParms,	"bgblue", &bgcols[2] ) &&
		 cpiGetMetaFloat32( myParms, "bgalpha", &bgcols[3] ) )
	{
		switch ( result->myContext.channels )
		{	
			case 1:
				if ( result->myContext.bitsPerPel == 8 )
					grid1chan( result, linewid, linegap, (CPI_Uint8)255 );
				else if ( result->myContext.bitsPerPel == 16 )
					grid1chan( result, linewid, linegap, (CPI_Uint16)65535 );
				else
					grid1chan( result, linewid, linegap, (CPI_Float32)1.0F );
				break;
			case 3:
				if ( result->myContext.bitsPerPel == 8 )
					grid3chan( result, linewid, linegap,
							   linecols, bgcols, (CPI_Uint8)255 );
				else if ( result->myContext.bitsPerPel == 16 )
					grid3chan( result, linewid, linegap,
							   linecols, bgcols, (CPI_Uint16)65535 );
				else
					grid3chan( result, linewid, linegap,
							   linecols, bgcols, (CPI_Float32)1.0F );
				break;
			case 4:
				if ( result->myContext.bitsPerPel == 8 )
					grid4chan( result, linewid, linegap,
							   linecols, bgcols, (CPI_Uint8)255 );
				else if ( result->myContext.bitsPerPel == 16 )
					grid4chan( result, linewid, linegap,
							   linecols, bgcols, (CPI_Uint16)65535 );
				else
					grid4chan( result, linewid, linegap,
							   linecols, bgcols, (CPI_Float32)1.0F );
				break;
			default:
				cpiError( "Image has incorrect number of channels" );
		}

		retval = CPI_TRUE;
	}
	else
		cpiError( "Couldn't get Grid parameter(s)" );
		
	return retval;
}


//////////////////////////////////////
// This defines for Rayz the output region of your image
// op. It just passes on the values from the GUI.

static CPI_Bool
GridOutputRegion( CPI_ImageContext	*output,
					 CPI_ImageContext	 inputs[],
					 CPI_Uint32			 numInputs,
					 CPI_Metadata		 parms )
{	
	CPI_Bool retval = CPI_FALSE;

	if ( cpiGetMetaInt32( parms, "sizex", &output->sizeX ) &&
		 cpiGetMetaInt32( parms, "sizey", &output->sizeY ) &&
		 cpiGetMetaInt32( parms, "channels", &output->channels ) &&
		 cpiGetMetaInt32( parms, "bitdepth", &output->bitsPerPel ) )
	{
		output->offsetX = 0;
		output->offsetY = 0;
		retval			= CPI_TRUE;
	}
	
	return retval;
}


/////////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM.

static RPI_ImageOpInfo opinfo = 
{
	{
		"grid",				// op name
		"",					// not used
		"Silicon Grail",	// author
		"1.0"				// version
	},
	GridProcess,			// image execution process
	NULL,					// InputRegion
	GridOutputRegion,		// output region
	NULL,					// control processing
	NULL,					// setup processing
	NULL,					// cleanup processing
	0,						// min inputs
	CPI_FALSE,				// does it multiproc?
	CPI_FALSE				// handle disjoint regions?
};


////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister node and image op

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "grid" );
	cpiUnregisterNode( "grid" );
}
