/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	Regions
//
// COMMENTS: This demonstrates region handling in Rayz. When
// 			 compositing images of different sizes, or with
//			 different offsets, there will be 2 or 3 different
//			 regions - 
//			 1) where they share pixels (intersect)
//			 2) where input A exists, but not B
//			 3) where input B exists, but not A
//			 Rayz provides routines for accessing these different 
//			 regions.
//
*/

#include <stdio.h>
#include <CPI.h>

////////////////////////////////////////////
// The init routine is called once, and typically is used to
// define the user input controls

static CPI_Bool
RegionsInit( void )
{
	cpiAddToggle( "fill", "Fill with Red",
				  CPI_FALSE,
				  CPI_PARM_DEFAULT,
				  "Fill With Color or Copy Outside!",
				  NULL );

	return CPI_TRUE;
}


/////////////////////////////////////////////
// shutdown does nothing in this case

static CPI_Bool
RegionsShutDown( void )
{
	return CPI_TRUE;
}


/////////////////////////////////////////////
// This is called before processing begins. In this case, it
// checks the inputs to make sure they match expectations.
// In general, it offers a way to generate a user error before
// the user attempts to generate an image, if appropriate.


static CPI_Bool
RegionsGetErrors( CPI_Float32 time )
{
	CPI_ImageContext	inputA;
	CPI_ImageContext	inputB;
	CPI_Bool			retval = CPI_TRUE;

	if ( cpiGetInputContext( &inputA, 0, time ) &&
		 cpiGetInputContext( &inputB, 1, time ) )
	{
		if ( inputA.channels != inputB.channels )
		{
			cpiError( "Inputs A and B must have the same number of channels" );
			retval = CPI_FALSE;
		}
		else if ( inputA.bitsPerPel != inputB.bitsPerPel )
		{
			cpiError( "Inputs A and B must have the same bit depth" );
			retval = CPI_FALSE;
		}
	}
	else
	{
		cpiError( "Couldn't get input image contexts" );
		retval = CPI_FALSE;
	}

	return retval;
}


/////////////////////////////////////////
// This is called when an image is required
// The basic structure of these routines is:
// 1. get the inputs
// 2. get the parameters from the user interface
// 3. allocate metadata and copy the parameters to it
// 4. add the image op(s) to the list of operations
// In this case, there is a single toggle - if true, it fills
// the outside of the image intersection with red

static CPI_ImageOp
RegionsExec( CPI_Float32	myTime,
			 CPI_Uint8		quality,
			 CPI_Uint32		nodeOutput,
			 CPI_Bool	 	viewerOutput,
			 CPI_Float32	scaleX,
			 CPI_Float32   	scaleY )

{
	CPI_Bool		 fillWithColor;
	CPI_ImageOp		 myInputs[2];
	CPI_Metadata	 myParms	= cpiCreateMetadata();
	CPI_ImageOp		 retval		= NULL;

	myInputs[0] = cpiGetInputOp( 0, myTime, quality, scaleX, scaleY );
	myInputs[1] = cpiGetInputOp( 1, myTime, quality, scaleX, scaleY );

	if ( cpiGetInteger( &fillWithColor, "fill", myTime ) )
	{
		cpiSetMetaBool( myParms, "color", 0 != fillWithColor );

		if ( NULL != myInputs[0] && NULL != myInputs[1] )
			retval = cpiAddImageOp( "regions", myParms, myInputs, 2 );

		cpiDeleteMetadata( myParms );
	}
	else
		cpiError( "Couldn't get toggle value" );

	return retval;
}


//////////////////////////////////////////////
// This structure defines aspects of the node - user subroutines,
// number of inputs, etc

static RPI_NodeInfo ninfo = {
	{
		"regions",			// node name
		"Regions",			// menu name
		"Silicon Grail",	// author
		"v0.1"				// version
	},
	RegionsInit,			// NodeInit
	RegionsShutDown,		// NodeShutDown
	NULL,					// ParmChanged
	NULL,					// InputChanged
	NULL,					// GetRegion
	NULL,					// GetFullSize
	NULL,					// GetRange
	RegionsGetErrors,		// GetErrors
	RegionsExec,			// Node Execute
	NULL,					// InputLabel
	NULL,					// OutputLabel
	NULL,					// ViewerLabel
	NULL,					// ViewerOutputs
	2,						// min inputs
	2,						// max inputs
	1,						// min outputs
	1,						// max outputs
	CPI_FALSE				// is growable?
};


///////////////////////////////////////////////////////////
// Image Op Functions /////////////////////////////////////
///////////////////////////////////////////////////////////

////////////////////////////////////////
// This templated function does the actual work. By making
// it a template, it is only defined once, and the compiler
// creates a version for each of 8 bit, 16 and float inputs.
// In this case, it multiplies the two inputs to produce the
// output. You are guaranteed that these images will match in
// size and offset if the 'disjoint regions' flag is set to
// CPI_FALSE in the Image Op info structure

template <class T> static
void mult( CPI_Image	*result,
		   CPI_Image	*inputA,
		   CPI_Image	*inputB,
		   T			 max )
{
	CPI_Uint32	 x, y, c;
	CPI_Uint32	 chans;
	T			*pr, *pa, *pb;
	CPI_Float32	 val;
	CPI_Uint32 	 endY  	= result->myContext.offsetY + result->myContext.sizeY;
	CPI_Uint32	 endX	= result->myContext.offsetX + result->myContext.sizeX;

	pr		= (T *)result->data;
	pa		= (T *)inputA->data;
	pb		= (T *)inputB->data;
	chans	= result->myContext.channels;

	for ( y = result->myContext.offsetY; y < endY; ++y )
	{
		pr = (T *)cpiGetLine( result, y );
		pa = (T *)cpiGetLine( inputA, y );
		pb = (T *)cpiGetLine( inputB, y );

		for ( x = result->myContext.offsetX; x < endX; ++x )
		{
			for ( c = 0; c < chans; c++ )
			{
				val = ( *pa++ + *pb++ );
				
				if ( val > max )
					*pr++ = (T)max;
				else
					*pr++ = (T)val;
 			}			
		}
	}
}


//////////////////////////////////////////
// The Control Processing function is used to put a top layer of
// compute control on your Image Operation. If this routine is
// defined, then control begins here, and it is up to this routine
// to compute the image(s) properly - in this example, we just
// do that by calling cpiCompute(), but you could call your
// Process function directly, or even not have a process function
// and just do all the work here.
// In this example, we use this routine to find the intersection
// of the two inputs, and then process the region(s) OUTSIDE that
// intersection.
// The *result param comes in set to your first input (by
// default) unless you have set it to something else in OutputRegion()

static CPI_Bool
RegionsControlProcessing( CPI_Image		*result,
						  CPI_Image		 inputs[],
						  CPI_Uint32	 numinputs,
						  CPI_Metadata	 myParms )
{
	CPI_ImageContext	shared;
	CPI_Bool			color;
	CPI_Bool			retval = CPI_FALSE;

	if ( numinputs == 2 )
	{
		// find the intersection of two image regions 
		if ( cpiIntersectRegions( inputs[0].myContext,
								  inputs[1].myContext,
								  &shared ) )
		{
			cpiGetMetaBool( myParms, "color", &color );  

			// if flag set, color outside area red
			// Magic: puts the result into *result arg for you
			if ( color )
				retval = cpiSetColorOutsideRegion( &shared,
												   1.0,
												   0.0,
												   0.0,
												   0.0 );
			else
				// Magic: puts the result into *result arg for you
				retval = cpiCopyOutsideRegion( &shared, &inputs[0] );
		}

		// this kicks off the processing that is defined by
		// the Process function, if any
		cpiCompute( true );	
	}

	return retval;
}


//////////////////////////////////////////
// This is called to execute the image op
// The basic structure of these routines is
// 1. unload parameter values from metadata
// 2. call the appropriate routine depending on bit depth

static CPI_Bool
RegionsProcess( CPI_Image		*result,
				CPI_Image	  	 inputs[],
				CPI_Uint32	  	 numinputs,
				CPI_Metadata	 myParms )
{
	CPI_Bool retval = CPI_TRUE;

	if ( result->myContext.bitsPerPel == 8 )
		mult( result, &inputs[0], &inputs[1], (CPI_Uint8)255 );
	else if ( result->myContext.bitsPerPel == 16 )
		mult( result, &inputs[0], &inputs[1], (CPI_Uint16)65535 );
	else if ( result->myContext.bitsPerPel == 32 )
		mult( result, &inputs[0], &inputs[1], (CPI_Float32)1.0 );
	else
	{
		cpiError( "Result image has an unknown bitdepth" );
		retval = CPI_FALSE;
	}

	return retval;
}


///////////////////////////////////////////


static RPI_ImageOpInfo opinfo = 
{
	{
		"regions",			// command name
		"",					// not used
		"Silicon Grail",	// author
		"0.1"				// version
	},
	RegionsProcess,			// image execution process
	NULL,					// InputRegion
	NULL,					// OutputRegion
	RegionsControlProcessing,
	NULL,					// setup processing
	NULL,					// cleanup processing
	2,						// minimum inputs req'd
	CPI_TRUE,				// can this plugin multiproc?
	CPI_FALSE,				// supports disjoint regions?
};

//////////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = true;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister node and image op

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "regions" );
	cpiUnregisterNode( "regions" ); 
}
