/*
// 
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME: 	Shift
//
// COMMENTS: Demonstrate context and regions
//
*/

#include <stdio.h>
#include <math.h>
#include <CPI.h>


// sets up a single parameter for shifting in X
// the extension to 2 directions is left as an exercise

static CPI_Bool
ShiftInit( void )
{
	cpiAddInteger( "pixels", "Pixels", 
				    0, -256, 256, CPI_PARM_ANIMATABLE,
					NULL, NULL);

	return CPI_TRUE;
}


//////////////////////////////////////////
// This is called when the node is actually executed

static CPI_ImageOp
ShiftExec( CPI_Float32  myTime,
		   CPI_Uint8	quality,
		   CPI_Uint32   nodeOutput,
		   CPI_Bool		viewerOutput,
		   CPI_Float32	scaleX,
		   CPI_Float32	scaleY )
{
	CPI_Int32		pixels;
	CPI_ImageOp		retval	= NULL;
	CPI_ImageOp 	myInputs;
	
	myInputs = cpiGetInputOp( 0, myTime, quality, scaleX, scaleY );

	// we are not insuring that the shift value makes sense, but
	// we should, and this would be the place to do it.
	if ( cpiGetInteger( &pixels, "pixels", myTime ) )
	{
		CPI_Metadata    nodeparms = cpiCreateMetadata();

		cpiSetMetaInt32( nodeparms, "pixels", pixels );

		if ( myInputs != NULL )
		   	retval = cpiAddImageOp( "shift", nodeparms, &myInputs, 1 );
		else
			retval = NULL;

		cpiDeleteMetadata( nodeparms );
	}
	else
	{
		cpiError( "Couldn't retrieve shift value" );
		retval = NULL;
	}

	return retval;
}

///////////////////////////////////////////
// GetRegion defines the output region of the node
// to any other nodes that request it

static CPI_Bool
ShiftGetRegion( CPI_ImageContext   *region,
				CPI_Float32         myTime,
				CPI_Uint32          output,
				CPI_Bool			viewerOutput )
{
	CPI_Int32			 shift;
	CPI_ImageContext	 input;
	CPI_Bool			 retval = CPI_FALSE;
	
	if ( cpiGetInteger( &shift, "pixels", myTime ) &&
		 cpiGetInputContext( &input, 0, myTime ) )
	{
		region->sizeX 		= input.sizeX;
		region->sizeY 		= input.sizeY;
		region->offsetX 	= input.offsetX + shift;
		region->offsetY 	= input.offsetY;
		region->channels 	= input.channels;
		region->bitsPerPel	= input.bitsPerPel;
		retval				= CPI_TRUE;
	}
	
	return retval;
}


////////////////////////////////////////////
// This structure defines aspects of the node - user subroutines,
// number of inputs, etc

static RPI_NodeInfo ninfo = {
	{
		"shift",				// node name
		"Shift",				// menu name
		"Silicon Grail",		// author
		"v1.0",					// version
		NULL, NULL
	},
	ShiftInit,			// NodeInit
	NULL,				// NodeShutDown
	NULL,				// ParmChanged
	NULL,				// InputChanged
	ShiftGetRegion,		// GetRegion
    NULL,				// GetFullSize
	NULL,				// GetRange
	NULL,				// GetErrors
	ShiftExec,			// Node Execute
	NULL,				// InputLabel
	NULL,				// OutputLabel
	NULL,				// ViewerLabel
	NULL,				// ViewerOutputs
	1, 					// min inputs
	1,					// max inputs
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable (eg Mcomp)
};


///////////////////////////////////////
// user-define pixel operation
// all we have to do is a simple copy of input to output
// the context manipulations have done the real work

template <class T> static
void imCopy( CPI_Image	   &input,
			 CPI_Image	   *result, 
			 T 			    max)
{
	CPI_Int32	 i, x, y, endX, endY, j = 0;
	CPI_Int32	 chans = result->myContext.channels;
	T			*pin, *pout;

	endX	= result->myContext.offsetX + result->myContext.sizeX;
	endY	= result->myContext.offsetY + result->myContext.sizeY;
	
	for ( y = result->myContext.offsetY; y < endY; y++ )
	{
		// We use cpiGetLine to avoid any sub-image issues
		// as rayz may ask us to process a small part of a larger image
		pin 	= (T *)cpiGetLine( &input, y );
		pout 	= (T *)cpiGetLine( result, y );
		j		= 0;

		for ( x = result->myContext.offsetX; x < endX; x++ )
		{
			// copy each channel
			for ( i = 0; i < chans; i++, j++ )
				pout[j] = pin[j];
		}
	}
}


/////////////////////////////////////////
// This is the image operation. 
// Note that by now, we don't care about the shift parameter, since
// the work has already been done.

static CPI_Bool
ShiftProcess( CPI_Image		*result,
			  CPI_Image	 	 inputs[],
			  CPI_Uint32	 numInputs,
			  CPI_Metadata   myParms )
{
	if ( result->myContext.bitsPerPel == 8 )
		imCopy( inputs[0], result, (CPI_Uint8)255 );
	else if ( result->myContext.bitsPerPel == 16 )
		imCopy( inputs[0], result, (CPI_Uint16)65535 );
	else
		imCopy( inputs[0], result, (CPI_Float32)1.0F );

		
	return CPI_TRUE;
}

// By default, the output region is the same as the input region.
// Since that won't be the case here, we have to define to Rayz
// what the output region looks like. That means getting the shift
// value and adding it to the incoming offset value

static CPI_Bool
ShiftOutputRegion( CPI_ImageContext    *output,
				   CPI_ImageContext     inputs[],
				   CPI_Uint32           numInputs,
				   CPI_Metadata         parms )
{
	CPI_Int32	 shift;
	CPI_Bool	 retval = CPI_FALSE;
	
	if ( cpiGetMetaInt32( parms, "pixels", &shift ) )
	{
		output->sizeX 		= inputs[0].sizeX;
		output->sizeY 		= inputs[0].sizeY;
		output->offsetX 	= inputs[0].offsetX + shift;
		output->offsetY 	= inputs[0].offsetY;
		output->channels 	= inputs[0].channels;
		output->bitsPerPel 	= inputs[0].bitsPerPel;
		retval				= CPI_TRUE;
	}
	else
		cpiError( "Couldn't get Shift parameter" );
		
	return retval;
}

///////////////////////////////////////
// InputRegion is called to determine what input region (context)
// will be required to produce the given output region.

static CPI_Bool
ShiftInputRegion( CPI_ImageContext *input,
				  CPI_ImageContext  output,
				  CPI_Uint32 		whichInput,
				  CPI_Metadata 		parms )
{
	CPI_Int32	 shift;
	CPI_Bool	 retval = CPI_FALSE;
	
	if ( cpiGetMetaInt32( parms, "pixels", &shift ) )
	{
		input->sizeX 		= output.sizeX;
		input->sizeY 		= output.sizeY;
		// undo the effect of the shift we are going to do
		input->offsetX 		= output.offsetX - shift;
		input->offsetY 		= output.offsetY;
		input->channels 	= output.channels;
		input->bitsPerPel 	= output.bitsPerPel;
		retval				= CPI_TRUE;
	}
	else
		cpiError( "Couldn't get the number of pixels to shift" );
		
	return retval;
}

/////////////////////////////////////
// Structure which defines the image op 

static RPI_ImageOpInfo opinfo = 
{
	{
		"shift",			// command name
		"",			        // not used
		"Silicon Grail",	// author
		"v1.0",				// version
		NULL, NULL
	},
	ShiftProcess,			// image execution process
	ShiftInputRegion,		// InputRegion
	ShiftOutputRegion,		// OutputRegion
	NULL,                   // ControlProcessing
	NULL,                   // SetupProcessing
	NULL,                   // CleanupProcessing
	1,						// min number of inputs required
	CPI_TRUE,				// Can this plugin multiproc?
	CPI_TRUE                // Disjoint regions?
	// This flag is set to CPI_FALSE by default. We need to set it
	// to CPI_TRUE for this example, because the input region and
	// the output region are different, and do not line up
};


CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;	
}


CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "shift" );
	cpiUnregisterNode( "shift" );
}
