/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:    Simple
//
// COMMENTS: Demonstrate the minimum plugin definition possible
//			 in Rayz. This node takes no parameters - it copies
//			 its input to its output.
//
*/

#include <stdio.h>
#include <math.h>
#include <CPI.h>


/////////////////////////////////////////
// Init is called when the node is created and is 
// typically used to define the user input controls.
// Here, there is nothing to do.

static CPI_Bool
SimpleInit( void )
{
	return CPI_TRUE;
}


//////////////////////////////////////////
// This is called when the node is actually executed
// This gets the input image, and then adds the image
// operation to the list of ops to be executed

static CPI_ImageOp
SimpleExec( CPI_Float32		myTime,
			CPI_Uint8		quality,
			CPI_Uint32		nodeOutput,
			CPI_Bool	 	viewerOutput,
			CPI_Float32		scaleX,
			CPI_Float32   	scaleY )

{
	CPI_ImageOp	retval	= NULL;
	CPI_ImageOp myInputs;
	
	myInputs = cpiGetInputOp( 0, myTime, quality, scaleX, scaleY );

	if ( myInputs != NULL )
	   	retval = cpiAddImageOp( "opname", NULL, &myInputs, 1 );

	return retval;
}


////////////////////////////////////////////
// This structure, which is required, defines to Rayz
// the various routines which are part of this plugin.
// Routines which are not required are set to NULL

static RPI_NodeInfo ninfo = {
	{
		"nodename",				// node name
		"menuname",				// menu name
		"Joe Programmer",		// author
		"v1.0"					// version
	},
	SimpleInit,			// NodeInit
	NULL,				// NodeShutDown
	NULL,				// ParmChanged
	NULL,				// InputChanged
	NULL,				// GetRegion
    NULL,				// GetFullSize
	NULL,				// GetRange
	NULL,				// GetErrors
	SimpleExec,			// Node Execute
	NULL,				// InputLabel
	NULL,				// OutputLabel
	NULL,				// ViewerLabel
	NULL,               // ViewerOutputs
	1, 					// min inputs
	1,					// max inputs
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable (eg Mcomp)
};



//////////////////////////////////////////////////////////////////////
// Image Op Functions ////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

// This is the actual image manipulation routine. It is a templated
// function so that we can write it once, and let the compiler build
// a version for each of 8, 16 and float pixel types.
// This just copies input to result

template <class T> static
void simple( CPI_Image	&input,
			 CPI_Image	*result, 
			 T 			 max)
{
	CPI_Int32	 x, y, endX, endY, i, j = 0;
	CPI_Int32	 chans = result->myContext.channels;
	T			*pin, *pout;

	endX	= result->myContext.offsetX + result->myContext.sizeX;
	endY	= result->myContext.offsetY + result->myContext.sizeY;
	
	for ( y = result->myContext.offsetY; y < endY; y++ )
	{
		// this detects the user requesting an early exit
		if ( cpiIsProcessingAborted() )
			break;

		// Notice that we use cpiGetLine to avoid any sub-image issues
		// as rayz may ask us to process a small part of a larger image
		pin		= (T *)cpiGetLine( &input, y );
		pout	= (T *)cpiGetLine( result, y );
		j		= 0;

		for ( x = result->myContext.offsetX; x < endX; x++ )
		{
			// copy each channel
			for ( i = 0; i < chans; i++, j++ )
				pout[j] = pin[j];
		}
	}
}


/////////////////////////////////////////
// This is called to execute the image op
// It just calls the appropriate copy routine, 
// based on the input bit depth

static CPI_Bool
SimpleProcess( CPI_Image	*result,
			   CPI_Image	 inputs[],
			   CPI_Uint32	 numInputs,
			   CPI_Metadata  myParms )
{
	if ( result->myContext.bitsPerPel == 8 )
		simple( inputs[0], result, (CPI_Uint8)255 );
	else if ( result->myContext.bitsPerPel == 16 )
		simple( inputs[0], result, (CPI_Uint16)65535 );
	else
		simple( inputs[0], result, (CPI_Float32)1.0F );

		
	return CPI_TRUE;
}


/////////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM

static RPI_ImageOpInfo opinfo = 
{
	{
		"opname",			// command name
		"",			        // not used
		"Joe Programmer",	// author
		"0.1"				// version
	},
	SimpleProcess,			// image execution process
	NULL,					// InputRegion
	NULL,					// OutputRegion,
	NULL,                   // ControlProcessing
	NULL,                   // SetupProcessing
	NULL,                   // CleanupProcessing
	1,						// min number of inputs required
	CPI_TRUE,				// Can this plugin multiproc?
	CPI_FALSE				// Disjoint regions?
};


////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister node and image op

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "opname" );
	cpiUnregisterNode( "nodename" );
}
