/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:    SimpleTwoInput
//
// COMMENTS: Show how to build a node with two inputs. The 2 input
//			 requirement shows up in the RPI_NodeInfo structure,
//			 as well as in the Process() function.
//
*/

#include <stdio.h>
#include <math.h>
#include <CPI.h>


/////////////////////////////////////////
// Init is called when the node is created, and is used
// to define the user input controls

static CPI_Bool
SimpleInit( void )
{
	cpiAddFloat( "dissolve", "Dissolve", 1.0, 0.0, 1.0,
				  CPI_PARM_ANIMATABLE | CPI_PARM_CLAMPMIN,
				  "Amount of dissolve from input A to input B",
				  NULL);

	return CPI_TRUE;
}

//////////////////////////////////////////
// label the node inputs - this is optional
// Input is the number of the input you want to provide a
// label for, starting at 0 - this label shows up in the
// status line when the user mouses over the connector.
// The first letter of each string is used to label the input
// connector.

static const char *
SimpleInputLabel( CPI_Uint32	input)
{
	static const char	input1[] = {"foreground image"};
	static const char	input2[] = {"background image"};

	if ( input == 0 )
		return input1;
	else
		return input2;
}


//////////////////////////////////////////
// This is called when an image is required
// The basic structure of these routines is:
// 1. get the inputs
// 2. get the parameters from the user interface
// 3. allocate metadata and copy the parameters to it
// 4. add the image op(s) to the list of operations

static CPI_ImageOp
SimpleExec( CPI_Float32		myTime,
			CPI_Uint8		quality,
			CPI_Uint32		nodeOutput,
			CPI_Bool	 	viewerOutput,
			CPI_Float32		scaleX,
			CPI_Float32   	scaleY )

{
	CPI_Float32		dissolve;
	CPI_ImageOp		retval	= NULL;
	CPI_ImageOp 	myInputs[2];
	
	// get the two input images
	myInputs[0] = cpiGetInputOp( 0, myTime, quality, scaleX, scaleY );
	myInputs[1] = cpiGetInputOp( 1, myTime, quality, scaleX, scaleY );

	// get the value from the user
	cpiGetFloat ( &dissolve, "dissolve", myTime );

	// create metadata and copy the value to it
	CPI_Metadata    nodeparms = cpiCreateMetadata();
	cpiSetMetaFloat32( nodeparms, "dissolve", dissolve );

	// add this image op to the list, with its inputs
	if ( myInputs[0] != NULL && myInputs[1] != NULL )
	   	retval = cpiAddImageOp( "twoinput", nodeparms, myInputs, 2 );

	cpiDeleteMetadata( nodeparms );

	return retval;
}


////////////////////////////////////////////
// This structure, which is required, defines to Rayz
// the various routines which are part of this plugin.
// Routines which are not required are set to NULL

static RPI_NodeInfo ninfo = {
	{
		"twoinput",				// node name
		"Twoinput",				// menu name
		"Joe Programmer",		// author
		"v1.0"					// version
	},
	SimpleInit,			// NodeInit
	NULL,				// NodeShutDown
	NULL,				// ParmChanged
	NULL,				// InputChanged
	NULL,				// GetRegion
    NULL,				// GetFullSize
	NULL,				// GetRange
	NULL,				// GetErrors
	SimpleExec,			// Node Execute
	SimpleInputLabel,	// InputLabel
	NULL,				// OutputLabel
	NULL,               // ViewerLabel
	NULL,               // ViewerOutputs
	2, 					// min inputs
	2,					// max inputs
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable (eg Mcomp)
};



//////////////////////////////////////////////////////////////////////
// Image Op Functions ////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

// This is the actual image manipulation routine. It is a templated
// function so that we can write it once, and let the compiler build
// a version for each of 8, 16 and float pixel types.

// You are guaranteed that the two images (and the result) will
// all be the same size if the 'disjoint regions' flag is CPI_FALSE
// in the RPI_ImageOpInfo structure below. See later note.

template <class T> static
void simple( CPI_Image	    inputs[],
			 CPI_Image	   *result, 
			 CPI_Float32	dissolve,
			 T 			    max)
{
	CPI_Uint32	 i, x, y, endX, endY;
	CPI_Uint32	 chans = result->myContext.channels;
	T           *pinA, *pinB, *pout;
	CPI_Float32	 oneminus;

	endX	= result->myContext.offsetX + result->myContext.sizeX;
	endY	= result->myContext.offsetY + result->myContext.sizeY;
	oneminus =  1.0 - dissolve;
	
	for ( y = result->myContext.offsetY; y < endY; y++ )
	{
		// Notice that we use cpiGetLine to avoid any sub-image issues
		// as rayz may ask us to process a small part of a larger image
		pinA	= (T *)cpiGetLine( &inputs[0], y );
		pinB	= (T *)cpiGetLine( &inputs[1], y );
		pout	= (T *)cpiGetLine( result, y );

		for ( x = result->myContext.offsetX; x < endX; x++ )
		{
			// do the math
			for ( i = 0; i < chans; i++ )
			{
				*pout = (T) ((*pinA * dissolve) + (*pinB * oneminus));
				pinA++;
				pinB++;
				pout++;
			}
		}
	}
}


/////////////////////////////////////////
// This is called to execute the image op
// The basic structure of these routines is
// 1. unload parameter values from metadata
// 2. call the appropriate routine depending on bit depth

static CPI_Bool
SimpleProcess( CPI_Image	*result,
			  CPI_Image	 	 inputs[],
			  CPI_Uint32	 numInputs,
			  CPI_Metadata   myParms )
{
	CPI_Float32	dissolve;

	// get the value
	cpiGetMetaFloat32( myParms, "dissolve", &dissolve );

	// call the routine with parameters
	if ( result->myContext.bitsPerPel == 8 )
		simple( inputs, result, dissolve, (CPI_Uint8)255 );
	else if ( result->myContext.bitsPerPel == 16 )
		simple( inputs, result, dissolve, (CPI_Uint16)65535 );
	else
		simple( inputs, result, dissolve, (CPI_Float32)1.0F );

		
	return CPI_TRUE;
}


/////////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM.

static RPI_ImageOpInfo opinfo = 
{
	{
		"twoinput",			// command name
		"",			        // not used
		"Joe Programmer",	// author
		"0.1"				// version
	},
	SimpleProcess,			// image execution process
	NULL,					// InputRegion
	NULL,					// OutputRegion,
	NULL,                   // ControlProcessing
    NULL,                   // SetupProcessing
    NULL,                   // CleanupProcessing
	2,						// min number of inputs required
	CPI_TRUE,				// Can this plugin multiproc?
	CPI_FALSE               // Disjoint regions?
	// the Disjoint regions flag means this:
	// if == CPI_FALSE, Rayz will compute the common pixels 
	// (intersection) of all your inputs, and just pass those
	// as inputs to the Process() routine.
	// The area outside the intersection will be set to the
	// Backing Color defined in the background's image context. 
	// If you want something other than that to happen, you 
	// need to define the ControlProcessing function. See the 
	// example called NewMult.C
	// if == CPI_TRUE, you will get handed the complete inputs,
	// and you are on your own as to how to process them
};


////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister node and image op

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "twoinput" );
	cpiUnregisterNode( "twoinput" );
}
