/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:    InteractiveOverlay
//
// COMMENTS: This demonstrates an overlay which is not attached
//           to any node, but which instead is attached to the
//           image viewer(s). After installing this plugin, you
//           will see the menu name appear on the right mouse 
//           menu in any image viewer. This also shows how such
//           an overlay can still be interactive.
//
*/

#include <stdio.h>
#include <math.h>
#include <CPI.h>

// typically, you define an overlayState structure to hold
// the current state of what you are drawing. This is then
// handed from function to function as private data.
// The constants are used to identify sections of the overlay
// for selection.

#define INNER_CIRCLE_ID		1
#define OUTER_CIRCLE_ID		2
#define CROSS_ID			3

typedef struct _overlayState
{
	CPI_Float32	centerX;
	CPI_Float32 centerY;
	CPI_Float32	radiusIn;
	CPI_Float32 radiusOut;
	CPI_Float32	selectedX;
	CPI_Float32 selectedY;
	CPI_Uint32	selectedObject;
	CPI_Float32	myTime;
} overlayState;


//////////////////////////////////////
// This creates the overlay you are working with, by allocating
// the defined data structure and passing it on.

static CPI_PrivateData
OverlayCreateInstance( void )
{
	overlayState	*retval;

	retval = (overlayState *)cpiAlloc( sizeof( overlayState ) );

	retval->centerX = 0.5;
	retval->centerY = 0.5;
	retval->radiusIn = 0.15;
	retval->radiusOut = 0.25;
	
	return retval;
}


///////////////////////////////////////
// cleans up before exiting

static void
OverlayDestroyInstance( CPI_PrivateData handle )
{
	overlayState *state = (overlayState *)handle;

	cpiFree( state );
}


///////////////////////////////////////
// You need to define a render routine. This will use the
// drawing commands which you have available - see
// include/CPI_Overlay.h for a complete list.

static void
OverlayRender( CPI_PrivateData handle, CPI_Bool isPicking )
{
	CPI_Float32	  inRad, outRad;
	overlayState *state = (overlayState *)handle;
	
	inRad = state->radiusIn * 0.5;
	outRad = state->radiusOut * 0.5;

	if ( inRad <= 0 )
		inRad = 0.075;
	
	if ( outRad <= 0 )
		outRad = 0.1;
	 
	if ( outRad > 1 )
		outRad = 1;

	// each collection of lines or shapes that you want to be 
	// separately selectable should be delimited by a call to
	// cpiStartNewObject. The ID that you pass will be returned
	// in the selection routine, so you know what the user has
	// selected.

	cpiStartNewObject( INNER_CIRCLE_ID );
	cpiSetColor( 0.0, 0.0, 1.0 );
	cpiDrawCircle( state->centerX, state->centerY, inRad );

	cpiStartNewObject( OUTER_CIRCLE_ID );
	cpiDrawCircle( state->centerX, state->centerY, outRad );

	cpiStartNewObject( CROSS_ID );
	cpiDrawLine( state->centerX - 0.05, 
				 state->centerY, 
				 state->centerX + 0.05, 
				 state->centerY );
	cpiDrawLine( state->centerX, 
				 state->centerY - 0.05,
				 state->centerX, 
				 state->centerY + 0.05 );
}


//////////////////////////////////////////
// If the node is asked to recompute for any reason (such
// as a parameter change, either in this node or elsewhere
// in Rayz), this routine is called. In this case, just
// update the time

static void
OverlaySourceOutOfDate( CPI_PrivateData handle, 
						CPI_ImageContext fullsize,
						CPI_Float32 theTime )
{
	overlayState *state = (overlayState *)handle;

	state->myTime = theTime;
}


////////////////////////////////////////////
// When an input image changes, this routine is called to
// update any values which may be based on that image.

static void
OverlaySourceChanged( CPI_PrivateData handle, 
					  CPI_ImageContext fullsize,
					  CPI_Float32 theTime )
{
	OverlaySourceOutOfDate( handle, fullsize, theTime );
}


///////////////////////////////////////////
// These routines are called to service the appropriate
// hardware activity. They are only called if you have a current
// object under the cursor. The input values are
// handle - your private structure 
// objectID - the ID of the overlay image that is under the cursor
// x,y - the current x,y position of the cursor, ranging from 0.0
// to 1.0
// keyMod - any modifier key which is pressed. These are defined
// in include/CPI_OverlayProvider.h, as
// CPI_CTRL_KEY  - control key
// CPI_ALT_KEY   - alt key
// CPI_SHIFT_KEY - shift key

static void
OverlayLeftMouseDown( CPI_PrivateData	handle,
					  CPI_Uint32		objectID,
					  CPI_Float32		x,
					  CPI_Float32		y,
					  CPI_Uint32		keyMod )
{
	overlayState	*state = (overlayState *)handle;

	state->selectedObject	= objectID;

	state->selectedX		= x;
	state->selectedY		= y;
	
	cpiSetSelectLock( CPI_TRUE );	
}


/////////////////////////////////////////
// same as before, but now we are dragging the cursor, so we
// need to update the overlay image with cpiDoRedraw()

static void
OverlayLeftMouseDrag( CPI_PrivateData	 handle,
					  CPI_Uint32		 objectID,
					  CPI_Float32		 x,
					  CPI_Float32		 y,
					  CPI_Uint32		 keyMod,
					  CPI_Bool			 isContinuous )
{
	overlayState	*state = (overlayState *)handle;

	if ( state->selectedObject == CROSS_ID )
	{
		state->centerX = x;
		state->centerY = y;
	}
	else if ( state->selectedObject == OUTER_CIRCLE_ID )
	{
		state->radiusOut = ( hypot( x - state->centerX, y - state->centerY ) *
							 2 );
	}
	else if ( state->selectedObject == INNER_CIRCLE_ID )
	{
		state->radiusIn = ( hypot( x - state->centerX, y - state->centerY ) * 
							2 );
	}

	cpiDoRedraw();   
}


/////////////////////////////////////////
// unlock the current selection and update current values

static void
OverlayLeftMouseUp( CPI_PrivateData	 handle, 
				    CPI_Uint32		 objectID,
				    CPI_Float32		 x,
				    CPI_Float32		 y,
				    CPI_Uint32		 keyMod )
{
	overlayState *state = (overlayState *)handle;

	cpiSetSelectLock( CPI_FALSE );
}


////////////////////////////////////////
// As with nodes and image ops, overlays are defined
// to Rayz with a structure. This structure must be
// present, but the routines it defines are optional.
// Missing routines must be marked with NULL

static RPI_OverlayInfo oinfo = {
	{
		"interact",			// internal name
		"Moving Overlay",	// menu name
		"Silicon Grail",	// author
		"v0.1"				// versioin
	},
	NULL,					// Init
	NULL,					// Shutdown
	OverlayCreateInstance,
	OverlayDestroyInstance,
	OverlaySourceChanged,
	NULL,					// OverlayFullSizeChanged
	OverlaySourceOutOfDate,
	OverlayRender,
	OverlayLeftMouseDown,
	OverlayLeftMouseDrag,
	OverlayLeftMouseUp,
	NULL,					// OverlayMouseEnter
	NULL,					// OverlayMouseDrag
	NULL,					// OverlayMouseExit
	NULL,					// node to attach to
	// since the node name is NULL, the overlay is
	// attached directly to any image viewers that exist
	CPI_FALSE
};


////////////////////////////////////
// register this plugin

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterOverlay( &oinfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister this plugin. Note the internal name is used.

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterOverlay( "interact" );
}
