/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 Seward St
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	CPI_DeviceProvider
//
// COMMENTS:
//	This is the interface to the device plugins, which defines the way rayz
// interacts with files..  It will handle opening, closing, reading and seeking
// in files.
//
// A Device plugin contains an opaque ( to rayz ) handle, which can contain
// any necessary information about the device.  This handle is then passed
// into each of the other functions to be manipulated as needed.
*/

#ifndef _CPI_DeviceProvider_h_
#define _CPI_DeviceProvider_h_

#include <CPI/CPI_Base.h>


/* //////////////////////////////////////// */


#ifdef __cplusplus
extern "C" {
#endif

/* These defines are used in the seek functions to define if it's necessary
 * to seek from the beginning, current position or from the end.
 */
#define CPI_SEEK_START 0
#define CPI_SEEK_CUR 1
#define CPI_SEEK_END 2

/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_Metadata data - Metadata which contains all of the necessary
//						information about the file to open such as the filename
// 						and read/write mode...
// Return Value:
//	CPI_PrivateData - user created structure which contains any information
//					  which will be needed to use this file in the future such
//					  as a file handle, max request size or user name.  This
//					  structure should be allocated in the open function.
*/
typedef CPI_PrivateData		(*RPI_DeviceOpenFn)( CPI_Metadata data );


/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData device - user created structure needed to access the
//							 currently open file.  NOTE: this structure
//							 should probably be freed in this function
// Return Value : void
*/
typedef void				(*RPI_DeviceCloseFn)( CPI_PrivateData device );


/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData device - user created structure needed to check if the
//							 handle is currently opened
// Return Value:
//  CPI_Bool - indicating whether the file is open or not
*/
typedef CPI_Bool			(*RPI_DeviceIsOpenFn)( CPI_PrivateData device );


/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData device - user created structure needed to sync the device
//
// Return Value:
//	void
*/
typedef void				(*UPI_DeviceSyncFn)( CPI_PrivateData device );


/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData device - user created structure needed to get the position
//							 of the handle in the file
// Return Value:
//	CPI_Uint64 - Position ( in bytes ) we are from the start of the file
*/
typedef CPI_Uint64			(*UPI_DeviceGetPosFn)( CPI_PrivateData device );


/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData device - user created structure needed to set the position
//							 of the handle in the file
//	CPI_Int64 pos - requested position ( in bytes ) to move the handle
//	CPI_Uint8 whence - position to move the handle from :
//			CPI_SEEK_START - move "pos" bytes from the start of the file
//			CPI_SEEK_CUR - move "pos" bytes from the current position in the
//						   file
//			CPI_SEEK_END - move "pos" bytes from the end of the file
//
// Return Value:
//	CPI_Uint64 - Position ( in bytes ) we are from the start of the file after
//				 we've moved to the new position ( -1 for failure )
*/
typedef CPI_Uint64			(*UPI_DeviceSetPosFn)( CPI_PrivateData device,
												   CPI_Int64		 pos,
												   CPI_Uint8		 whence );
/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData device - user created structure needed to get the length
//						     ( in bytes ) of the current file.
//	CPI_Uint64 *length - length ( in bytes ) of the current file
//
// Return Value:
//	CPI_Bool - returning whether we successfully got the length of the file
*/
typedef CPI_Bool			(*UPI_DeviceLengthFn)( CPI_PrivateData	 device,
												   CPI_Uint64		*length );


/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData device - user created structure needed to read from the
//							 file
//	void *buffer - buffer to read the data into
//	CPI_Uint64 count - number of bytes to read from the file
*/
typedef CPI_Uint64			(*UPI_DeviceReadFn)( CPI_PrivateData	 device,
												 void				*buffer,
												 CPI_Uint64			 count );


/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData device - user created structure needed to write data to the
//							 file
//	void *buffer - buffer to write the data from
//  CPI_Uint64 count - number of bytes to write to the file
*/
typedef CPI_Uint64			(*UPI_DeviceWriteFn)( CPI_PrivateData	 device,
												  const void		*buffer,
												  CPI_Uint64		 count );


/*
// For any optional (upi labeled) functions you don't wish to provide,
// fill in NULL
*/
typedef struct
{

	RPI_PluginInfo		 rpiPluginInfo;

	/* Functions for the device plugin */
	RPI_DeviceOpenFn	 rpiOpen;
	RPI_DeviceCloseFn	 rpiClose;
	RPI_DeviceIsOpenFn	 rpiIsOpen;
	UPI_DeviceSyncFn	 upiSync;
	UPI_DeviceGetPosFn	 upiGetPos;
	UPI_DeviceSetPosFn	 upiSetPos;
	UPI_DeviceLengthFn	 upiFileLength;
	UPI_DeviceReadFn	 upiRead;
	UPI_DeviceWriteFn	 upiWrite;
} RPI_DeviceInfo;

#ifdef __cplusplus
}
#endif

#endif
