/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 Seward St
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	CPI_FileManagerProvider
//
// COMMENTS:
//  Defines the interface to the file manager plugins, which are basically
// the way rayz can get a listing of files from a variety of file
// systems - either the normal filesystem, an ftp device, an accom. whatever.
// A DBMS front end could even be implemented...
//
// The File Manager primarily deals with the concept of a FileID, or file
// identifier. When talking about a normal filesystem, the FileID is just
// the file or path name. However, it is never used as such, which allows
// the file manager to apply arbitrary variables to it. The one we care
// about most is the time ( in frame numbers or partial frame numbers ).
// As an example, if there is a sequence of files on the disk called
// foobar.1.tiff through foobar.200.tiff, the file manager would compress
// this sequence into a FileID that is something like foobar.$F.tiff. Or
// foobar.#.tiff - whatever... There are requests to convert this FileID
// into metadata, which low level devices can use to open and read from
// for a particular time...
*/

#ifndef _CPI_FileManagerProvider_h_
#define _CPI_FileManagerProvider_h_

#include <CPI/CPI_Base.h>
#include <CPI/CPI_FileID.h>


/* //////////////////////////////////// */


#ifdef __cplusplus
extern "C" {
#endif

/*
 * Allows a function to be passed in to filter out function names
 * If this returns CPI_TRUE, the file name is OK to add...
 *
 * The extension should be passed in WITHOUT with the ., so "jpg", not
 * ".jpg"
 *
 * It is OK to pass in NULL for the file extension (i.e. there is no extension).
 * This will match when the user has selected 'All Files', or if the
 * filename is all numbers and user has selected 'All Image Files'.
 */
typedef CPI_Bool			 (*CPI_ExtListFn)( const char *filename,
											   const char *fileext );

/*
 * These flags will be OR'ed together to construct flags for what
 * to put in the fileID list (and what to do to the list)
 */

/*
 * CPI_NO_HIDDEN_IDS means that no hidden files should be added. In unix,
 * this means that anything beginning with a . should be ignored.
 */
#define CPI_NO_HIDDEN_IDS		0x0001


/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	None:
//
// Return Value:
//	CPI_PrivateData - This is an opaque (to rayz) pointer which will be passed
//					  into to all of the file manager functions.  This should
//					  allow you to carry around any necessary information
//					  about a filesystem( i.e: an ftp connection, etc.. ).
//					  This pointer should probably be allocated here and
//					  cleaned up in the shutdown function.
//
// Undefined Behavior - The 'handle' parameter which gets passed to all of the
//						other file manager functions will be NULL.
*/
typedef CPI_PrivateData			(*UPI_InitFn)( void );


/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData handle - User created handle which was created in the
//							 init function.. If allocated there, it should
//							 be freed here..
// Return Value:
//	None.
//
*/
typedef void					(*UPI_ShutDownFn)( CPI_PrivateData handle );


/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData - This is an opaque (to rayz) pointer which will be passed
//					  into to all of the file manager functions.  This should
//					  allow you to carry around any necessary information
//					  about a filesystem( i.e: an ftp connection, etc.. ).
//  CPI_FileIDList dirIDList - ID list that is to hold any sub directories
//                             of the dirID passed in. Directories should
//                             always be added. In terms of a normal
//                             filesystem, it is expected that in general
//                             you shouldn't add "." to this list. If you
//                             choose to add "..", then you are expected
//                             to handle this in your ConcatFileID function
//                             below (if provided - the standard filesystem
//                             concat function will handle this).
//                             Accessed using functions in CPI_FileID.h
//
//                             NB: In general, should be sorted. There is a
//                                 sort function in CPI_FileID.h which is
//                                 a convenience function, where you can
//                                 specify an optional sort function to
//                                 change from standard alpha-numeric
//
//  CPI_FileIDList fileIDList - ID list that holds the list of normal files
//                              that could be opened, based on the flags
//                              and optional extension filter function below.
//                              Accessed using funcitons in CPI_FileID.h
//
//                              NB: In general, should be sorted. There is a
//                                  sort function in CPI_FileID.h which is
//                                  a convenience function, where you can
//                                  specify an optional sort function to
//                                  change from standard alpha-numeric. This
//                                  list will be passed unchanged in to the
//                                  sequence compression function, so being
//                                  sorted usually makes sequences easier to
//                                  find...
//
//  const char *dirID - FileID for the current directory which when
//                      concatenated with the returned FileIDs will
//                      result in a fully qualified FileID
//
//  CPI_Uint32 flags - various flags controlling what and how things should
//                     be put in the file list. See defines above for a possible
//                     set of OR'ed together options
//
//  CPI_ExtListFn extListFn - pointer to a function that (if NOT NULL) should
//                            be called with the extension for the file in
//                            question to see if it matches. If the func
//                            returns CPI_TRUE, the file is OK to add.
//
//                            This is meant ONLY for the fileIDList,
//                            and will be used to do things like filter out
//                            things like all files that extension match
//                            known image formats.
//
//                            NB: Can be NULL, in which case all files
//                                should be considered to be OK
//                            NB: Extension should be passed WITHOUT leading "."...
//
// Return Value:
//  CPI_Bool - return CPI_TRUE if able to open dirID and retrieve listing
//
*/
typedef CPI_Bool		 (*RPI_GetFileIDListFn)( CPI_PrivateData handle,
												 CPI_FileIDList	 dirIDList,
												 CPI_FileIDList	 fileIDList,
												 const char		*dirID,
												 CPI_Uint32		 flags,
												 CPI_ExtListFn	 fileExtFn );

/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData - This is an opaque (to rayz) pointer which will be passed
//					  into to all of the file manager functions.  This should
//					  allow you to carry around any necessary information
//					  about a filesystem( i.e: an ftp connection, etc.. ).
//  const char *fileID - fileID in to convert to metadata
//  CPI_Float32 - floating point representation of the frame number to be
//                used in expanding any time related variables in the fileID
//                to actual values in the metadata
//
// Return Value:
//  CPI_Metadata - metadata structure that should no longer be referenced and
//                 will be freed, representing the meta data for an actual
//                 file on whatever device is specified in the metadata, or
//                 the default device returned by the UPI_GetDeviceNameFn
//                 declared below
*/
typedef CPI_Metadata	 (*RPI_GetMetadataToOpenFn)( CPI_PrivateData handle,
													 const char		*fileID,
													 CPI_Float32	 curSample );

/*
////////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData - This is an opaque (to rayz) pointer which will be passed
//					  into to all of the file manager functions.  This should
//					  allow you to carry around any necessary information
//					  about a filesystem( i.e: an ftp connection, etc.. ).
//  CPI_FileSeqList - opaque structure which you can add string / metadata
//                    pairs to. meta data should have at least start, end
//                    and gaps information for the sequence being added
//                    any other info will be displayed in a list format
//                    on the file browser. If the sequence info is not given
//                    then all 'sequences' without that info are assumed to
//                    be a single frame....
//
//  const CPI_FileIDList - the list from a previous GetFileIDList call.
//
// Return Value:
//  CPI_Bool - should return CPI_TRUE if the operation succeeded
//
// Undefined behavior:
//  If this function is not provided, then no sequences will be available.
// Every file will be treated as a single entity. This is probably the
// case if a file manager is being written for a rayz file browser or
// channel file or something.
*/
typedef CPI_Bool		 (*UPI_FillSeqInfoFn)( CPI_PrivateData		 handle,
											   CPI_FileIDSeqList	 seqlist,
											   const CPI_FileIDList	 filelist );

/*
////////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData - This is an opaque (to rayz) pointer which will be passed
//					  into to all of the file manager functions.  This should
//					  allow you to carry around any necessary information
//					  about a filesystem( i.e: an ftp connection, etc.. ).
//  const char *oldFileID - FileID that can be converted to the native
//                          filesystem. For example, the default filesystem
//                          plugins will try to convert FileIDs between
//                          Unix and Windows NT naming conventions. They
//                          will do this by looking for a set of
//                          preferences accessed through the cpiPref* calls
//                          and finding drive letter - mount point mappings
//                          and converting backslashes to forward slashes
//                          or vice versa
//
//                          Also, if there is any cleanup to the filename,
//                          it should be done here. For example, if the user
//                          types /usr/////foo/bar/ and that is not valid,
//                          or at least unaesthetic, this function should
//                          clean it up and return something like /usr/foo/bar
//
// Return Value:
//  char * - a string which should no longer be referenced upon return, and
//           will be freed by the consumer, representing the new, converted
//           file identifier. If this is NULL, then it is assumed no
//           translation was necessary
//           Should be alloced with cpiAlloc - will be freed with cpiFree
//
// Undefined behavior:
//  If this function is not provided, then the default filesystem file manager
// will be used to do the conversion. If you are constructing file IDs that
// are consistent with the normal native file system, you do not need to
// provide this function
*/
typedef char			*(*UPI_ConvertFileIDFn)( CPI_PrivateData	 handle,
												 const char		*oldFileID );


/*
////////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData - This is an opaque (to rayz) pointer which will be passed
//					  into to all of the file manager functions.  This should
//					  allow you to carry around any necessary information
//					  about a filesystem( i.e: an ftp connection, etc.. ).
//  const char *dirID - the file identifier representing the directory, which
//                      should be placed in the resulting file ID such that
//                      a fully qualified fileID can be used to construct
//                      metadata to open a file.
//  const char *fileID - the file ID to add to the directory
//
// Return Value:
//  char * - a string which should no longer be referenced upon return, and
//           will be freed by the consumer, representing the fully qualified
//           file identifier. For example, for the Unix filesystem file
//           manager, a dirID of "/usr/tmp" and a fileID of "foobar" would
//           result in "/usr/tmp/foobar"
//           If this value is NULL, then it is assumed that there was some
//           sort of error.
//           Should be alloced with cpiAlloc - will be freed with cpiFree
//
// Undefined behavior:
//  If this function is not provided, then the default filesystem file manager
// will be used
*/
typedef char			*(*UPI_ConcatFileIDFn)( CPI_PrivateData	 handle,
												const char		*dirID,
												const char		*fileID );


/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData - This is an opaque (to rayz) pointer which will be passed
//					  into to all of the file manager functions.  This should
//					  allow you to carry around any necessary information
//					  about a filesystem( i.e: an ftp connection, etc.. ).
//  char		**dirID - Destination string for dir part: passed in NULL,
//                         alloc with cpiAlloc - will be freed by caller
//  char		**fileID - Destination string for file part: passed in NULL,
//                         alloc with cpiAlloc - will be freed by caller
//  const char	 *fileAndDirID - the file ID that is to be split into a directory
//                             and a file (if possible)
//
// Return Value:
//  CPI_Bool - returns whether the file was able to be split into component
//             parts. For example, if /usr/tmp/foobar is passed in,
//             this should split it into "/usr/tmp" and "foobar" returning
//             CPI_TRUE, but if "/" is passed in (or something like "D:\" under NT),
//             it should fail, returning CPI_FALSE. If "/usr" (or "/usr/") is
//             passed in, should return "/" and "usr" (a directory can be
//             considered a file ID)
//
// Undefined behavior:
//  If this function is not provided, then the default filesystem file manager
// will be used
*/
typedef CPI_Bool		(*UPI_SplitFileIDFn)( CPI_PrivateData	 handle,
											  char			   **dirID,
											  char			   **fileID,
											  const char		*fileAndDirID );


/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//  char		**fileType - Destination string for file type: passed in NULL,
//                         alloc with cpiAlloc - will be freed by caller
//  const char	 *fileID   - the file ID that is to be split into a directory
//                           and a file (if possible)
//
// Return Value:
//  CPI_Bool - returns whether the file type was able to be determined. The
//             file type is most commonly known as file extension - i.e.
//             "foobar.tiff" would return "tiff" as the fileType. Some
//             filemanagers may not have file names with actual extensions
//             and will insert those as appropriate - i.e. older Accom
//             devices just had frame numbers as the file names, but there
//             is still a file type associated with that file.
//
// Undefined behavior:
//  If this function is not provided, then the default filesystem file manager
// will be used
*/
typedef CPI_Bool		(*UPI_GetFileTypeFn)( CPI_PrivateData	  handle,
											  char				**fileType,
											  const char		 *fileID );


/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData - This is an opaque (to rayz) pointer which will be passed
//		  		  	  into to all of the file manager functions.  This should
//					  allow you to carry around any necessary information
//					  about a filesystem( i.e: an ftp connection, etc.. ).
//	char **newID	- Destination string to place the changed string into
//					  It's passed in as NULL and shouldbe alloc'ed with
//					  cpiAlloc - It will be freed by the caller
//	const char  *oldID	- The File Id which the filetype should be changed for.
//						  If you cannot determine the file type you should
//						  append the new file type.
//	const char	*newType - The File type which 'oldID' should be changed to
//						   This will be passed in as the file extension - i.e.
//						   "foobar.tiff" would pass in "tiff" as the type
//	const char	*oldType - Rayz will ask your plugin for what type it thinks
//						   'oldID' is and compares it against the list of
//						   types it knows about, if it matches, that file type
//						   will be passed in here, otherwise NULL will be
//						   passed and you will be expected to append the new
//						   type to oldID.
// Return Value:
//	CPI_Bool - returns whether the file type was able to be changed.  This is
//			   just a string mangling function.  No change should be made to
//			   any file on disk.  The user has just changed the filetype he
//			   wants to write out, so we want to make the 'fileID' match what
//			   he's selected
//
// Undefined behavior:
//	If this functions is not provided, then the default filesystem file manager
//	will be used
*/
typedef CPI_Bool		(*UPI_ChangeFileTypeFn)( CPI_PrivateData 	handle,
												 char		 	   **newID,
												 const char			*oldID,
												 const char			*newType,
												 const char			*oldType );
/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData - This is an opaque (to rayz) pointer which will be passed
//					  into to all of the file manager functions.  This should
//					  allow you to carry around any necessary information
//					  about a filesystem( i.e: an ftp connection, etc.. ).
//  const char *fileID - File ID for information to be retrieved. Commonly,
//                       if the file ID is a sequence of files, then the disk
//						 would be scanned for the sequence info, and any other
//                       relevant, common info. If the fileID is a single
//                       frame, then other info like the file size might be
//                       returned.
//
// Return Value:
//  CPI_Metadata - metadata that can't necessarily be used to open a file with
//                 a device, but can be used to provide information to the
//                 user, or to intialize values, like the sequence range for
//                 a sequence of frames or stat information ( last modified
//				   date, file size, etc.. )
*/
typedef CPI_Metadata	 (*UPI_GetFileIDInfoFn)( CPI_PrivateData	 handle,
												 const char		*fileID );

/*
////////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//  NONE
//
// Return Value:
//  const char * - an internal string which will not be freed, representing
//                 the default device name for this file manager plugin.
//                 For example, the default filesystem File Manager will
//                 return the default filesystem device plugin "filesystem"
*/
typedef const char		*(*UPI_GetDeviceNameFn)( void );

/*
////////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData - This is an opaque (to rayz) pointer which will be passed
//					  into to all of the file manager functions.  This should
//					  allow you to carry around any necessary information
//					  about a filesystem( i.e: an ftp connection, etc.. ).
//  const char *dirID - directory identifier to be created. If any intermediate
//                      directories need to be created, then they should be
//                      created as well.
//	CPI_Bool createParent - If this variable is CPI_TRUE, then you should
//							create all directories leading up to the
//							specified directory.
//
// Return Values:
//  CPI_Bool - indicating whether the operation was successful or not.
*/
typedef CPI_Bool		 (*UPI_CreateDirIDFn)( CPI_PrivateData	 handle,
											   const char		*newdirID,
											   CPI_Bool			 createParent);

/*
////////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData - This is an opaque (to rayz) pointer which will be passed
//					  into to all of the file manager functions.  This should
//					  allow you to carry around any necessary information
//					  about a filesystem( i.e: an ftp connection, etc.. ).
//  const char *dirID - directory identifier to be removed. That director, and
//                      all its contents, should be removed.
//
// Return Values:
//  CPI_Bool - indicating whether the operation was successful or not.
*/
typedef CPI_Bool		 (*UPI_RemoveDirIDFn)( CPI_PrivateData	 handle,
											   const char		*dirID );

/*
////////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData - This is an opaque (to rayz) pointer which will be passed
//					  into to all of the file manager functions.  This should
//					  allow you to carry around any necessary information
//					  about a filesystem( i.e: an ftp connection, etc.. ).
//  const char *fileID - fileID to be removed. Remember that this can represent
//                       a sequence, so it is possible to pass in something
//                       like "foobar.$F.tiff", and in terms of a perl-like
//                       regular expression, "foobar.[1-9]?[1-0]*.tiff" should
//                       be removed.
//
// Return Values:
//  CPI_Bool - indicating whether the operation was successful or not.
*/
typedef CPI_Bool		 (*UPI_RemoveFileIDFn)( CPI_PrivateData	 handle,
												const char		*fileID );


/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData - This is an opaque (to rayz) pointer which will be passed
//					  into to all of the file manager functions.  This should
//					  allow you to carry around any necessary information
//					  about a filesystem( i.e: an ftp connection, etc.. )
//	const char *oldID - source fileID.
//	const char *newID - target fileID.
//
// Return Value:
//	CPI_Bool - indicating whether the operation was successful or not.
*/
typedef CPI_Bool		 (*UPI_RenameFileIDFn)( CPI_PrivateData	 handle,
												const char		*oldID,
												const char		*newID );


/*
///////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	CPI_PrivateData - This is an opaque (to rayz) pointer which will be passed
//					  into to all of the file manager functions.  This should
//					  allow you to carry around any necessary information
//					  about a filesystem( i.e: an ftp connection, etc.. )
//	const char *fileID - This is the fileID you'd like to create a temporary
//						 name for.
//
// Return Value:
//	char *tmpFileID - This is the new fileID that rayz will use to create a
//					  a temporary file during renders.  The temporary file will
//					  be used to render into, and if there are no rendering
//					  errors, it will be renamed to the correct fileID.
//					  Therefore rendering errors will not cause the destruction
//					  of old files..
//					  NOTE: This char * should be cpiAlloc'ed in here and will
//							be freed internally by Rayz
//					  NOTE: If you don't have a renameID function this function
//					 		will NOT be called..
*/
typedef	char			*(*UPI_TemporaryFileIDFn)( CPI_PrivateData	 handle,
												   const char		*fileID );



/*
////////////////////////////////////////////////////////////////////////////////
//
// Arguments:
//	None;
//
// Return Value:
//  char * - a string which should no longer be referenced upon return, and
//           will be freed by the consumer, representing the Dir ID to
//           'start' at - for example, the Unix filesystem will return
//           the results from getcwd()...
//
//           Should be alloced with cpiAlloc - will be freed with cpiFree
//
// Undefined behavior:
//  If this function is not provided, then the default filesystem file manager
// will be used to do the operation
*/
typedef char			*(*UPI_GetDefaultDirIDFn)( void );


/*
// For any optional (upi labeled) functions you don't wish to provide,
// fill in NULL
*/
typedef struct
{
	RPI_PluginInfo			rpiPluginInfo;
	UPI_InitFn				upiInit;
	UPI_ShutDownFn			upiShutDown;
	RPI_GetFileIDListFn		rpiGetFileIDList;
	RPI_GetMetadataToOpenFn	rpiGetMetadataToOpen;
	UPI_FillSeqInfoFn		upiFillSeqInfo;
	UPI_ConvertFileIDFn		upiConvertFileID;
	UPI_ConcatFileIDFn		upiConcatFileID;
	UPI_SplitFileIDFn		upiSplitFileID;
	UPI_GetFileTypeFn		upiGetFileType;
	UPI_ChangeFileTypeFn	upiChangeFileType;
	UPI_GetFileIDInfoFn		upiGetFileIDInfo;
	UPI_GetDeviceNameFn		upiGetDeviceName;
	UPI_CreateDirIDFn		upiCreateDirID;
	UPI_RemoveDirIDFn		upiRemoveDirID;
	UPI_RemoveFileIDFn		upiRemoveFileID;
	UPI_RenameFileIDFn		upiRenameFileID;
	UPI_TemporaryFileIDFn	upiTemporaryFileID;
	UPI_GetDefaultDirIDFn	upiGetDefaultDirID;
} RPI_FMInfo;

#ifdef __cplusplus
}
#endif

#endif
