/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 Seward St
//  Hollywood, CA
//  USA 90038
//  323-871-9100
*/

#ifndef _CPI_Util_h_
#define _CPI_Util_h_

#ifdef __cplusplus
extern "C" {
#endif


/*******************************************************************************
 ** Error Reporting
 ******************************************************************************/


CPIDSOIMPORT void				 cpiError( const char *errormsg );
CPIDSOIMPORT void				 cpiWarning( const char *warningmsg );


/*******************************************************************************
 ** Memory Management tools
 ******************************************************************************/


CPIDSOIMPORT void				*cpiInternalAlloc( size_t		 size,
												   const char	*file,
												   int			 line );
CPIDSOIMPORT void				 cpiFree( void *ptr );
CPIDSOIMPORT void				*cpiInternalReAlloc( void			*ptr,
													 size_t			 size,
													 const char		*file,
													 int			 line);
CPIDSOIMPORT void				 cpiDoMemoryTracking( CPI_Bool on );
CPIDSOIMPORT void				 cpiMemoryReport( void );

/* Defines so you don't have to type __FILE__ and __LINE__... */
#define cpiAlloc(size) cpiInternalAlloc(size, __FILE__, __LINE__)
#define cpiReAlloc(ptr, size) cpiInternalReAlloc( ptr, size, __FILE__, __LINE__)

/*******************************************************************************
 ** Machine related info
 ******************************************************************************/

/*
 * Returns whether the current architecture is little endian
 * indicating that you may have to byte swap data during a file read
 */
CPIDSOIMPORT CPI_Bool			 cpiIsLittleEndian( void );


/*******************************************************************************
 ** Local filesystem searching and other info
 ******************************************************************************/

/*
 * Returns the full path to the shared library the specified plugin was
 * defined in
 */
CPIDSOIMPORT const char			*cpiGetLibraryName( const char *pluginName );

/*
 * Returns the directory in which the library which declared the
 * specified plugin was found
 */
CPIDSOIMPORT const char			*cpiGetLibraryPath( const char *pluginName );

/*
 * Searches the local file system to find things like configuration files
 * and such that need to be present on the filesystem, meaning they are
 * only available from the filesystem device readers, and not useable
 * from things like an ftp device reader
 *
 * The recurse flag specifies that at each path entry, it should recurse
 * sub directories, looking for the pattern
 */
#define CPI_SEARCH_PLUGIN_PATH	0
#define CPI_SEARCH_SYSTEM_PATH	1
CPIDSOIMPORT CPI_Bool			 cpiFindLocalFile( char			*filePath,
												   CPI_Uint32	 bufSize,
												   const char	*pattern,
												   CPI_Int32	 whichPath,
												   CPI_Bool		 recurse );

/*
 * Fills in the tmpdir argument with the current temporary directory path
 * which is unique per process - it is a subdirectory of the main temp
 * path
 */
CPIDSOIMPORT void				 cpiGetTmpDirectory( char		*tmpdir,
													 CPI_Uint32	 size );

CPIDSOIMPORT CPI_Bool			 cpiPatternMatch( const char *stringtomatch,
												  const char *pattern );

/*******************************************************************************
 ** Specify icons for the user interface
 ******************************************************************************/

/*
 * Defines a set of icons that have the name of form:
 *
 * "baseName/Disabled"
 * "baseName/Located"
 * "baseName/Normal"
 * "baseName/Selected"
 * "baseName/SelectedDisabled"
 * "baseName/SelectedLocated"
 *
 * Any of the files can be NULL, meaning there is no icon defined for that
 * widget state. If an icon is NULL, Rayz will default to using the
 * normalFile icon if defined
 *
 * Rayz searches for icons in the path controlled by the "Theme Search Path"
 * preference, and will look for the icon in <path>/Icons/<baseName>
 *
 * The image must be of a format Rayz can read, 8 bit, 3 channels
 *
 */
CPIDSOIMPORT void				 cpiDefineIcon( const char	*baseName,
												const char	*normalFile,
												const char	*locatedFile,
												const char	*disabledFile,
												const char	*selectedFile,
												const char	*selectedLocatedFile,
												const char	*selectedDisabledFile,
												CPI_Uint8	 transColorRed,
												CPI_Uint8	 transColorGreen,
												CPI_Uint8	 transColorBlue );

/*
 * Used in conjunction with the above cpiDefineIcon, to give Rayz compiled in
 * data for an icon. This is often used to avoid having to distribute
 * an icon file, and just having it compiled in to the program.
 *
 * The name should be the same as the filename given in cpiDefineIcon
 * i.e. if you specify foobar.tiff for the normalFile, foobar.tiff should
 * be used as the name here. This is so the user can opt to override icons
 * if she wants to, just as they are able to change the user interface
 * colors...
 *
 * The iconData that is passed in must be sizeX x sizeY x 3 channels
 * interleaved RGB
 */
CPIDSOIMPORT void				 cpiRegisterIconData( const char		*name,
													  CPI_Int32			 sizeX,
													  CPI_Int32			 sizeY,
													  const CPI_Uint8	*iconData,
													  CPI_Uint8			 transColorRed,
													  CPI_Uint8			 transColorGreen,
													  CPI_Uint8			 transColorBlue );


/*******************************************************************************
 ** Miscellaneous Image related functions
 ******************************************************************************/

CPIDSOIMPORT void				 cpiApplyLut8( CPI_Image			*image,
											   const CPI_Uint8		*lut,
											   CPI_Int8				 channel );
// lut of size 65536
CPIDSOIMPORT void			 	 cpiApplyLut16( CPI_Image			*image,
												const CPI_Uint16	*lut,
												CPI_Int8			 channel );

CPIDSOIMPORT void				*cpiGetLine( CPI_Image	*image,
											 CPI_Int32	 linenum );

CPIDSOIMPORT CPI_Bool			 cpiIsProcessingAborted( void );
	
CPIDSOIMPORT void				 cpiKernelToString( const CPI_Float32 *kernel,
													const CPI_Int32 xsize,
													const CPI_Int32 ysize,
													char *string,
													const CPI_Int32 maxstring );

CPIDSOIMPORT CPI_Uint32			 cpiRound( CPI_Float32 sample );

#define cpiClampFloat(v, min, max) ( v < min ? min : v > max ? max : v )
#define cpiClampInt(v, min, max) ( v < min ? min : v > max ? max : v )
#define cpiSwapShort(in) ( ( in & 0x00ff ) << 8 ) | ( (in >> 8 ) & 0x00ff )
#define cpiSwapLong(in) ( ( ( in & 0x000000ff ) << 24 ) | \
						  ( ( in & 0x0000ff00 ) << 8 ) | \
						  ( ( in & 0x00ff0000 ) >> 8 ) | \
						  ( ( in & 0xff000000 ) >> 24 ) )

#define cpiSwap64(in)   ( ( ( in & 0x00000000000000ffLL ) << 56 ) | \
                          ( ( in & 0x000000000000ff00LL ) << 40 ) | \
                          ( ( in & 0x0000000000ff0000LL ) << 24 ) | \
                          ( ( in & 0x00000000ff000000LL ) << 8  ) | \
                          ( ( in & 0x000000ff00000000LL ) >> 8  ) | \
                          ( ( in & 0x0000ff0000000000LL ) >> 24 ) | \
                          ( ( in & 0x00ff000000000000LL ) >> 40 ) | \
                          ( ( in & 0xff00000000000000LL ) >> 56 ) )

CPIDSOIMPORT CPI_Float32 cpiSwapFloat( CPI_Float32 in );


/*******************************************************************************
 ** Misc functions to work with node execution
 ******************************************************************************/

CPIDSOIMPORT CPI_ImageOp		cpiGetInputOp( CPI_Uint32	input,
											   CPI_Float32	theTime,
											   CPI_Uint8	quality,
											   CPI_Float32	scaleX,
											   CPI_Float32	scaleY );
CPIDSOIMPORT CPI_ImageOp		cpiAddImageOp( const char	*opName,
											   CPI_Metadata	 parms,
											   CPI_ImageOp	 inputs[],
											   CPI_Uint32	 numinputs );

/******************************************************************************
 ** Useful image process control functions..
 ******************************************************************************/

CPIDSOIMPORT CPI_Bool	 	cpiCompute( CPI_Bool horizontal );
CPIDSOIMPORT CPI_Bool		cpiIntersectRegions( const CPI_ImageContext  one,
												 const CPI_ImageContext	 two,
												 CPI_ImageContext		*result );
CPIDSOIMPORT CPI_Bool		cpiMergeRegions( const CPI_ImageContext	 one,
											 const CPI_ImageContext	 two,
											 CPI_ImageContext		*result );
CPIDSOIMPORT CPI_Bool		cpiCopyRegion( CPI_ImageContext	*regionToCopy,
										   const CPI_Image	*from );
CPIDSOIMPORT CPI_Bool		cpiCopyOutsideRegion( CPI_ImageContext	*regionToSkip,
												  const CPI_Image	*from );
CPIDSOIMPORT CPI_Bool		cpiSetColorOutsideRegion( CPI_ImageContext	*regionToSkip,
													  CPI_Float32		 r,
													  CPI_Float32		 g,
													  CPI_Float32		 b,
													  CPI_Float32		 a );

CPIDSOIMPORT CPI_Bool		cpiSetColorInsideRegion( CPI_ImageContext	*regionToFill,
													 CPI_Float32		 r,
													 CPI_Float32		 g,
													 CPI_Float32		 b,
													 CPI_Float32		 a );

/******************************************************************************
 ** Useful function for script option plugins
 ******************************************************************************/

CPIDSOIMPORT const char		*cpiGetUniqueNodeName( void );

#ifdef __cplusplus
}
#endif

#endif /* _CPI_Util_h_ */

